import collections
import glob
import json
import re
import sys

"""
    {
        "event": "start",
        "http_host": "www.bellyfashion.nl",
        "method": "GET",
        "nginx_start": "1408694999",
        "pid": 15631,
        "remote_addr": "64.39.102.103",
        "time": 2.00940203666687,
        "timestamp": 1408695001.3356,
        "uri": "/l-voedingstuniek-daniella-olive",
        "useragent": "Mozilla/5.0 (compatible; AhrefsBot/5.0; +http://ahrefs.com/robot/)",
        "worker_status": "GONE"
    },
"""


def main():  # pragma: nocover
    try:
        incident_dir = sys.argv[1]
    except IndexError:
        print("Usage: {} <incident directory (quoted)>".format(sys.argv[0]))
        sys.exit(1)

    analysis, description = analyze(incident_dir)
    print("Analysis: {}".format(analysis))
    print("Description: {}".format(description))
    print("")
    print(json.dumps(get_oldest_fpm_status(incident_dir), indent=4))


def analyze(incident):
    """
    Analyze the earliest FPM status file and return a conclusion about the disruption cause.

    :param str incident: the path to the directory containing the incident logs
    :return: a tuple containing the incident category, a human-readable description about the
             incident and the FPM status itself as a dict
    :rtype: (string, string)
    """
    fpm_status = get_oldest_fpm_status(incident)
    remoteaddr = collections.Counter([worker.get("remote_addr") or '-' for worker in fpm_status])
    useragents = collections.Counter([worker.get("useragent") or '-' for worker in fpm_status])

    most_frequent_agent, most_frequent_agent_count = useragents.most_common(1)[0]
    most_frequent_agent_match = re.search(r'([\w-]*(?:bot|href)[\w-]*)',
                                          most_frequent_agent, re.I)  # add more search terms here

    if most_frequent_agent_match:
        return "bot", analyze_bot(fpm_status, most_frequent_agent, most_frequent_agent_match.group(1))

    # TODO: must have at least X requests running, so only "if most_frequent_address_count > X"
    (most_frequent_address, most_frequent_address_count) = remoteaddr.most_common(1)[0]
    return "topuser", most_frequent_address


def analyze_bot(status, useragent, match):
    ips = sorted(list(set(request.get('remote_addr')
                          for request in status
                          if request.get('useragent') == useragent)))
    return {
        "useragent_match": match,
        "useragent": useragent,
        "remote_address": ips,
    }


def lookup_oldest_fpm_status(incident):
    """
    Lookup the earliest FPM status JSON file in the incident directory

    :param incident: the path to the directory containing the incident logs
    :return: the name of the oldest incident JSON fpm status log
    :rtype: string
    """
    return sorted(glob.glob("{}/*-fpm-status.json".format(incident)))[0]


def get_oldest_fpm_status(incident):
    """
    Parse the earliest FPM status JSON file in the incident directory.

    :param incident: the path to the directory containing the incident logs
    :return: the parsed FPM status
    :rtype: list[dict]
    """
    fpm_status_log = lookup_oldest_fpm_status(incident)
    with open(fpm_status_log) as fh:
        return json.loads(fh.read())
