import argparse
import collections
import json
import os.path
import re
import sys

NGINX_LOG_DIR = '/var/log/nginx'
EMAIL_RE = re.compile(r'[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+')
URL_RE = re.compile(r'((http[s]?://)|(www.))(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
FILTERS = ["chrome", "edge", "msie", "firefox", "safari", "opera", "windows",
           "win", "iphone", "ipad", "mac", "android", "iceweasel", "linux", "x11"]


def get_default_nginx_log_path():
    return os.path.join(NGINX_LOG_DIR, 'access.log')


def parse_arguments(cmdline=None):
    parser = argparse.ArgumentParser(description="Analyzes the access.log for bots. Returning bots with the count "
                                                 "of the frequency in which they occurred.")
    return parser.parse_args(cmdline)


def main():
    parse_arguments()

    bots_found = get_bots_from_logs()

    counts_per_bot = collections.Counter(bots_found).most_common(15)

    print_bots_result(dict(counts_per_bot))


def get_bots_from_logs():
    for accesslog_line in get_accesslog_lines():
        botname = get_bot_from_line(accesslog_line)
        if botname is not None:
            yield botname


def get_accesslog_lines():
    with open(get_default_nginx_log_path(), "r", errors='ignore') as in_file:
        lines = in_file.read().splitlines()
        return lines


def get_bot_from_line(accesslog_line):
    match = re.search(r'user_agent":"(.*?)"', accesslog_line)

    if not match:
        return

    useragent = match.group(1)
    if _is_bot_useragent(useragent):
        return useragent


def _is_bot_useragent(useragent):
    if EMAIL_RE.search(useragent) or URL_RE.search(useragent):
        return True

    for keyword in FILTERS:
        if keyword in useragent.lower():
            return False

    return True


def print_bots_result(all_bots):
    sys.stdout.write(json.dumps(all_bots))


if __name__ == '__main__':
    main()
