import os
import json
from time import time
from collections import defaultdict
from glob import glob
from contextlib import suppress
from logging import getLogger

import GeoIP
from psutil import Process, NoSuchProcess, AccessDenied
from kamikaze3.common.filesystem import read_first_existing_file, read_last_lines


gi = GeoIP.new(GeoIP.GEOIP_MEMORY_CACHE)
logger = getLogger(__name__)


def determine_worker_status(proc):
    """
    Get status of a FPM worker process

    The status is based on the state of the TCP connection it has to port 9000. If the process is
    idle, there will only be one connection in the LISTEN state. If the process is busy it has more
    connections. When the connection is CLOSE_WAIT, the client has already left but the process
    doesn't realise it yet or keeps working without closing its side of the connection.

    :param psutil.Process proc: Worker process object
    :rtype: str
    """
    try:
        conns = [conn for conn in proc.connections() if conn.laddr == ('127.0.0.1', 9000)]
    except (NoSuchProcess, AccessDenied):
        # Proc is already gone
        return '-'

    if any(conn.status == 'ESTABLISHED' for conn in conns):
        return 'BUSY'
    if any(conn.status == 'CLOSE_WAIT' for conn in conns):
        return 'GONE'
    return 'IDLE'


def worker_as_text(worker):
    """Return the status of the worker as a line of text

    :param dict worker: last request info dict of the worker
    :return: line of the worker status
    :rtype: str
    """
    # Python 3.6 str.format(**kwargs) is not playing well with defaultdict, still
    # rasing KeyErrors on missing keys.
    # So instead of a defaultdict, we'll use a normal dict and ensure required keys
    # exist and are empty string by default.
    safe_worker = dict()
    safe_worker.update(worker)

    for key in ('pid', 'worker_status', 'time', 'remote_addr', 'method', 'http_host',
                'uri', 'useragent', 'country'):
        if key not in safe_worker or safe_worker[key] is None:
            safe_worker[key] = ''

    with suppress(ValueError, TypeError):
        safe_worker['time'] = "{:5.1f}".format(safe_worker['time'])

    text = "{pid:05d} {worker_status:4.4s} {time:5s}s {country:2.2s} ".format(**safe_worker)
    return text + "{remote_addr:15s} {method:4s} {http_host}{uri}   ({useragent})".format(**safe_worker)


class FpmSource(object):
    """The source of information regarding FPM processes"""

    def __init__(self):
        self.pid_base_path = '/var/run'
        self.log_base_path = '/var/log/hypernode/fpm-inspector'

    def _pid_path(self, sub_path):
        """Return the absolute path to the provided subpath from PID base path.
        Note: Does not check if the path exists.

        :param sub_path: sub path under the base PID path
        :rtype: str
        """
        return os.path.abspath(os.path.join(self.pid_base_path, sub_path))

    def get_possible_fpm_pidfiles(self):
        return glob(self._pid_path('php/php[0-9].[0-9]-fpm.pid'))

    def get_fpm_process(self):
        """Get master FPM process
        :rtype: psutil.Process
        """
        with suppress(IOError, NoSuchProcess, ValueError):
            pid = int(read_first_existing_file(self.get_possible_fpm_pidfiles()))
            return Process(pid)
        return None

    def get_process_last_logs(self, pid, max_logs=2, line_size=1024):
        """Get the last logs of an FPM process
        Return None if failed to read the log file

        :Note: the estimated line_size is 1KB, but it may happen to have larger
        useragents. So we may need to increase the line size in future.

        :param pid: The pid of the process
        :param max_logs: max number of logs
        :param line_size: estimated size of each line
        :return: The last logs
        :rtype: list[dict[str, V]] | None
        """
        filename = os.path.join(self.log_base_path, '{}.log'.format(pid))
        block_size = (max_logs * line_size) + max_logs  # consider EOL chars

        last_lines = read_last_lines(filename, max_logs, block_size)
        if last_lines is None:
            return None

        def try_parse_json(line):
            with suppress(ValueError, UnicodeDecodeError):
                return json.loads(line)
            return None

        last_logs = [try_parse_json(line) for line in last_lines]
        return list(filter(None, last_logs))

    def get_last_request_info(self, pid):
        """Get details of the last request event of an FPM worker.

        :param int pid: Process ID of the FPM worker
        :return: Dictionary containing all kinds of request properties
        :rtype: dict[str, V]
        """
        logs = self.get_process_last_logs(pid, 2)
        request_info = defaultdict(lambda: '-')
        if not logs:
            return request_info

        """
        Fields come from: https://github.com/ByteInternet/hypernode-control/blob/master/ansible-playbooks/roles/magweb/files/fpm-inspector/prepend.php
            timestamp = $_SERVER['REQUEST_TIME_FLOAT']

        # last is an end-request line
        """

        last_log = logs[-1]
        last_log_timestamp = last_log.get('timestamp')
        if last_log['event'] == 'start':
            # Last request has not finished yet
            last_log['time'] = '-' if (last_log_timestamp is None) else (time() - last_log_timestamp)
        elif len(logs) > 1:
            previous_log = logs[-2]
            assert previous_log['event'] == 'start'
            previous_log_timestamp = previous_log.get('timestamp')
            last_log['time'] = '-' if (previous_log_timestamp is None) else (last_log_timestamp - previous_log_timestamp)
        else:
            # Apparently we only have the "end" event in the log file, so can't calculate the time
            last_log['time'] = '-'

        for key, value in last_log.items():
            request_info[key] = value if value is not None else '-'
        return request_info

    def get_scoreboard(self):
        """Return the information on last requests of all worker process,
        sorted by fastest to slowest.

        :rtype: list[dict]
        """
        master = self.get_fpm_process()
        if master is None:
            logger.warning("Unable to get FPM status. Could not find master process!")
            return []

        result = []
        for worker in master.children():
            request = self.get_last_request_info(worker.pid)
            request['pid'] = worker.pid
            request['worker_status'] = determine_worker_status(worker)
            request['country'] = '-'
            if 'remote_addr' in request and request['remote_addr']:
                with suppress(Exception):
                    country = gi.country_code_by_addr(request['remote_addr'])
                    if country:
                        request['country'] = country
            result.append(request)

        # Python 3 does not support sorting of heterogeneous collections.
        # 'time' is numeric or '-' if unclear, for sorting let's assume '-' is 0
        result.sort(key=lambda x: 0 if x['time'] == '-' else x['time'], reverse=True)
        return result

    def get_status_report(self):
        return "\n".join(map(worker_as_text, self.get_scoreboard()))

    def get_status_report_json(self):
        return json.dumps(self.get_scoreboard(), indent=4, sort_keys=True)


def print_fpm_report():
    print(FpmSource().get_status_report())


def print_fpm_report_json():
    print(FpmSource().get_status_report_json())
