import argparse
import os
import pwd
import logging
import logging.handlers
from contextlib import suppress

MB = 2 ** 20


def configure_logging():
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)

    console = logging.StreamHandler()

    formatter = logging.Formatter(__file__ + ': %(message)s')
    syslog = logging.handlers.SysLogHandler(address='/dev/log')
    syslog.setFormatter(formatter)
    syslog.setLevel(logging.INFO)

    log.addHandler(console)
    log.addHandler(syslog)

    return log

logger = configure_logging()


def purge(files, space_to_purge, dry_run=False):
    """
    Loop through an already sorted array of files, and purge first
    until space threshold (MB) is reached.
    """

    space_to_purge *= MB
    bytes_already_deleted = 0

    for f in files:
        if bytes_already_deleted >= space_to_purge:
            break

        if f[1].st_uid == 0:
            raise RuntimeError("Refusing to delete root files (%s)" % f[0])

        filename = f[0]
        filesize = f[1].st_size
        filetime = f[1].st_mtime
        owner = f[1].st_uid

        bytes_still_required = space_to_purge - bytes_already_deleted - filesize

        logger.info("[ %(filetime)d / %(filesize)8d bytes / uid %(owner)s ] will unlink %(filename)s (then still %(bytes_still_required)d bytes to go)" % locals())

        if dry_run:
            logger.info("Dry run, but would delete: %s" % filename)
        else:
            try:
                os.unlink(filename)
            except OSError as e:
                logger.error("Cant unlink %s due to %s, will skip" % (filename, e.strerror))
                continue

        bytes_already_deleted += filesize

    return bytes_already_deleted


def check_and_purge(path, minfree, user, dry_run=False):
    """
    Purge files owned by user, oldest files first.

    First, check whether it is at all possible to achieve minfree available space (if not, bail out).
    """

    total, avail = df(path)
    logger.debug("Free: %s MB, need: %s MB" % (avail, minfree))

    if minfree < avail:
        # everything ok!
        logger.debug("Everything ok, no need to purge)")
        return

    space_to_purge = minfree - avail

    files = find_user_files(path, user)
    space_in_use = sum([file[1].st_size for file in files]) / MB

    logger.debug("Need to purge %s MB, in use by this user is %s MB" % (space_to_purge, space_in_use))

    if space_in_use < space_to_purge:
        logger.error("Cannot purge the required %(space_to_purge)s MB, there was only %(space_in_use)s MB found for user %(user)s in %(path)s. Other users taking up storage? Storage taken up elsewhere?" % locals())
        return

    files = sorted(files, key=lambda x: x[1].st_mtime)
    purge(files=files, space_to_purge=space_to_purge, dry_run=dry_run)


def find_all_files_in_path(path):
    return [os.path.join(dirpath, filename)
            for (dirpath, dirs, files) in os.walk(path, followlinks=False)
            for filename in files]


def try_lstat(filename):
    """
    lstat file, return None if it's already gone
    :param filename:
    :return: obj stat_result | None
    """
    with suppress(OSError):
        return os.lstat(filename)


def list_file_and_stat_information(filenames):
    return list(filter(all, [(filename, try_lstat(filename)) for filename in filenames]))


def filter_root_files(files, user):
    uid = pwd.getpwnam(user).pw_uid
    if user != 'root' and uid == 0:
        raise RuntimeError("Refusing to delete uid 0 files")
    files = [file for file in files if file[1].st_uid == uid]
    return files


def find_user_files(path, user=None):
    """ find $path -type f """

    filenames = find_all_files_in_path(path)

    files = list_file_and_stat_information(filenames)

    if user:
        files = filter_root_files(files, user)

    return files


def df(path):
    """
    posix.statvfs_result(f_bsize=4096, f_frsize=4096, f_blocks=1224100L,
    f_bfree=97845L, f_bavail=35663L, f_files=311296L, f_ffree=153367L,
    f_favail=153367L, f_flag=4096, f_namemax=255)
    """
    stat = os.statvfs(path)

    total = int(stat.f_bsize * stat.f_blocks / MB)
    avail = int(stat.f_bsize * stat.f_bavail / MB)

    return total, avail


def main():
    parser = argparse.ArgumentParser(
        description="Purge old tmp files to ensure minimum available space.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--minfree', type=int, required=True, help="How many MB should be free?")
    parser.add_argument('--path', default='/tmp', help='Where to clean')
    parser.add_argument('--user', default='app', help='Restrict to files owned by this user')
    parser.add_argument('--dry-run', default=False, action="store_true", help="Just print")
    args = parser.parse_args()

    check_and_purge(
        path=args.path,
        minfree=args.minfree,
        user=args.user,
        dry_run=args.dry_run,
    )

if __name__ == '__main__':
    main()
