"""
common.api
==========
Abstraction to HNAPI
"""

import logging
import requests
import os
import pwd
from requests import HTTPError
from kamikaze3.common.filesystem import read_json_file

log = logging.getLogger(__name__)
HYPERNODE_API_URL = 'https://api.hypernode.com'
SYSTEMCTL_USER_AGENT = 'hypernode-systemctl'


def get_hypernode_api_token():
    """DEPRECATED: please use get_app_api_token() instead

    @TODO: replace usages of this function with get_app_api_token
    """
    try:
        with open('/etc/hypernode/hypernode_api_token', 'r') as f:
            hypernode_api_token = f.readline()
    except FileNotFoundError:
        log.error('Hypernode API token not provisioned on this node yet.')
        exit(1)
    else:
        return hypernode_api_token.replace('\n', '')


def handle_api_response(response, success_message=None):
    try:
        response.raise_for_status()
    except HTTPError:
        log.error('Looks like something went wrong: {}'.format(response.text))
        return False
    else:
        if success_message:
            log.info(success_message)
        return True


def get_user_api_token() -> str:
    """Return the support API token if the support or root user is calling this,
    otherwise the app token
    NOTE: This deprecates "get_hypernode_api_token" because:
      - it reads new app.json which contains the token instead of /etc/hypernode/hypernode_api_token
      - it supports reading the support token
      - it handles errors better (a function that exits is not reusable)

    :raise: RuntimeError on missing API token from proper accessible app.json
    """
    username = pwd.getpwuid(os.getuid()).pw_name
    if username in ('support', 'root'):
        return get_support_api_token()
    else:
        return get_app_api_token()


def get_app_api_token() -> str:
    """Return Hypernode API token from app.json
    :raise: RuntimeError on missing API token from proper accessible app.json
    """
    log.debug("reading app.json to get Hypernode API token ...")
    try:
        app_json = read_json_file("/etc/hypernode/app.json")
        if "hypernode_api_token" not in app_json:
            raise RuntimeError("Key 'hypernode_api_token' is missing from '/etc/hypernode/app.json'")
    except RuntimeError as err:
        log.error("no api token in app.json: {}".format(err))
        raise err

    return str(app_json["hypernode_api_token"]).strip()


def get_support_api_token() -> str:
    """Return Hypernode API token from support.json

    :raise: RuntimeError on missing API token from proper accessible app.json
    """
    log.debug("reading support.json to get Hypernode API token ...")
    try:
        support_json = read_json_file("/etc/hypernode/support.json")
        if "hypernode_api_support_token" not in support_json:
            raise RuntimeError("Key 'hypernode_api_support_token' is missing from '/etc/hypernode/support.json'")
    except RuntimeError as err:
        log.error("no api token in support.json: {}".format(err))
        raise err

    return str(support_json["hypernode_api_support_token"]).strip()


class HttpApiClient:
    default_base_url = HYPERNODE_API_URL  # type: str
    """The default base URL of the API client. override in subclasses to customize"""

    def __init__(self, token: str = None, headers=None):
        """HttpApiClient abstracts accessing HTTP REST API.
        A base class for other API clients, or can be used directly.

        Usage:
        data = HttpApiClient().get("/endpoint/path/")
        data = HttpApiClient().post("/endpoint/path/", data={"key": "value"})

        Custom Client:
        HttpApiClient(token)  # explicit token
        HttpApiClient(token="")  # unauthorized request, no token
        HttpApiClient(headers={"Key": "Value"})  # explicit headers, can override defaults

        :param str token: explicit API token, empty to use no auth token (unauthorized requests)
        :param dict headers: dictionary of extra headers, could override defaults if needed
        """
        self.base_url = self.__class__.default_base_url
        self.token = self._default_api_token() if token is None else token  # type: str
        self.headers = self.default_headers()
        if headers:
            self.headers.update(**headers)
        if "Authorization" not in self.headers and self.token:
            self.headers["Authorization"] = "Token {}".format(self.token)

    @staticmethod
    def _default_api_token():
        """Provide the defualt API token. Override in subclasses to customize the default token"""
        return get_user_api_token()

    @staticmethod
    def default_headers():
        """Return default headers used for all requests, if not overwritten by constructor"""
        return {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": "hypernode-kamikaze3",
        }

    def get(self, path: str):
        """Send a GET request to the specified path in API and return JSON parsed data.

        :param str path: the path to the end point
        :raises: HTTPError
        :return: parsed out data from JSON response
        """
        return self._api_call("GET", path)

    def post(self, path: str, data=None):
        """Send a POST request to the specified path in API and return JSON parsed data.

        :param str path: the path to the end point
        :param dict data: the data to post
        :raises: HTTPError
        :return: parsed out data from JSON response
        """
        return self._api_call("POST", path, data)

    def put(self, path: str, data=None):
        """Send a PUT request to the specified path in API and return JSON parsed data.

        :param str path: the path to the end point
        :param dict data: the data to put
        :raises: HTTPError
        :return: parsed out data from JSON response
        """
        return self._api_call("PUT", path, data)

    def patch(self, path: str, data):
        """Send a PATCH request to the specified path in API and return JSON parsed data.

        :param str path: the path to the end point
        :param dict data: the data to patch
        :raises: HTTPError
        :return: parsed out data from JSON response
        """
        return self._api_call("PATCH", path, data)

    def delete(self, path: str):
        """Send a DELETE request to the specified path in API and return JSON parsed data.

        :param str path: the path to the end point
        :raises: HTTPError
        :return: parsed out data from JSON response
        """
        return self._api_call("DELETE", path)

    def _api_call(self, method: str, path: str, data=None):
        """Call the provided path in the API and return the results

        :param str method: the HTTP method
        :param str path: the extra path of the API (after base URL)
        :param dict data: data to send in request body (not in GET)
        :raises: HTTPError
        :return: parsed out data from JSON response
        """
        url = "{}/{}".format(self.base_url, path)
        log.debug("requesting {} to API {}".format(method, url))
        response = self._request(method, url, data)
        log.debug("response of {} API {}. status: {}".format(method, url, getattr(response, "status_code")))
        response.raise_for_status()
        return response.json()

    def _request(self, method: str, url: str, data=None):
        """Perform an HTTP request to the URL with the method and send the data, returns the response object.

        :raises ValueError: unsupported method not in (get, post, put, delete, patch)
        """
        method = method.strip().upper()
        if method not in ("GET", "POST", "PATCH", "PUT", "DELETE"):
            raise ValueError("Invalid method specified {}".format(method))
        return requests.request(method, url, headers=self.headers, json=data)


class ApiV2Client(HttpApiClient):
    """ApiV2Client abstracts accessing API v2. Could be used directly, or a base class for other API clients.

    High level usage:
        app_info = ApiV2Client().get_app('app_name')
        ...

    Low level usage:
        data = ApiV2Client().get("/appname/extra/path")
        data = ApiV2Client().post("/appname/extra/path", data={"key": "value"})

    Custom Client:
        ApiV2Client(token)  # explicit token
        ApiV2Client(token="")  # unauthorized request, no token
        ApiV2Client(headers={"Key": "Value"})  # explicit headers, can override defaults

    :param str token: explicit API token, empty to use no auth token (unauthorized requests)
    :param dict headers: dictionary of extra headers, could override defaults if needed
    """

    default_base_url = "{}/v2".format(HYPERNODE_API_URL)  # type: str
    dns_record_types = ("A", "CNAME", "TXT")  # matches supported types in API

    def get_app(self, app_name: str):
        """Return app information"""
        app_path = self._app_path(app_name)
        log.debug("getting app info from v2 API")
        return self.get(app_path)

    def get_all_domain_records(self, domain_name: str):
        """Return all the records registered for the given domain

        :param str domain_name: the unique domain name, like hypernode.com
        :return: list[dict] all available records
        """
        api_path = "dns/{}/".format(domain_name)
        return self.get(api_path)

    def get_dns_records_of_type(self, record_name: str, type_: str = "A"):
        """Return all records of the specified types for the given record name

        :param str record_name: the unique record name, like api.hypernode.com
        :param str type_: the record type (A, CNAME, TXT, ...)
        :return: list[dict] all the available records of the specified type
        :raises: ValueError on unsupported record types
        """
        type_ = self._sanitize_dns_record_type(type_)
        api_path = "dns/{}/type/{}".format(record_name, type_)
        return self.get(api_path)

    def set_dns_record_contents(self, record_name: str, type_: str, contents):
        """Set the DNS records for the provided record name and type.
        Accepts a list of contents, and TTLs. Priorities are not supported yet (by the API)

        :param str record_name: the unique record name, like api.hypernode.com
        :param str type_: the record type (A, CNAME, TXT, ...)
        :param list[dict] contents: list of dictionaries, containing "content", "ttl" keys
        :return: ApiV2Client self reference for chain calls
        :raises: ValueError on unsupported record types or invalid contents
        """
        type_ = self._sanitize_dns_record_type(type_)
        for content in contents:
            self._validate_dns_record_content(content)
        api_path = "dns/{}/type/{}".format(record_name, type_)
        self.put(api_path, data=contents)
        return self

    def _app_path(self, app_name: str) -> str:
        """Return the path to API endpoints working with app"""
        return "app/{}".format(app_name)

    @classmethod
    def _sanitize_dns_record_type(cls, type_: str):
        type_ = type_.strip().upper()
        if type_ not in cls.dns_record_types:
            raise ValueError("Unsupported DNS record type '{}'. should be in {}".format(type_, ",".join(cls.dns_record_types)))
        return type_

    @staticmethod
    def _validate_dns_record_content(content):
        """Validate a DNS record content, raise ValueError on invalid data"""
        if not isinstance(content, dict):
            raise ValueError("DNS record content should be a dictionary")
        for key in ("content", "ttl"):
            if key not in content:
                raise ValueError("DNS record is missing '{}' key".format(key))
