import os
import json
import re
import socket
import subprocess
from contextlib import suppress
from http import HTTPStatus
from os.path import devnull, isfile
from random import shuffle
from subprocess import check_output, CalledProcessError

import requests

from kamikaze3.common.settings import COMMON_INFO_PATH, PUBLIC_IP_URLS, MAGWEB_INFO_PATH, APP_INFO_PATH
from kamikaze3.common.system import walk_with_depth

GET_IP_TIMEOUT = 10


def parse_app_name_from_host_name():
    """
    Try parsing the app name from the hostname
    Returns None if hostname can not be parsed
    :return str | None: app_name
    """
    hostname_components = socket.gethostname().split('-')
    if len(hostname_components) > 1:
        return hostname_components[1]


def get_hn_var(info_file, *field_names, default_val=None):
    """
    Retrieve a field from the given info file. Return the default value if the field is not found if any,
    otherwise crash
    :param str info_file: JSON file that contains the fieldnames
    :param [str, ..] field_names: Field names that are to be retrieved from the info file
    :param obj|None default_val: Default value to return if the field is not found. If none and the field
    is not found, the error is raised
    :rtype: obj
    """
    try:
        # try to get the app name from the common.json file, this is the most accurate
        with open(info_file, 'r') as info:
            d = json.load(info)
            # Go as deep as the number of field names given
            for field_name in field_names:
                d = d[field_name]
            return d
    except (IOError, ValueError, KeyError) as e:
        # when the common.json file is not available (app user can not access it)
        # or the fieldname is not found in the json file
        if default_val is not None:
            return default_val
        raise e


def get_hn_app_var(*field_names, default_val=None):
    """
    Retrieve a field from the app.json file. Return the default value if the field is not found if any,
    otherwise crash
    :param [str, ..] field_names: Field names that are to be retrieved from the common file
    :param obj|None default_val: Default value to return if the field is not found. If none and the field
    is not found, the error is raised
    :rtype: obj
    """
    return get_hn_var(APP_INFO_PATH, default_val=default_val, *field_names)


def get_hn_common_var(*field_names, default_val=None):
    """
    Retrieve a field from the common.json file. Return the default value if the field is not found if any,
    otherwise crash.
    NOTE: This requires root privileges!
    :param [str, ..] field_names: Field names that are to be retrieved from the common file
    :param obj|None default_val: Default value to return if the field is not found. If none and the field
    is not found, the error is raised
    :rtype: obj
    """
    return get_hn_var(COMMON_INFO_PATH, default_val=default_val, *field_names)


def get_hn_magweb_var(*field_names, default_val=None):
    """
    Retrieve a field from the magweb.json file. Return the default value if the field is not found if any,
    otherwise crash.
    NOTE: This requires root privileges!
    :param [str, ..] field_names: Field names that are to be retrieved from the magweb file
    :param obj|None default_val: Default value to return if the field is not found. If none and the field
    is not found, the error is raised
    :rtype: obj
    """
    return get_hn_var(MAGWEB_INFO_PATH, default_val=default_val, *field_names)


def get_app_name():
    """
    Get the app name from the node on which this script is running
    :str return: app_name
    """
    return get_hn_common_var('tag', default_val=parse_app_name_from_host_name())


def host_is_a_vagrant():
    """
    Returns True if the host on which the importer is running
    is a hypernode-vagrant. That's the customer development
    environment people can download here:
    https://github.com/ByteInternet/hypernode-vagrant
    :return bool is_a_vagrant: True if it is a hypernode-vagrant
    instance running this code, False if not
    """
    return isfile('/etc/hypernode/is_vagrant')


def host_is_a_docker():
    """
    Returns True if the host on which the importer is running
    is a hypernode-docker. That's the customer development
    environment people can download here:
    https://github.com/ByteInternet/hypernode-docker
    :return bool is_a_docker: True if it is a hypernode-docker
    instance running this code, False if not
    """
    return isfile('/etc/hypernode/is_docker')


def get_hypernode_name():
    """
    Get the Hypernode hostname, e.g. dummy.hypernode.io

    Assumes the hostname is like "rar1s5-dummy-magweb-xls.nodes.hypernode.io"

    :return: the Hypernode hostname, e.g. dummy.hypernode.io
    :return str: hypernode name like: vdloo.hypernode.io
    """
    app_name = 'hypernode' if host_is_a_docker() else get_app_name()
    if not app_name:
        raise ValueError("Could not determine full Hypernode name based on current hostname {}".format(
            socket.gethostname())
        )
    return "{}.hypernode.{}".format(
        app_name, 'local' if host_is_a_vagrant() or host_is_a_docker() else 'io'
    )


def get_public_ip():
    """
    Retrieve the public IP
    :return str public_ip: The public IP
    """
    # Randomized order just in case a lot of nodes reboot at the exact same time
    # for some reason and all need to determine if routing through the outbound
    # floating IP works as intended. Shuffling will spread the load, decrease
    # service failure impact (time before trying the next endpoint) but it will
    # also make it non-deterministic. For now this should be OK though.
    randomized_public_ip_urls = PUBLIC_IP_URLS.copy()
    shuffle(randomized_public_ip_urls)
    for public_ip_url in randomized_public_ip_urls:
        with suppress(Exception):
            response = requests.get(public_ip_url, timeout=GET_IP_TIMEOUT)
            if response.status_code == HTTPStatus.OK and response.text.strip():
                return response.text.strip()
    raise RuntimeError(
        "Failed to retrieve public IP! Perhaps something is wrong with the "
        "networking on this node or all the public IP endpoints are down?"
    )


def is_varnish_enabled():
    """
    Check if varnish is enabled on the system

    :return bool enabled: True if enabled, False if not
    """
    varnish_settings_lines = []
    with suppress(FileNotFoundError, IOError):
        with open('/etc/default/varnish', 'rt') as fh:
            varnish_settings_lines = fh.read().split('\n')
    return 'START=yes' in varnish_settings_lines


def is_elasticsearch_enabled():
    """
    Check if elasticsearch is enabled on the system

    :return bool enabled: True if enabled, False if not
    """
    return isfile('/etc/systemd/system/multi-user.target.wants/elasticsearch.service')


def is_rabbitmq_server_enabled():
    """
    Check if rabbitmq-server is enabled on the system

    :return bool enabled: True if enabled, False if not
    """
    return isfile('/etc/systemd/system/multi-user.target.wants/rabbitmq-server.service')


def is_persistent_redis_enabled():
    """
    Check if persistent Redis is enabled on the system

    :return bool enabled: True if enabled, False if not
    """
    return isfile('/etc/systemd/system/redis-server-persistent.service')


def is_supervisor_enabled():
    """
    Check if supervisor is enabled on the system

    :return bool enabled: True if enabled, False if not
    """
    return isfile('/etc/systemd/system/multi-user.target.wants/supervisor.service')


def get_active_php_version_info():
    """
    Return the active PHP version as a tuple. This is guranteed to be in
    (major, minor, release) format where each value is a string.
    If failed to detect such information a RuntimeError is raised.

    :return tuple php_version: example ('7', '0', '28')
    :raises: RuntimeError
    """
    try:
        php_version = check_output(['php', '-r', 'echo PHP_VERSION;'], stderr=open(devnull, 'wb')).decode().strip()
    except (CalledProcessError, OSError, ValueError) as err:
        raise RuntimeError('Failed to detect PHP version: {}'.format(err))

    matched = re.search(r'^(\d+)\.(\d+)\.(\d+)', php_version)
    if not matched:
        raise RuntimeError('Failed to detect a valid PHP version from "{}"'.format(php_version))

    return matched.groups()


def is_magento_1():
    """
    Check if the Hypernode is currently hosting Magento 1 or 2
    :return bool: True if Magento 1, False if not
    """
    # We can't just check /data/web/public/app/Mage.php, sometimes people have
    # unconventional constructions where their Magento installation is hidden in a
    # deeper directory. This find checks up to 3 levels deep (so if it is
    # /data/web/public/shop/app/etc/Mage.php then that is also OK). We don't check
    # any deeper because then it will start hitting things like the media directory
    # and its subdirectories which can potentially house a lot of files and this
    # script is ran in a cron, so that could potentially cause a lot of unwanted IOPS.
    command = "find /data/web/public/ -maxdepth 3 -type f | grep 'app/Mage.php'"
    return_code = subprocess.call(command, shell=True, stdout=subprocess.PIPE)

    return return_code == 0


def get_magento_2_install_dir(path):
    """
    Searches in `path` for 'static.php' files excluding any 'static.php' files found for vendors. All static.php
    files are then resolved (i.e. symlinks) to determine the real path. If only one path is found we return it since
    we know for sure this will be the magento install dir. If more paths are found it probably means there are multiple
    magento installations in the provided `path` and we can't return any logical value since we don't know what the
    user is looking for. In those cases we return None.
    :param path: The path to search for a magento installation.
    :return: None if no/multiple install dirs found|path to magento 2 install dir if one path found
    """
    static_php_paths = set()
    # Walk with a maximum depth of 3, otherwise we will start accessing media directories
    # which are always stuffed with files which will cause high IO load which might crash
    # some nodes.
    for path, subdirs, files in walk_with_depth(path, depth=3):
        for name in files:
            if name == 'static.php' and 'vendor' not in path:
                static_php_paths.add(os.path.realpath(os.path.join(path, name)))

    if len(static_php_paths) == 1:
        result = static_php_paths.pop()
        return result.replace('static.php', '')
