"""
kamikaze3.common.postfix
========================
functionality related to postfix, processing logs, etc.
"""
import re
from os import path
from datetime import datetime
from logging import getLogger
from typing import List
from kamikaze3.common.syslog import log_file_lines

LOG_FILE = "/var/log/mail.log"

logger = getLogger(__name__)


def is_mail_failed_log(log: str) -> bool:
    """Return True if the log line is for a deferred/bounced mail message"""
    log = log.upper()
    return ("STATUS=DEFERRED" in log) or ("STATUS=BOUNCE" in log)


def get_transip_auth_issue(log: str) -> str:
    """If the log relates to TransIP mail authentication failure, return the details of the issue parsed out from
    the log line, otherwise return empty string
    """

    # sample log line:
    # Aug 17 11:28:11 5vi9lw-erikhyperdev-magweb-cmbl postfix/smtp[5532]: 5C5DC40E4F: to=<test@example.org>, \
    # relay=vps.transip.email[86.105.244.7]:587, delay=4.3, delays=0.02/0.01/1.1/3.1, dsn=4.7.0, status=deferred \
    # (host vps.transip.email[86.105.244.7] said: 421 4.7.0 No valid 'x-transip-mail-auth' DNS record found for \
    # 'erikhyperdev.hypernode.io'. Please see https://www.transip.eu/knowledgebase/entry/349 for more information. \
    # (4BVWxl3c21z7t9L) (in reply to end of DATA command))
    match = re.search(r"(No valid 'x-transip-mail-auth' DNS record found for.+)\s+\(.+\)\s+\(in reply to", log,
                      re.IGNORECASE)
    return match.group(1).strip() if match else ""


def get_transip_spf_issue(log: str) -> str:
    """If the log relates to TransIP mail SPF check failure, return the details of the issue parsed out from
    the log line, otherwise return empty string
    """

    # we know of 2 types of errors:
    # sample log line 1:
    # Apr 22 13:44:30 wifamk-testfarzad-magweb-cmbl postfix/smtp[5628]: 013A860BF0: to=<test@example.org>, relay=vps.transip.email[86.105.244.7]:587, \
    # delay=0.51, delays=0.03/0.03/0.18/0.26, dsn=4.7.0, status=deferred (host vps.transip.email[86.105.244.7] said: 421 4.7.0 SPF records for testfarzad.hypernode.io \
    # have not been configured correctly. Please see https://www.transip.eu/knowledgebase/entry/350 for more information. \
    # (4FQzDZ2Csbz18H4J) (in reply to end of DATA command)
    #
    # sample log line 2:
    # Apr 23 13:01:09 cj8jyi-vdloo-magweb-aws postfix/smtp[16986]: 9C20C20DF6: to=<test@example.org>, relay=vps.transip.email[86.105.244.7]:587, \
    # delay=0.48, delays=0/0/0.25/0.22, dsn=4.7.0, status=deferred (host vps.transip.email[86.105.244.7] said: 421 4.7.0 An Internal Server Error has occurred. \
    # SPF records for testfarzad.hypernode.io could not be verified. Please try again later. Please see https://www.transip.eu/question/110000767 for more information. \
    # (4FRZD471jBz2Bd3L) (in reply to end of DATA command)
    match = re.search(r"(SPF records for.+)\s+\(.+\)\s+\(in reply to", log,
                      re.IGNORECASE)
    return match.group(1).strip() if match else ""


def get_transip_dkim_issue(log: str) -> str:
    """If the log relates to TransIP mail DKIM verification failure, return the details of the issue parsed out from
    the log line, otherwise return empty string
    """

    # sample log line:
    # May 17 08:08:16 wifamk-testfarzad-magweb-cmbl postfix/smtp[5821]: 0DBAD60AEB: to=<test@example.org>, \
    # relay=vps.transip.email[86.105.244.7]:587, delay=0.36, delays=0.03/0.01/0.18/0.14, dsn=5.7.1, status=bounced \
    # (host vps.transip.email[86.105.244.7] said: 550 5.7.1 \
    # message failed DKIM verification with a permanent error (4FkBb41y7Gz1BG2s) (in reply to end of DATA command))
    match = re.search(r"(message failed DKIM verification with a permanent err.+)\s+\(.+\)\s+\(in reply to", log,
                      re.IGNORECASE)
    return match.group(1).strip() if match else ""


def find_failed_transip_mails_from_logs(after: datetime = None, limit: int = 0) -> List[str]:
    """Find the first occurrence of an authentication failure to TransIP mail in postfix logs.
    Returns the description of the failure as logged by postfix.
    Return empty string if no such log was found.
    Search faster by specifying a datetime object for "after".
    """
    mail_issues = []  # type: List[str]
    if not path.exists(LOG_FILE):
        logger.info("no mail log file exists on '{}'".format(LOG_FILE))
        return mail_issues

    for line, line_num in log_file_lines(LOG_FILE, filter_=is_mail_failed_log, after=after):
        mail_issue = get_transip_auth_issue(line)
        if mail_issue:
            logger.info("found mail transip auth failure log in line {} of {}".format(line_num, LOG_FILE))
            mail_issues.append(mail_issue)
        else:
            mail_issue = get_transip_spf_issue(line)
            if mail_issue:
                logger.info("found mail transip SPF failure log in line {} of {}".format(line_num, LOG_FILE))
                mail_issues.append(mail_issue)
            else:
                mail_issue = get_transip_dkim_issue(line)
                if mail_issue:
                    logger.info("found mail transip DKIM failure log in line {} of {}".format(line_num, LOG_FILE))
                    mail_issues.append(mail_issue)

        if limit and len(mail_issues) >= limit:
            logger.warning("reached the limit on transip mail failures of {} in file {}. skipping the rest".format(limit, LOG_FILE))
            break

    if len(mail_issues) < 1:
        logger.info("didn't find any mail issue logs in {}".format(LOG_FILE))

    return mail_issues
