from contextlib import contextmanager
from sys import stdout
from subprocess import Popen, PIPE
from threading import Timer

DEFAULT_SHELL_TIMEOUT = 1200  # 20 minutes


def terminate_and_kill(process, timeout, command):
    """
    Terminate and kill a subprocess.Popen process
    :param obj process: The object to terminate and kill
    :param int timeout: The amount of time to allow
    it to run.
    :param list | str command: The command as a list or as string (when shell).
    I.e. ['/bin/ls', '/root'] or "/bin/ls /root"
    :return None:
    """
    process.terminate()
    process.kill()
    raise TimeoutError(
        "Subprocess timed out after {} seconds. "
        "Command was: {}".format(timeout, command)
    )


@contextmanager
def terminate_on_timeout(process, timeout, command):
    """
    Run a thread to watch the passed process. If it is running longer than the allowed amount of
    time, send a terminate and kill to the process.
    :param obj process: the running process to kill after the timeout expires.
    :param int timeout: The amount of time to allow it to run.
    :param list | str command: The command as a list or as string (when shell).
    I.e. ['/bin/ls', '/root'] or "/bin/ls /root"
    :yield obj timer: The timer object
    """
    timer = Timer(
        timeout, terminate_and_kill,
        args=[process, timeout, command]
    )
    timer.start()
    yield
    timer.cancel()


def write_real_time_output(process):
    """
    Write real time output from the running process
    :param obj process: the running process to do
    blocking reads on the stdout pipe on.
    :return None:
    """
    for line in iter(process.stdout.readline, b''):
        stdout.buffer.write(line)
        stdout.flush()


def run_local_command(command, shell=False, environment=None, timeout=DEFAULT_SHELL_TIMEOUT):
    """
    Run a command on the local system.
    :param iter command: an iterable that composes the command or a string if shell=True
    :param bool shell: True if a shell should be used, False is not
    :param dict environment: The environment to pass to the command
    :param int timeout: The amount of time to allow it to run.
    :return int return_code: The command return code
    """
    process = Popen(
        command, stdout=PIPE, shell=shell,
        bufsize=0, env=environment
    )
    with terminate_on_timeout(process, timeout, command):
        write_real_time_output(process=process)
    process.wait()
    return process.returncode


def replace_in_file(filename, original, replace):
    """
    Replace a string in a file
    :param str filename: The file to replace the string in
    :param bstr original: The string to replace
    :param bstr replace: The string to replace it with
    :return None:
    """
    with open(filename, 'r+b') as file_to_replace_in:
        contents = file_to_replace_in.read()
        contents = contents.replace(original, replace)
        file_to_replace_in.seek(0)
        file_to_replace_in.truncate()
        file_to_replace_in.write(contents)
