"""
kamikaze3.common.system
~~~~~~~~~~~~~~~~~~~~~~~
Common system/os related functionality
"""
import sys
import os
from grp import getgrnam


def strip_unsafe_env_vars(env=None):
    """Filter out unsafe environment varialbes from the environment vars dict
    (provided param, or current process environment variables).
    Environment variables that may affect runtime behavior are considered unsafe,
    for example PYTHON* or LD* variables.

    :param dict env: environment variables, or None to use process env (default=None)
    :return: dict
    """
    if env is None:
        env = os.environ
    return {k: v for k, v in env.items() if not k.startswith('PYTHON') and not k.startswith('LD')}


def setgid_exec_python(group_name, entrypoint, cmd_name=None, cmd_args=None):
    """
    Run the specified entrypoint Python file by calling a setgid syscall
    to the specified group name.

    Note: this function uses os.exec* semantics, so won't return and
    current process would be replaced, and open file objects and descriptors
    are not flushed.

    :param str group_name: name of the group to setgid to
    :param str entrypoint: path to Python file to run
    :param str command_name: new process name, replacing this one (default is sys.arv[0])
    :param list args: args passed to sub command (default is sys.argv[1:])
    """
    # for security reasons we want to make sure this function
    # only executes the correct python interpreter running the specified script.
    # using execve() instead of execvp() ensures user can not force
    # to point to another Python (locally installed), and won't
    # be able to pass any extra Python flags (like -c "foo"),
    # and we can strip environment variables that may effect Python

    if not cmd_name:
        cmd_name = sys.argv[0]

    if cmd_args is None:
        cmd_args = sys.argv[1:]

    args = [entrypoint] + cmd_args

    # first item in args should be the name for the new command
    args.insert(0, cmd_name)

    # prevent passing unsafe environment variables (PYTHON*, ec.t) to be passed
    # to avoid replacing PYTHONPATH or other vars that might affect Python
    # behavior
    env = strip_unsafe_env_vars()

    target_gid = getgrnam(group_name).gr_gid
    os.setregid(target_gid, target_gid)
    os.execve('/usr/bin/python3', args, env)


def walk_with_depth(path, depth=1):
    """
    Walk function that performs the same as `os.walk` with the added ability of specifying a depth. Specifying a depth
    of 1 means we will walk the directory of `path` and all subdirectories, but no further. Same can be applied for
    depth 2, accept we go one more subdirectory deeper etc.
    :param path: The path to walk
    :param depth: Depth to search for. Depth of 1 means that we will only go one directory deep for searches.
    :return: regular `os.walk` return values
    """
    # Based off of https://stackoverflow.com/a/234329/1294864
    # We need to strip any trailing `/` here, otherwise our root separator count won't be accurate.
    # I.e. if we receive `/my/path/` instead of `/my/path` it means we will check for a separator count of
    # 4 (root_separator_count + depth). This means we would accidentally search one directory too deep, because
    # os.walk will return current_path values with trailing `/` stripped. So the separator count for
    # `current_path` would be 2, and our maximum count would be 4. Meaning we will search at a depth of 2
    # instead of 1.
    path = path.rstrip(os.path.sep)
    root_separator_count = path.count(os.path.sep)
    for current_path, dirs, files in os.walk(path):
        yield current_path, dirs, files
        current_separator_count = current_path.count(os.path.sep)
        if root_separator_count + depth <= current_separator_count:
            del dirs[:]


def user_is_root():
    """
    Returns true if the users if root, returns false if the user is not root
    :return: boolean
    """
    return os.getuid() == 0
