from logging import getLogger
from subprocess import check_output, CalledProcessError, DEVNULL
from ipaddress import IPv4Network

from kamikaze3.utils import request_confirmation


# We allow a subset of https://gist.github.com/mcastelino/c38e71eb0809d1427a6650d843c42ac2
VALID_IPTABLES_JUMP_TARGETS = ('RETURN', 'ACCEPT', 'DROP', 'REJECT')


class FirewallException(RuntimeError):
    pass


class Firewall:
    def __init__(self, wait_time=3, safe_mode=False):
        """
        :param wait_time: int/str: Max lock acquire time
        :param safe_mode: bool: Ask user confirmation before doing anything dangerous
        """
        self.logger = getLogger('kamikaze3.firewall.iptables')
        self.wait_time = wait_time
        self.blacklist_chain = 'hypernode-blacklist'
        self.whitelist_chain = 'hypernode-whitelist'
        self.safe_mode = safe_mode

    @staticmethod
    def _remove_leading_iptables_placement_arguments(rule):
        """
        Remove any leading placement arguments
        :param rule: str: The rule to check for leading placement arguments
        :return: str: The stripped rule
        """
        options = ('-A', '-P', '-I', '-N')
        for option in options:
            if rule.startswith(option):
                rule = rule.replace(option, '', 1)
                break
        return rule

    def construct_iptables_rule(self, ip, cidr=32, chain='INPUT', action='ACCEPT'):
        """
        Make an IPtables rule
        :param ip: str: The IP address (1.1.1.1)
        :param cidr: str/int: The CIDR (32) (without the slash)
        :param chain: str: The IPtables chain to use
        :param action: str: The IPtables action to take
        :return: str: The constructed rule, or None if validation failed
        """
        network = '{}/{}'.format(ip, cidr)
        if not self.validate_network(network) or action not in VALID_IPTABLES_JUMP_TARGETS:
            return None
        # We use shorthand arguments here, because they need to match the information given by IPtables -S
        return '{chain} -s {source} -j {jump}'.format(
            chain=chain,
            source=network,
            jump=action
        )

    def prepare_rule(self, rule, action='add'):
        """
        Wrapper class to perform preparation for rule execution
        :param rule: str: The rule to prepare
        :param action: str: The action to take on the rule (just for user message)
        :return str: The prepared rule
        """
        rule = self._remove_leading_iptables_placement_arguments(rule)
        if self.safe_mode:
            request_confirmation('Do you want to {} the following rule: {}?'.format(action, rule))
        return rule

    def validate_network(self, network):
        """
        Validate an IP network for soundness (can also validate a single IP)
        :param network: str: The network to validate
        :return: bool: True if OK, False otherwise
        """
        try:
            # Hypernode only uses v4 atm, so explicitly say we want a v4 network
            network_obj = IPv4Network(network)
            if not network_obj.is_global or network_obj.is_multicast:
                raise ValueError('Given network is not public')
        except ValueError as e:
            self.logger.error('{} is not a valid IPv4 network, got error: {}'.format(network, e))
            return False
        # Check for too small CIDR
        if int(network.split('/')[1]) < 16:
            self.logger.warning('CIDR lower than 16 detected, failing network validation to prevent blocking too much')
            return False
        # Finally return everything OK -> True
        return True

    def iptables_command(self, command, silent=False, numeric=False):
        """
        Run an iptables command
        :param command: str: The command to run
        :param silent: bool: Be quiet if iptables returns an error
        :param numeric: bool: use iptables --numeric (-n)
        :return: str: The output from the iptables command
        """
        command = '/sbin/iptables --wait {wait} {numeric} '.format(
            wait=self.wait_time,
            numeric='--numeric' if numeric else ''
        ) + command
        self.logger.debug('Running IPtables command: {}'.format(command))
        try:
            output = check_output(command.split(), stderr=DEVNULL if silent else None).decode('utf-8').strip()
        except CalledProcessError as e:
            if not silent:
                self.logger.error('IPtables command error; {}'.format(e))
            return None
        return output

    def ensure_firewall_chains(self):
        """
        Make sure that the iptables environment on the host matches the setup we use for putting in entries
        :return: bool: True if everything OK, False if not OK
        """
        # Blacklist
        self.logger.debug('Ensuring blacklist mangle table exists')
        if not self.iptables_command(
                '--table mangle --list {}'.format(self.blacklist_chain),
                silent=True,
                numeric=True
        ):
            self.logger.debug('IPtables {} mangle chain does not exist, creating...'.format(self.blacklist_chain))
            if self.iptables_command('--table mangle --new-chain {}'.format(self.blacklist_chain)) is not None:
                self.logger.debug('Successfully created the {} mangle chain'.format(self.blacklist_chain))
            else:
                self.logger.error('Failed to create {} mangle chain, throwing error'.format(self.blacklist_chain))
                return False

        # Whitelist
        self.logger.debug('Ensuring whitelist filter table exists')
        # Filter is the default table
        if not self.iptables_command('--list {}'.format(self.whitelist_chain), silent=True, numeric=True):
            self.logger.debug('IPtables {} filter table does not exist, creating...'.format(self.whitelist_chain))
            if self.iptables_command('--new-chain {}'.format(self.whitelist_chain)) is not None:
                self.logger.debug('Successfully created the {} filter table'.format(self.whitelist_chain))
            else:
                self.logger.error('Failed to create {} filter table, throwing error'.format(self.whitelist_chain))
                return False
        # Finally return everything OK -> True
        return True

    def ensure_firewall_chain_references(self):
        """
        Make sure that the iptables firewall chains we use are referenced from the main chains
        :return: bool: True if everything OK, False otherwise
        """
        # Blacklist
        self.logger.debug('Ensuring {} chain is referenced from mangle table PREROUTING chain'.format(
            self.blacklist_chain)
        )
        if self.iptables_command(
                '--table mangle --check PREROUTING --jump {}'.format(self.blacklist_chain), silent=True
        ) is None:
            self.logger.debug('Blacklist chain is not referenced from mangle table PREROUTING chain, referencing it...')
            if self.iptables_command(
                    '--table mangle --insert PREROUTING --jump {}'.format(self.blacklist_chain)
            ) is not None:
                self.logger.debug('Successfully referenced blacklist chain in mangle PREROUTING chain')
            else:
                self.logger.error('Failed to reference blacklist chain in mangle table, throwing error')
                return False

        # Whitelist
        self.logger.debug('Ensuring {} chain is referenced from the INPUT chain'.format(self.whitelist_chain))
        if self.iptables_command('--check INPUT --jump {}'.format(self.whitelist_chain), silent=True) is None:
            self.logger.debug('Whitelist chain is not referenced from the INPUT chain, referencing it...')
            if self.iptables_command(
                    '--insert INPUT --jump {}'.format(self.whitelist_chain)
            ) is not None:
                self.logger.debug('Successfully referenced whitelist chain in INPUT chain')
            else:
                self.logger.error('Failed to reference whitelist chain in INPUT chain, throwing error')
                return False
        # Finally return everything OK -> True
        return True

    def get_chain_entries(self, chain, table='filter', list_rules=True, remove_chain_name=True,
                          remove_positioning_argument=False):
        """
        Get all iptables entries for a specific chain
        :param chain: str: The chain to get the IPtables rules for
        :param table: str: The table in which the chain is located
        :param list_rules: bool: use IPtables --list-rules (-S) (default) instead of --list (-L)
        :param remove_chain_name: bool: If true: remove the policy rule from the return set
        :param remove_positioning_argument: bool: If true tries to remove the leading positioning argument from the rule
        :return: tuple: The IPtable rules of that chain
        """
        self.logger.debug('Getting IPtables rules for chain {}'.format(chain))
        chain_rules = self.iptables_command(
            '--table {table} {list_command} {chain}'.format(
                table=table,
                list_command='--list-rules' if list_rules else '--list',
                chain=chain
            ),
            numeric=False if list_rules else True
        )
        if chain_rules:
            chain_rules_list = chain_rules.split('\n')
            if remove_chain_name or remove_positioning_argument:
                for idx, chain_rule in enumerate(chain_rules_list):
                    if remove_chain_name:
                        if list_rules:
                            if chain_rule.startswith('-N') or chain_rule.startswith('-P'):
                                del chain_rules_list[idx]
                                continue
                        else:
                            if 'Chain' in chain_rule and '(' in chain_rule:
                                del chain_rules_list[idx]
                                continue
                    if remove_positioning_argument:
                        chain_rule = self._remove_leading_iptables_placement_arguments(chain_rule)
                        chain_rules_list[idx] = chain_rule

            # We return a set because this is a IPtables 'fact' of the current state, so no mutation allowed
            return tuple(chain_rules_list)
        else:
            self.logger.warning('No rules were found (does the chain exist?)')
            return None

    def delete_iptables_rule(self, rule, table='filter'):
        """
        Delete an IPtables rule
        :param rule: str: The raw rule to delete, example(-A INPUT -i lxcbr0 -p udp -m udp --dport 53 -j ACCEPT)
        :param table: str: The table to work on
        :return: True if successful, False if failure
        """
        rule = self.prepare_rule(rule, action='delete')
        if self.iptables_command('--table {table} --delete {rule}'.format(table=table, rule=rule)) is not None:
            self.logger.debug('Successfully deleted IPtables rule: {}'.format(rule))
            return True
        else:
            self.logger.error('Unable to delete IPtables rule: {} (did it even exist?)'.format(rule))
            return False

    def add_iptables_rule(self, rule, table='filter'):
        """
        Add an IPtables rule
        :param rule: str: The raw rule to add, example(-A INPUT -i lxcbr0 -p udp -m udp --dport 53 -j ACCEPT)
        :param table: str: The table to work on
        :return: True if successful, False if failure
        """
        rule = self.prepare_rule(rule, action='add')
        if self.iptables_command('--table {table} --append {rule}'.format(table=table, rule=rule)) is not None:
            self.logger.debug('Successfully added IPtables rule: {}'.format(rule))
            return True
        else:
            self.logger.error('Unable to add IPtables rule: {} (is it correctly formatted?)'.format(rule))
            return False

    def flush_iptables_chain(self, chain, table='filter'):
        """
        Empty an IPtables chain
        :param chain: str: The chain to empty
        :param table: str: The table were the chain is located
        :return: bool: True if successful, False if not
        """
        if self.safe_mode:
            request_confirmation('Do you want to flush the following chain {} in table {}?'.format(chain, table))
        self.logger.debug('Flushing IPtables {} chain'.format(chain))
        if self.iptables_command('--table {table} --flush {chain}'.format(table=table, chain=chain)) is not None:
            self.logger.debug('Successfully flushed the {} chain'.format(chain))
            return True
        else:
            self.logger.error('Unable to flush the {} chain (maybe there was no chain?)'.format(chain))
            return False
