import logging
import os
import requests
from argparse import ArgumentParser
from socket import getfqdn
from subprocess import check_output, CalledProcessError

from kamikaze3.firewall.iptables import Firewall, FirewallException
from kamikaze3.firewall.log import configure_logging
from kamikaze3.common.env import host_is_a_docker, host_is_a_vagrant
from kamikaze3.common.api import get_hypernode_api_token
from kamikaze3.settings import USR_CNF

FLAG = '-f'


logger = logging.getLogger('kamikaze3.firewall')


def _get_firewall_url():
    try:
        format_url = check_output((USR_CNF, FLAG)).decode('utf-8').strip()
    except (FileNotFoundError, CalledProcessError):
        raise RuntimeError('Unable to retrieve hn-api URL from USR_CNF')
    return format_url


def parse_args(args=None):
    parser = ArgumentParser(description='Script to pull the centralized Hypernode firewall '
                                        'to this node and apply via iptables')
    parser.add_argument('-v', '--debug', action='store_true', help='Enable debug logging')
    parser.add_argument('-d', '--dry-run', action='store_true', help='Write rules to STDOUT instead '
                                                                     'of applying to iptables')
    parser.add_argument('--block-only', action='store_true', help='Only process block actions')
    parser.add_argument('--allow-only', action='store_true', help='Only process allow actions')
    parser.add_argument('--flush-blacklist', action='store_true',
                        help='Flush the blacklist table and exit (for debugging)')
    parser.add_argument('--flush-whitelist', action='store_true',
                        help='Flush the whitelist table and exit (for debugging)')
    parser.add_argument('--safe-mode', action='store_true', help='Enable firewall safe-mode '
                                                                 '(ask before making dangerous changes)')

    return parser.parse_args(args=args)


def get_firewall_entries_from_hn_api(url):
    """
    Tries to get a JSON object from hn-api <URL>
    :param url: str: The URL to get
    :return: json-object / object: If json then json object, otherwise the thing that was returned by hn-api
    """
    response = requests.get(url, headers={
        'Authorization': 'Token {}'.format(get_hypernode_api_token()),
        'Content-Type': 'application/json',
        'Accept': 'application/json',
        'User-Agent': '{}:{}'.format(
            getfqdn(), 'hypernode-kamikaze3'
        )
    })
    response.raise_for_status()
    if not response:
        raise FirewallException('hn-api URL {} return no response!'.format(url))
    return response.json()


def add_chain_rules(requested, firewall, table='filter'):
    """
    Process firewall changes
    First remove old entries (not found in requested, but present in current)
    Then add new entries (not found in current, but present in requested)
    :param requested: tuple: The request rules (or requested state)
    :param table: str: The IPtables table to work on
    :param firewall: iptables.Firewall object
    """
    # This method assumes that the firewall is already correctly configured and that the rules are correct!
    # Add new entries
    for rule in requested:
        logger.debug('Adding rule: {} to the {} table'.format(rule, table))
        if not firewall.add_iptables_rule(rule, table):
            logger.warning('Adding firewall rule failed')


def construct_requested_rules_tuple(entries, firewall):
    """
    Make a set of requested rules based on the info retrieved from hn-api
    :param entries: list of dicts: The requested state for this action
    :param firewall: iptables.Firewall object
    :return: set: The requested state for this action
    """
    requested_state = []
    for entry in entries:
        if entry['action'] == 'deny':
            rule = firewall.construct_iptables_rule(
                entry['ip'],
                cidr=entry['cidr'],
                chain=firewall.blacklist_chain,
                action='DROP'
            )
        else:
            rule = firewall.construct_iptables_rule(
                entry['ip'],
                cidr=entry['cidr'],
                chain=firewall.whitelist_chain,
                action='ACCEPT'
            )
        if rule is None:
            logger.warning(
                'Entry {} received from hn-api does not seem to be valid, skipping...'.format(entry['ip'])
            )
            continue
        requested_state.append(rule)
    return tuple(requested_state)


def flush_chain(chain, firewall, table='filter'):
    """
    Flush an IPtables chain
    :param chain: str: The IPtables chain to flush
    :param table: str: The IPtables table to operate on
    :param firewall: iptables.Firewall object
    :return bool: True if successful, False otherwise
    """
    logger.debug('Flushing the {} chain in table {}'.format(chain, table))
    if not firewall.flush_iptables_chain(chain, table=table):
        logger.error('Flushing the chain {} failed'.format(chain))
        return False
    return True


def process_chain_entries(rule_type, firewall, dry_run=False):
    """
    Retrieve <type> entries from hn-api
    Flush the current <type> chain
    Fill the current <type> chain with entries received from hn-api
    :param rule_type: str: Either allow or deny
    :param firewall: iptables.Firewall object
    :param dry_run: bool: don't change anything
    :return: The entries received from hn-api, None if none
    """
    if rule_type not in ('allow', 'deny'):
        raise FirewallException('Type {} not allowed'.format(rule_type))
    chain = firewall.blacklist_chain if rule_type == 'deny' else firewall.whitelist_chain
    table = 'mangle' if rule_type == 'deny' else 'filter'
    logger.debug('Retrieving {} entries from hn-api'.format(rule_type))
    entries = get_firewall_entries_from_hn_api('{}/?action={}'.format(_get_firewall_url(), rule_type))
    if not entries:
        logger.warning('No entries received from hn-api')
        # Return early, because there might be something wrong with hn-api at this point so let's keep the old entries
        return
    if not dry_run:
        # Make the rules
        rules = construct_requested_rules_tuple(entries, firewall)
        logger.debug('Adding the {} rules received from hn-api'.format(rule_type))
        # Empty the current chain
        if not flush_chain(chain, firewall, table=table):
            raise FirewallException('Flushing chain {} failed'.format(chain))
        add_chain_rules(rules, firewall, table=table)
        logger.debug('{} entries processed'.format(rule_type))
    # Finally return the entries for later use
    return entries


def main(args=None):
    args = parse_args(args)
    configure_logging(debug=(args.debug or args.dry_run))
    logger.debug('Logger configured')

    if args.dry_run:
        logger.debug('Dry run flag set, not applying any IPtables')

    # ROOT only!
    if os.getuid() != 0:
        raise FirewallException('This script can only be run by root!')

    # Non of these virtual imposter's
    if host_is_a_docker() or host_is_a_vagrant():
        raise FirewallException('This command can only be executed on non virtual environments')

    # Setup the firewall and ensure everything is as we expect it to be
    logger.debug('Ensuring IPtables firewall is in the correct state')
    firewall = Firewall(safe_mode=args.safe_mode)
    # Don't configure the firewall on a dry_run
    if not args.dry_run and (not firewall.ensure_firewall_chains() or not firewall.ensure_firewall_chain_references()):
        raise FirewallException('Firewall not configured correctly and unable to rectify, please investigate')

    # FLUSH?
    if args.flush_blacklist or args.flush_whitelist:
        if args.flush_blacklist:
            if not flush_chain(firewall.blacklist_chain, firewall, table='mangle'):
                exit(1)
        if args.flush_whitelist:
            if not flush_chain(firewall.whitelist_chain, firewall):
                exit(1)
        exit(0)

    # ALLOWs
    if not args.block_only:
        logger.debug('Processing allow entries')
        allow_entries = process_chain_entries('allow', firewall, args.dry_run)
    else:
        logger.debug('Skipping ALLOWs because block_only given')

    # DENYs
    if not args.allow_only:
        logger.debug('Processing deny entries')
        deny_entries = process_chain_entries('deny', firewall, args.dry_run)
    else:
        logger.debug('Skipping DENYs because allow_only given')

    # Dump entries if dry_run
    if args.dry_run:
        logger.debug('Dry-running, printing the info')
        current_deny_entries = firewall.get_chain_entries(firewall.blacklist_chain, table='mangle')
        current_allow_entries = firewall.get_chain_entries(firewall.whitelist_chain)
        logger.debug('Current IPtables entries')
        print({
            'current_allow_entries': current_allow_entries,
            'current_deny_entries': current_deny_entries,
        })
        if not args.block_only:
            logger.debug('Allow info received from hn-api')
            print(allow_entries)
        if not args.allow_only:
            logger.debug('Deny info received from hn-api')
            print(deny_entries)
