import argparse
import getpass
import json
import logging
from operator import itemgetter
import os
import subprocess
import sys
import grp
import pwd

FTPASSWD_LOCATION = "/data/var/ftp/ftpasswd"
FTPASSWD_BINARY = "/usr/sbin/ftpasswd"


logging.basicConfig(level=logging.DEBUG)


def main(argv=None):
    argv = argv or sys.argv[1:]
    args = parse_args(argv)

    if args.subcommand == "remove":
        remove_user(args.username)
    elif args.subcommand == "add":
        password = args.password or get_password()
        add_user(args.username, password, args.homedir)
    elif args.subcommand == "list":
        users = list_users()

        if args.json:
            print(json.dumps(users, indent=4))
        else:
            print_users(users)


def parse_args(args):
    parser = argparse.ArgumentParser(description="Manage FTP users on your Hypernode.")
    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    json_parser = subparsers.add_parser('list')
    json_parser.add_argument("--json", "-j", action='store_true', help="Print users in JSON format.")

    remove_parser = subparsers.add_parser('remove')
    remove_parser.add_argument("--username", "-u", required=True, help='The FTP user to remove.')

    add_parser = subparsers.add_parser('add')
    add_parser.add_argument("--username", "-u", required=True, help='The FTP user to add.')
    add_parser.add_argument("--password", "-p", required=False, help="The password for the user.")
    add_parser.add_argument("--homedir", "-d", required=True, help="The directory the user will be restricted to.")

    return parser.parse_args(args)


###
# List users
###

def list_users():
    try:
        with open(FTPASSWD_LOCATION) as fh:
            def get_user_and_homedir(line):
                entries = line.split(":")
                return {"user": entries[0], "homedir": entries[5]}

            return list(map(get_user_and_homedir, fh.readlines()))
    except IOError:
        return []


def print_users(users):
    if not users:
        print("No users in ftpasswd file")
    else:
        print("Users found:")
        max_userlength = max(map(len, map(itemgetter("user"), users)))
        for user in users:
            print("{1:<0}   {2}".format(max_userlength, user["user"], user["homedir"]))


###
# Add user
###

def get_password():
    password1 = getpass.getpass("Please choose a strong password: ")

    if not password1:
        print(sys.stderr, "Password cannot be empty")
        sys.exit(1)

    password2 = getpass.getpass("Retype password: ")
    if password1 != password2:
        print(sys.stderr, "Passwords do not match, try again")
        sys.exit(1)

    return password1


def add_user(user, password, homedir):
    if not homedir.startswith("/"):
        print(sys.stderr, "Cannot add user with relative home directory. Please specify absolute path.")
        sys.exit(1)

    if not os.path.isdir(homedir):
        print(sys.stderr, "Given home directory '{}' does not exist or is not a directory.".format(homedir))
        sys.exit(1)

    uid, gid = get_app_uid_and_gid()
    command = get_add_command(user, homedir, uid, gid)
    proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    proc.communicate(password.encode())

    if proc.returncode != 0:
        print(sys.stderr, "Failed to set FTP user: {}".format(proc.returncode))
        sys.exit(1)

    print("Account has been added.\n\nPlease note that you must add your IP to the whitelist in our ServicePanel "
          "before you can use the FTP \naccount. This will reconfigure the firewall so that this specific IP is "
          "allowed to connect via FTP.")


def get_app_uid_and_gid():
    # will throw errors by design when app user does not exist :)
    uid = pwd.getpwnam("app").pw_uid
    gid = grp.getgrnam("app").gr_gid
    return uid, gid


def get_add_command(user, homedir, uid, gid):
    return ["sudo", FTPASSWD_BINARY, "--passwd", "--name", user, "--home", homedir, "--file", FTPASSWD_LOCATION, "--shell", "/bin/false", "--uid", str(uid), "--gid", str(gid), "--md5", "--stdin"]


###
# Remove users
###

def remove_user(user):
    if not os.path.exists(FTPASSWD_LOCATION):
        print(sys.stderr, "No users present in ftpasswd file.")
        sys.exit(1)

    command = get_remove_command(user)
    try:
        subprocess.check_call(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if e.returncode == 1:
            print(sys.stderr, "User '{}' does not exist.".format(user))
            sys.exit(1)


def get_remove_command(user):
    return ["sudo", FTPASSWD_BINARY, "--passwd", "--name", user, "--file", FTPASSWD_LOCATION, "--delete-user"]
