from abc import ABC, abstractmethod


class ImporterSubClass(ABC):
    importer = None

    def check_runtime_parameters(self):
        """
        Check the constructor received an object to use as the Importer
        and get the global state from.
        :return None:
        """
        if self.importer is None:
            raise RuntimeError(
                "You must pass the Importer object to the constructor "
                "of every ImporterSubClass. This object keeps the global "
                "state of the program and prevents us from having to pass "
                "around variables through many functions in every part of "
                "the program."
            )

    # Any class that subclasses this abstract class must implement __init__
    # We don't really need to require __init__ because it is perfectly valid
    # to subclass this class and not want to override __init__, but because
    # python 3 abstract classes don't throw exceptions when being instantiated
    # even if the are abstract but no abstract methods are specified we need
    # to work around that by adding at least one abstract method.
    # Also see: https://stackoverflow.com/a/13646263/4158804
    @abstractmethod
    def __init__(self, importer):
        """
        Instantiate an ImporterSubClass object
        :param obj importer: The Importer object containing the
        global state.
        :return None:
        """
        self.importer = importer
        self.check_runtime_parameters()
        super(ImporterSubClass, self).__init__()
