import logging
import json
from argparse import ArgumentParser, SUPPRESS

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.settings import CMS_DEFAULT_LOCATIONS
from kamikaze3.log.log import configure_logging
from kamikaze3.importer.perform_import import Importer

log = logging.getLogger(__name__)


def add_verbosity_arguments(parser):
    """
    Add the verbosity arguments to the argparser
    :param obj parser: The instantiated ArgumentParser to add the verbosity
    arguments to
    :return None: Nothing. This function mutates the original parser object.
    """
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument(
        '--verbose', '-v',
        help='Display debug output',
        action='store_true'
    )
    verbosity.add_argument(
        '--quiet', '-q',
        help='Do not show sync progress',
        action='store_true',
        dest='quiet'
    )


def expand_settings_file(parsed_args):
    """
    If there is a settings file specified, add those arguments
    on top of the parsed args
    :param obj parsed_args: The parsed args that will be mutated
    if the --use-settings-file flag was used.
    :return None:
    """
    if parsed_args.use_settings_file:
        try:
            with open(parsed_args.use_settings_file) as settings_file:
                cli_kwargs = json.load(settings_file)
            for arg, value in cli_kwargs.items():
                setattr(parsed_args, arg, value)
        except (OSError, ValueError):
            log.error("The settings file you specified contains no valid JSON! Skipping..")


def parse_arguments(args=None):
    """
    Parse the hypernode-importer arguments
    :param list args: Optional, defaults to sys.argv
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser()

    parser.add_argument(
        '--host',
        help='SSH hostname to fetch site from'
    )
    parser.add_argument(
        '--port',
        help='SSH port to fetch site from',
        default=22,
        type=int
    )
    parser.add_argument(
        '--user',
        help='Username used for SSH connection',
        default='app'
    )
    parser.add_argument(
        '--path',
        help='Remote path to look for the CMS directory'
    )
    parser.add_argument(
        '--set-default-url',
        help='Set secure URL of default storefront and admin '
             'area to http://{}'.format(get_hypernode_name()),
        action='store_true'
    )

    skip_mutex_group = parser.add_mutually_exclusive_group()
    skip_mutex_group.add_argument(
        '--skip-db-import',
        help='Skip importing the database. Might be handy if '
             'you only want to sync files',
        action='store_true'
    )
    skip_mutex_group.add_argument(
        '--skip-file-import',
        help='Skip importing the webroot. Might be handy if you '
             'only want to sync the db',
        action='store_true'
    )

    parser.add_argument(
        '--skip-nginx-import',
        help="Never copy any found NGINX configurations from the "
             "source. This might be handy when you run the importer "
             "for a second time with --set-maintenance-on-source to "
             "perform a real migration after you have already corrected "
             "the nginx configs on the destination node",
        action='store_true'
    )
    parser.add_argument(
        '--dry-run',
        help="Don't really import but check if the source is OK",
        action='store_true'
    )
    parser.add_argument(
        '--skip-autodetect',
        help='Do not attempt to find CMS in the default locations '
             'if no CMS can be found in the provided path. The '
             'default locations we look in are {}'.format(CMS_DEFAULT_LOCATIONS),
        action='store_true'
    )
    parser.add_argument(
        '--protected-file',
        help="An optional file to pass as a 'protect' filter to rsync "
             "to exclude files during the sync",
        action='append',
        default=list()
    )
    parser.add_argument(
        '--dev',
        # Skips failure notifications
        help=SUPPRESS,
        action='store_true'
    )
    parser.add_argument(
        '--use-settings-file',
        # for when we want to import without passing arguments as
        # parameters so we don't have to deal with escaping
        help="Parse a json file for keys and values that will be used "
             "as arguments for the importer. Can bee handy for "
             "scripting purposes. Example content: "
             "{\"path\": \"/data/web/public\"}"
    )
    parser.add_argument(
        '--force',
        help="Overwrite any local changes. If you run hypernode-importer, "
             "edit files locally and run the import again, all newer local "
             "files would be untouched by default. This flag overwrites that "
             "and imports remote files over newer local changes.",
        action='store_true'
    )
    parser.add_argument(
        '--single-threaded',
        help="Normally we import the database and the filesystem at the same time. "
             "If this is too rough on the source or destination server you can choose "
             "to run those steps serially instead of in parallel. This will take longer "
             "but decrease the load on the servers.",
        action='store_true'
    )
    parser.add_argument(
        '--set-maintenance-on-source',
        action='store_true',
        help='Set the original shop in maintenance mode. You probably want this '
             'if you are moving the original shop instead of only copying it.'
    )

    add_verbosity_arguments(parser)
    parsed_args = parser.parse_args(args)
    expand_settings_file(parsed_args)

    if not parsed_args.host:
        parser.error("No --host specified, specifying a host is required")
    return parsed_args


def main():
    """
    Run the hypernode-importer
    :return int exit_code: The importer exit code. Zero on success. Nonzero on failure.
    Check out a list of all possible error codes and their meaning here:
    https://github.com/ByteInternet/hypernode-kamikaze3/blob/master/kamikaze3/importer/error_codes.py
    """
    args = parse_arguments()
    configure_logging('kamikaze3.importer', args.verbose)
    importer = Importer(arguments=args)
    return importer.run()
