from configparser import ConfigParser
from copy import deepcopy
from os import remove
from os.path import join
from uuid import uuid4

import MySQLdb
import logging

from abc import ABC, abstractmethod

from MySQLdb.constants.CR import CONN_HOST_ERROR, CONNECTION_ERROR, UNKNOWN_HOST, SERVER_LOST
from MySQLdb.constants.ER import ACCESS_DENIED_ERROR
from MySQLdb.cursors import DictCursor
from contextlib import contextmanager, suppress

from kamikaze3.importer.abstract import ImporterSubClass
from kamikaze3.importer.error_codes import NO_LOCAL_DB_ACCESS_RC, NO_REMOTE_DB_ACCESS_RC, NO_REMOTE_DB_CONN_RC, \
    DB_USERNAME_TOO_LONG_RC
from kamikaze3.importer.error_codes import NO_LOCAL_DB_CONN_RC
from kamikaze3.importer.utils import catch_and_raise_known_errors, get_free_port
from kamikaze3.settings import MYSQL_CONNECT_TIMEOUT

log = logging.getLogger(__name__)

# Normally this would come from MySQLdb.constants.ER but python3-mysqldb 1.3.7-1build2
# which we run on Hypernodes does not have this yet, so we provide the constant here instead.
WRONG_STRING_LENGTH = 1470

# These exception classes are used for flow control and saving
# a specific exit status based on which class is calling the
# database connection code. For example, if the LocalDBConnection
# class catches a CannotAccessDatabaseError we should set a
# NO_LOCAL_DB_ACCESS_RC return code but when the RemoteDBConnection
# catches it we should set a NO_REMOTE_DB_ACCESS_RC return code.


class AmbiguousDatabaseError(RuntimeError):
    pass


class CannotAccessDatabaseError(RuntimeError):
    pass


class CannotConnectToDatabaseError(RuntimeError):
    pass


class DatabaseUsernameTooLongError(RuntimeError):
    pass


KNOWN_DATABASE_ERRORS = (
    AmbiguousDatabaseError,
    CannotAccessDatabaseError,
    CannotConnectToDatabaseError,
    DatabaseUsernameTooLongError
)


def quote_password(password):
    return '"{}"'.format(password.replace('"', '\\"'))


class DBDriver(ABC):
    # Note(vdloo): The TestCase in tests/importer/db/test_db_driver.py
    # tests this abstract entirely at once. Because we need to test the abstract
    # method functionality it makes no sense to test these methods individually
    # as they contain very little business logic. We just need to check if
    # Python actually throws TypeErrors if all abstract methods we require are
    # not implemented.
    def check_parameters(self):
        """
        Check the constructor received an object to use as the DB Connection
        and get the connection details from.
        :return None:
        """
        if self.db_connection is None:
            raise RuntimeError(
                "You must pass the DBConnection object to the constructor "
                "of an DBDriver object because the configured attributes are "
                "used to determine the connection details of the remote "
                "server."
            )

    def __init__(self, db_connection):
        """
        Initialize the DBDriver object and store the passed db
        connection on the self. Check if the constructor received
        the required parameters.
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        """
        self.db_connection = db_connection
        self.check_parameters()

    @abstractmethod
    def exec_query(self, query, params=None):
        """
        All DB Drivers must implement exec_query which takes a
        query as a string and executes it on the mysql instance.
        :param str query: The query to execute on the mysql instance
        :param iter params: Prepared parameters for the query.
        Safely fills in any %s in the query string.
        :return tuple(dict, ..): A tuple of dicts representing the mysql
        rows returned from the query
        """
        pass

    @abstractmethod
    def drop_database(self, db_name=None):
        """
        Delete a database if it exists
        :param str db_name: The name of the DB to drop.
        If None, the importer.remote_db_info['db_name'] will be used.
        :return:
        """
        pass

    @abstractmethod
    def create_database(self, db_name=None):
        """
        Create a database on the local mysql instance
        :param str db_name: The name of the DB to create.
        If None, the remote_db_info['db_name'] will be used.
        :return None:
        """
        pass

    def update_my_cnf_if_needed(self):
        """
        Update the .my.cnf if needed
        :return None:
        """
        username = self.db_connection.importer.remote_db_info['username']
        if username == 'app':
            password = self.db_connection.importer.remote_db_info['password']
            # disable ConfigParser interpolation, so values like MySQL credentials can contain '%'
            parser = ConfigParser(interpolation=None)
            parser.read("/data/web/.my.cnf")
            parser.set('client', 'user', username)
            parser.set('client', 'password', quote_password(password))
            parser.write(open("/data/web/.my.cnf", 'w'))


class DBPyDriver(DBDriver):
    """
    Use python3-mysql (MySQLdb) to connect to a mysql instance and
    allow for the execution of queries on the connected mysql instance.
    """
    connection = None
    default_file = None

    @staticmethod
    def catch_known_operational_errors(e):
        """
        This method checks a caught _mysql_exceptions.OperationalError for known
        errors that we can translate to something more user-friendly for the user
        of the program. Raises generic (not OperationalError) errors for the DBDriver
        class if the error is known.
        :param obj OperationalError e: A caught OperationalError as e.
        :return None:
        """
        if e.args[0] in (CONN_HOST_ERROR, CONNECTION_ERROR):
            raise CannotConnectToDatabaseError(
                "Could not connect to the database. Is it running?"
            )
        if e.args[0] == SERVER_LOST:
            raise CannotConnectToDatabaseError(
                "Could not connect to the database. Is the port correct?"
            )
        if e.args[0] == UNKNOWN_HOST:
            raise CannotConnectToDatabaseError(
                "Could not connect to the database. Host did not resolve."
            )
        if e.args[0] == ACCESS_DENIED_ERROR:
            raise CannotAccessDatabaseError(
                "Could not log in to the database. Access denied"
            )
        if e.args[0] == WRONG_STRING_LENGTH:
            raise DatabaseUsernameTooLongError(
                "The MySQL username is too long! It should be no longer than 16 characters."
            )

    def get_connection(self):
        """
        Establishing a connection with the mysql instance using python3-mysql.
        :return None:
        """

        # Try the body of the context and catch any _mysql_exceptions.OperationalError
        # and translate them to something more humanly readable before re-raising
        # any not caught operational_error errors.
        with catch_and_raise_known_errors(
                exceptions=(MySQLdb.OperationalError, ),
                catch_function=self.catch_known_operational_errors
        ):
            self.connection = MySQLdb.connect(
                read_default_file=self.default_file,
                connect_timeout=MYSQL_CONNECT_TIMEOUT
            )

    @contextmanager
    def exec_query_context(self, query, params=None):
        """
        Execute a query and yield the cursor. The cursor is lazy and
        we have to close it afterwards. By yielding it we can prepare
        a query and have it results one by one from the mysql instance
        and we don't have to keep a lot in memory.
        :param str query: The query to execute
        :param iter params: Prepared parameters for the query.
        Safely fills in any %s in the query string.
        :yield: obj cursor: The mysql cursor.
        """
        cursor = self.connection.cursor(DictCursor)
        cursor.execute(query, params)
        self.connection.commit()
        yield cursor
        cursor.close()

    def exec_query(self, query, params=None):
        """
        Execute a query and return the output
        :param str query: The query to execute
        :param iter params: Prepared parameters for the query.
        Safely fills in any %s in the query string.
        :return tuple(dict, ..): A tuple of dicts representing the mysql
        rows returned from the query
        """
        with self.exec_query_context(query, params=params) as cursor:
            return cursor.fetchall()

    def drop_database(self, db_name=None):
        """
        Delete a database if it exists
        :param str db_name: The name of the DB to drop.
        If None, the importer.remote_db_info['db_name'] will be used.
        :return:
        """
        db_name = db_name or self.db_connection.importer.remote_db_info['db_name']
        log.info("Dropping any existing {} database".format(db_name))
        if self.db_connection.importer.local_db_conn.driver.exec_query(
            "SHOW DATABASES LIKE '{}'".format(db_name)
        ):
            self.db_connection.importer.local_db_conn.driver.exec_query(
                "DROP DATABASE IF EXISTS `{}`".format(db_name)
            )

    def create_database(self, db_name=None):
        """
        Create a database on the local mysql instance
        :param str db_name: The name of the DB to create.
        If None, the remote_db_info['db_name'] will be used.
        :return None:
        """
        db_name = db_name or self.db_connection.importer.remote_db_info['db_name']
        log.debug("Creating database '{}'".format(db_name))
        self.db_connection.importer.local_db_conn.driver.exec_query(
            "CREATE DATABASE `{}`".format(db_name)
        )
        log.debug("Allowing {} to log in to the newly created database".format(
            self.db_connection.importer.remote_db_info['username']
        ))

        mysql_version = self.db_connection.importer.local_db_conn.driver.exec_query(
            "SHOW VARIABLES WHERE Variable_name = \"version\""
        )
        # Strip the minor version number off of the mysql version
        mysql_major_version = '.'.join(mysql_version[0]['Value'].split('.')[:2])

        if mysql_major_version in ('5.6', '5.7'):
            self.db_connection.importer.local_db_conn.driver.exec_query(
                "GRANT ALL PRIVILEGES ON `{}`.* TO %s@'%%' IDENTIFIED BY %s".format(db_name),
                (
                    self.db_connection.importer.remote_db_info['username'],
                    self.db_connection.importer.remote_db_info['password']
                )
            )

        elif mysql_major_version == '8.0':
            self.db_connection.importer.local_db_conn.driver.exec_query(
                "CREATE USER IF NOT EXISTS %s@'%%' IDENTIFIED BY %s",
                (
                    self.db_connection.importer.remote_db_info['username'],
                    self.db_connection.importer.remote_db_info['password']
                )
            )
            self.db_connection.importer.local_db_conn.driver.exec_query(
                "GRANT ALL PRIVILEGES ON `{}`.* TO %s@'%%'".format(db_name),
                (
                    self.db_connection.importer.remote_db_info['username'],
                )
            )

    def check_attributes(self):
        """
        Check that all required attributes are set on the object
        :return None:
        """
        if self.default_file is None:
            raise RuntimeError(
                "You must set self.default_file "
                "before calling the DBPyDriver init"
            )

    def __init__(self, db_connection):
        """
        Initialize the DBPyDriver. Will throw an exception if the
        connection will not be able to run queries remotely.
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        """
        log.debug("Attempting to connect to the mysql instance using MySQLdb")
        super(DBPyDriver, self).__init__(
            db_connection=db_connection
        )
        self.check_attributes()
        self.get_connection()
        log.debug("Checking if we can execute queries on the database")
        self.exec_query("select 1 limit 0;")


class RemoteDBPyDriver(DBPyDriver):
    default_file = None

    def create_default_file(self, db_connection):
        """
        Create a MySQL default file with the remote db connection details
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        :return None:
        """
        self.default_file = join('/tmp', str(uuid4()) + '-default.cnf')
        # disable ConfigParser interpolation, so values like MySQL credentials can contain '%'
        config = ConfigParser(interpolation=None)
        config.add_section('client')
        config.set('client', 'user', db_connection.remote_db_info['username'])
        config.set('client', 'password', quote_password(db_connection.remote_db_info['password']))
        config.set('client', 'host', db_connection.remote_db_info['host'])
        config.set('client', 'port', str(db_connection.remote_db_info['port']))
        with open(self.default_file, 'w') as f:
            config.write(f)

    def __del__(self):
        """
        Clean up the created temporary default file as soon as the object
        leaves scope.
        :return None:
        """
        if self.default_file:
            with suppress(FileNotFoundError):
                remove(self.default_file)

    def __init__(self, db_connection):
        """
        Initialize the RemoteDBPyDriver. Will throw an exception if the
        connection will not be able to run queries remotely.
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        """
        if not db_connection.importer.remote_db_info:
            raise RuntimeError(
                "No remote database information was "
                "discovered to be used to connect!"
            )
        self.create_default_file(db_connection=db_connection)
        super(RemoteDBPyDriver, self).__init__(
            db_connection=db_connection
        )


class DirectRemoteDBPyDriver(RemoteDBPyDriver):
    @property
    def ambiguous_db_hosts(self):
        """
        Database hosts that might be ambiguous in case we connect
        to them directly from the destination node.
        :return iterable ambiguous_db_hosts: An iterable of ambiguous hosts.
        """
        return (
            'localhost',
            '127.0.0.1',
            'mysqlmaster'
        )

    def check_db_host_ambiguous(self, db_connection):
        """
        Check if the db host in the remote database info dict
        is a host that could also be wrongly available on the
        local machine. For example, if the the shop on the
        remote host has 'localhost' as the database host,
        then we should not try to connect to the local localhost
        because that is a different database server and we don't
        want to accidentally connect to that in any situation.
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        :return None:
        """
        remote_host = db_connection.importer.remote_db_info['host']
        if remote_host in self.ambiguous_db_hosts:
            log.debug("The remote DB host is ambiguous, not connecting directly")
            raise AmbiguousDatabaseError(
                "The provided database has an ambiguous address, "
                "will not attempt to connect directly."
            )

    def __init__(self, db_connection):
        """
        Initialize the DirectRemoteDBPyDriver. Will throw an exception if the
        connection will not be able to run queries remotely.
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        """
        if db_connection.importer.remote_db_info:
            self.check_db_host_ambiguous(db_connection=db_connection)
        db_connection.remote_db_info = deepcopy(db_connection.importer.remote_db_info)
        super(DirectRemoteDBPyDriver, self).__init__(
            db_connection=db_connection
        )


class TunneledRemoteDBPyDriver(RemoteDBPyDriver):
    tunnel_port = None
    tunnel_proc = None

    def forward_database(self, db_connection):
        """
        TCP forward the remote database to the local host
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        :return None:
        """
        self.tunnel_port = get_free_port()
        log.debug(
            "Forwarding remote port {} of host {} to local port {}".format(
                db_connection.importer.remote_db_info['port'],
                db_connection.importer.remote_db_info['host'],
                self.tunnel_port
            )
        )
        self.tunnel_proc = db_connection.importer.ssh_conn.driver.tcp_forward(
            remote_host=db_connection.importer.remote_db_info['host'],
            remote_port=db_connection.importer.remote_db_info['port'],
            local_port=self.tunnel_port
        )

    def __init__(self, db_connection):
        """
        Initialize the TunneledRemoteDBPyDriver. Will throw an exception
        if the connection will not be able to run queries remotely.

        TCP forwards the remote DB and then attempts to connect to it
        through the tunnel.

        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        """
        self.forward_database(db_connection=db_connection)
        db_connection.remote_db_info = deepcopy(db_connection.importer.remote_db_info)
        db_connection.remote_db_info['host'] = '127.0.0.1'
        db_connection.remote_db_info['port'] = self.tunnel_port
        super(TunneledRemoteDBPyDriver, self).__init__(
            db_connection=db_connection
        )

    def __del__(self):
        """
        Close the tunnel as soon as the object leaves scope
        :return None:
        """
        with suppress(OSError):
            self.tunnel_proc.kill()
        super(TunneledRemoteDBPyDriver, self).__del__()


class LocalDBPyDriver(DBPyDriver):
    default_file = None

    def __init__(self, db_connection):
        """
        Initialize the LocalDBPyDriver. Will throw an exception if the
        connection will not be able to run queries remotely.
        :param obj db_connection: The DB connection object with
        the details to use to connect to the mysql server with.
        """
        if not db_connection.importer.local_db_cnf_file:
            raise RuntimeError(
                "No local database information was "
                "discovered to be used to connect!"
            )

        self.default_file = db_connection.importer.local_db_cnf_file
        super(LocalDBPyDriver, self).__init__(
            db_connection=db_connection
        )


class RemoteDBConnection(ImporterSubClass):
    driver = None

    def __init__(self, importer):
        """
        Initialize the RemoteDBConnection.
        :param obj importer: The Importer object where we will put the
        keep the global state and will put the remote db connection object
        if we can connect successfully.
        :return None:
        """
        super(RemoteDBConnection, self).__init__(
            importer=importer
        )

    def attempt_connect(self):
        """
        Connect to the remote DB server using any of the configured
        remote DB connection drivers. In case a direct connection does
        not work we will attempt to TCP forward and use the database
        as if we were connecting from the remote machine directly just
        like described in the database configuration of the shop.
        :return None:
        """
        try:
            log.debug("Attempting to connect to the remote mysql server directly")
            self.driver = DirectRemoteDBPyDriver(db_connection=self)
        except KNOWN_DATABASE_ERRORS:
            log.debug(
                "Attempting to connect to the remote mysql "
                "server by TCP forwarding remote port"
            )
            self.driver = TunneledRemoteDBPyDriver(db_connection=self)

    def catch_known_database_errors(self, e):
        """
        This method checks a caught database error for known
        errors that we can translate to something more user-friendly
        for the user of the program by setting a return code.
        :param obj RuntimeError e: A caught RuntimeError as e.
        :return None:
        """
        if isinstance(e, CannotAccessDatabaseError):
            self.importer.return_code = NO_REMOTE_DB_ACCESS_RC
        if isinstance(e, CannotConnectToDatabaseError):
            self.importer.return_code = NO_REMOTE_DB_CONN_RC
        if isinstance(e, DatabaseUsernameTooLongError):
            self.importer.return_code = DB_USERNAME_TOO_LONG_RC

    def connect(self):
        """
        Connect to the remote mysql instance. Returns the DBConnection
        object.

        :return obj RemoteDBConnection: The RemoteDBConnection object
        """
        log.info("Connecting to remote database")
        with catch_and_raise_known_errors(
            exceptions=KNOWN_DATABASE_ERRORS,
            catch_function=self.catch_known_database_errors
        ):
            self.attempt_connect()
        return self


class LocalDBConnection(ImporterSubClass):
    driver = None
    default_file = None

    def __init__(self, importer):
        """
        Initialize the LocalDBConnection.
        :param obj importer: The Importer object where we will put the
        keep the global state and will put the local db connection object
        if we can connect successfully.
        :return None:
        """
        super(LocalDBConnection, self).__init__(
            importer=importer
        )

    def catch_known_database_errors(self, e):
        """
        This method checks a caught database error for known
        errors that we can translate to something more user-friendly
        for the user of the program by setting a return code.
        :param obj RuntimeError e: A caught RuntimeError as e.
        :return None:
        """
        if isinstance(e, CannotAccessDatabaseError):
            self.importer.return_code = NO_LOCAL_DB_ACCESS_RC
        if isinstance(e, CannotConnectToDatabaseError):
            self.importer.return_code = NO_LOCAL_DB_CONN_RC
        if isinstance(e, DatabaseUsernameTooLongError):
            self.importer.return_code = DB_USERNAME_TOO_LONG_RC

    def connect(self):
        """
        Connect to the local mysql instance. Returns the DBConnection
        object.

        :return obj LocalDBConnection: The LocalDBConnection object
        """
        log.info("Connecting to local database")
        with catch_and_raise_known_errors(
                exceptions=KNOWN_DATABASE_ERRORS,
                catch_function=self.catch_known_database_errors
        ):
            self.driver = LocalDBPyDriver(db_connection=self)
        return self
