"""
When an expected exception is encountered in the code the global return_code
set. Various exceptions can cause the same return_code, these exit codes
map to a set of errors that can be displayed to the user to inform them of
what they are doing wrong.

For example: the hyperform should not parse the importer output but instead just
look at the exit code and use that to decide what type of error to display.

0    - No error, command succeeded
1    - Catchall for general errors
150  - User aborted the importer with a keyboard interrupt
151  - Circular import was attempted. Can not import the same Hypernode to itself
152  - Authentication failed. Could not log in with either key auth or the provided password
153  - Unsupported special characters in SSH password while running in safe mode
154  - Key authentication failed and no alternative password was provided
155  - No Magento installation found on remote host
156  - Can not access the database defined in app/etc/local.xml
157  - Unsupported special characters in database credentials while running in safe mode
158  - No file copying tool on the remote host. Please install rsync
"""
from collections import defaultdict

NOERROR_RC = 0
CATCHALL_RC = 1
USER_ABORTED_RC = 150
CIRCULAR_IMPORT_RC = 151
SSH_AUTH_FAIL_RC = 152
SSH_SPECIAL_CHAR_RC = 153
SSH_NO_PASS_RC = 154
NO_CMS_RC = 155
NO_LOCAL_DB_ACCESS_RC = 156
DB_SPECIAL_CHAR_RC = 157
NO_FILE_COPYING_TOOL_RC = 158
MAG_2_NOT_SUPPORTED_RC = 159
SSH_TIMEOUT_RC = 160
REDIS_FLUSH_ERR_RC = 161
MAGE_FLUSH_ERR_RC = 162
NO_SUCH_DNS_RC = 163
NO_REMOTE_DB_ACCESS_RC = 164
SSH_NO_ROUTE_RC = 165
NO_LOCAL_DB_CONN_RC = 166
REMOTE_DB_CONF_NOT_FOUND_RC = 167
REMOTE_DB_CONF_PARSE_ERR_RC = 168
CAN_NOT_SYNC_FILES_RC = 169
NO_REMOTE_DB_CONN_RC = 170
DB_NAME_SPECIAL_CHAR_RC = 171
DB_TABLE_PREFIX_SPECIAL_CHAR_RC = 172
REMOTE_DB_M2_CONF_PARSE_ERR_RC = 173
REMOTE_DB_M2_CONF_NOT_FOUND_RC = 174
SSH_WRONG_HN_USER_PROVIDED_RC = 175
SSH_CANT_SET_MAINTENANCE_RC = 176
SSH_CANT_UNSET_MAINTENANCE_RC = 177
DISK_TOO_SMALL_RC = 178
SSH_INFO_INVALID_RC = 179
WRONG_PATH_RC = 181
IPV6_RC = 182
SSH_PORT_NOT_AVAILABLE_RC = 183
DB_USERNAME_TOO_LONG_RC = 184
DB_NAME_TOO_LONG_RC = 185
REMOTE_DB_SW5_CONF_PARSE_ERR_RC = 186
REMOTE_DB_SW6_CONF_PARSE_ERR_RC = 187
REMOTE_DB_SW6_CONF_NOT_FOUND_RC = 188
REMOTE_DB_SW5_CONF_NOT_FOUND_RC = 189
REMOTE_APP_NAME_NOT_FOUND_SW6_RC = 190
EXEC_QUERY_CANT_SET_REMOTE_MAINTENANCE = 191
EXEC_COMMAND_CANT_CLEAR_CACHE_REMOTE_MAINTENANCE = 192
CAN_NOT_SET_BASE_URL = 193

IMPORTER_ERROR_MESSAGES = defaultdict(lambda: "of an uncaught error",)
IMPORTER_ERROR_MESSAGES.update({
    NOERROR_RC: None,
    USER_ABORTED_RC: 'the importer was killed',
    CIRCULAR_IMPORT_RC: 'you can not import a hypernode to itself',
    SSH_AUTH_FAIL_RC: 'the ssh credentials were wrong (key + password)',
    SSH_SPECIAL_CHAR_RC: "there were specials chars in ssh password, our automation can not handle this",
    SSH_NO_PASS_RC: "the ssh key was wrong. no password provided.",
    SSH_TIMEOUT_RC: "we could not connect to the provided SSH host (timed out)",
    SSH_NO_ROUTE_RC: "we could not connect to the provided SSH host (no route to host)",
    NO_CMS_RC: "we couldn't find a webshop installation",
    NO_LOCAL_DB_ACCESS_RC: "we can not access the local database. Access denied. The connection details "
                           "in the ~/.my.cnf file do not allow us to log in.",
    NO_LOCAL_DB_CONN_RC: "we can not connect to the local database. Failed to connect.",
    DB_SPECIAL_CHAR_RC: "there were special chars in the db password, our automation can not handle this",
    DB_USERNAME_TOO_LONG_RC: "the database username was too long. It can be 16 characters max.",
    DB_NAME_TOO_LONG_RC: "the database name was too long. It can be 64 characters max.",
    DB_NAME_SPECIAL_CHAR_RC: "there were special chars in the database name, "
                             "please rename the database to something without "
                             "special characters to prevent problems",
    DB_TABLE_PREFIX_SPECIAL_CHAR_RC: "there were special chars in the table prefix, "
                                     "please change the table prefix to something without "
                                     "special characters to prevent problems",
    NO_FILE_COPYING_TOOL_RC: "rsync was not installed on the other server",
    CAN_NOT_SYNC_FILES_RC: "failed to synchronize test file from remote server. Are your permissions correct? "
                           "Does your app/Mage.php exist?",
    MAG_2_NOT_SUPPORTED_RC: "the shop was magento 2 and this is not supported by the importer yet",
    REDIS_FLUSH_ERR_RC: "we failed to flush the Redis cache, the rest of the import might still have succeeded. "
                        "This was one of the last steps of the importer. You can try to flush it manually with "
                        "'redis-cli flushall'. Also make sure to flush your Magento cache with 'magerun cache:flush'",
    MAGE_FLUSH_ERR_RC: "we failed to flush the Magento cache, the rest of the import might still have succeeded. "
                       "Flushing the cache is the last step of the importer. You can try to flush it manually with "
                       "'magerun cache:flush'",
    NO_SUCH_DNS_RC: "the provided host can not be resolved. Is the DNS correct? Name or service not known.",
    NO_REMOTE_DB_ACCESS_RC: "we can not access the remote database with the detected credentials",
    NO_REMOTE_DB_CONN_RC: "we can not connect to the remote database. Failed to connect.",
    REMOTE_DB_CONF_NOT_FOUND_RC: "Could not find a connection block in the remote database config. "
                                 "Is your .xml correct?",
    REMOTE_DB_CONF_PARSE_ERR_RC: "Could not parse the remote database config. Is your .xml correct?",
    REMOTE_DB_M2_CONF_PARSE_ERR_RC: "Could not parse the remote database config. Is your env.php correct? "
                                    "Please use one statement per line and single quotes around the values.",
    REMOTE_DB_M2_CONF_NOT_FOUND_RC: "Could not find the remote php env file. Is your path correct?",
    SSH_WRONG_HN_USER_PROVIDED_RC: "Wrong user provided for hypernode import. Make sure you've supplied the correct "
                                   "user and the spelling is correct. Generally for Hypernode domains the SSH user "
                                   "should be 'app'.",
    SSH_CANT_SET_MAINTENANCE_RC: "Could not set maintenance flag on the remote host. Are the permissions correct?",
    SSH_CANT_UNSET_MAINTENANCE_RC: "Could not unset maintenance flag on the remote host. Are the permissions correct?",
    DISK_TOO_SMALL_RC: "The disk size on the hypernode is lower than the size of the CMS installation.",
    SSH_INFO_INVALID_RC: "The SSH information entered does not seem valid. Is it correct? Perhaps you made a typo?",
    WRONG_PATH_RC: "Could not find an installation in the provided path, however an installation"
                   " was found in a different path.",
    IPV6_RC: "Could not resolve remote address, because it was most likely an IPv6 address. Is it possible to "
             "change this address to an IPv4 address?",
    SSH_PORT_NOT_AVAILABLE_RC: "Something went wrong. Most likely this is because we couldn't SSH through the provided "
                               "SSH port. Is it possible to enable access to this port or supply a different one? "
                               "Usually port 22 is used by default for SSH.",
    REMOTE_DB_SW5_CONF_PARSE_ERR_RC: "Could not parse the remote database config. Is your config.php correct? "
                                     "Please use one statement per line and single quotes around the values.",
    REMOTE_DB_SW6_CONF_PARSE_ERR_RC: "Could not parse the remote database config. Is your .env correct? "
                                     "Please use one statement per line and single quotes around the values.",
    REMOTE_DB_SW6_CONF_NOT_FOUND_RC: "Could not find the shopware 6 .env config file, does it exist?",
    REMOTE_DB_SW5_CONF_NOT_FOUND_RC: "Could not find the shopware 5 config.php config file, does it exist?",
    REMOTE_APP_NAME_NOT_FOUND_SW6_RC: "Was not able to acquire the remote app name (APP_URL) from the .env config",
    EXEC_QUERY_CANT_SET_REMOTE_MAINTENANCE: "We were not able to set maintenance mode on the source shop",
    EXEC_COMMAND_CANT_CLEAR_CACHE_REMOTE_MAINTENANCE: "We were not able to set maintenance mode on the source shop,"
                                                      " because we were not able to find the Shopware console",

})
