import logging
from abc import ABC, abstractmethod
from contextlib import suppress
from os import remove
from os.path import join, isfile
from shlex import quote
from uuid import uuid4

from kamikaze3.importer.abstract import ImporterSubClass
from kamikaze3.importer.error_codes import NO_FILE_COPYING_TOOL_RC, CAN_NOT_SYNC_FILES_RC
from kamikaze3.importer.utils import catch_and_raise_known_errors, compose_sshpass_env
from kamikaze3.common.shell import run_local_command

log = logging.getLogger(__name__)


class FileSyncNotAvailable(RuntimeError):
    pass


class RemoteRsyncNotAvailable(FileSyncNotAvailable):
    pass


class RemoteRsyncVerificationFailed(FileSyncNotAvailable):
    pass


class RsyncFailed(FileSyncNotAvailable):
    pass


class FileTransferDriver(ABC):
    # Note(vdloo): The TestCase in tests/importer/file_transfer/file_transfer_driver/
    # test_file_transfer_driver.py tests this abstract class entirely at once. Because
    # we need to test the abstract method functionality it makes no sense to test these
    # methods individually as they contain very little business logic. We just need to
    # check if Python actually throws TypeErrors if all abstract methods we require are
    # not implemented.
    def check_parameters(self):
        """
        Check the constructor received an object to use as the SSH Connection
        and get the connection details from.
        :return None:
        """
        if self.file_transfer_connection is None:
            raise RuntimeError(
                "You must pass the FileTransferConnection object to the constructor "
                "of a FileTransferDriver object because the configured attributes are "
                "used to determine the connection details of the remote server."
            )

    def __init__(self, file_transfer_connection):
        """
        Initialize the FileTransferDriver object and store the file
        transfer connection object on the self.
        :param obj file_transfer_connection: The File transfer connection
        object with the details to use to connect to sync files from the
        remote server with.
        """
        self.file_transfer_connection = file_transfer_connection
        self.check_parameters()
        self.verify_can_download()

    @abstractmethod
    def download(self, source, destination):
        """
        All FileTransfer Drivers must implement download which takes
        a source and destination path.
        :param str source: The source path on the remote host
        :param str destination: The destination path on the local host
        :return None:
        """
        pass

    def verify_download_file(self, source_file, temp_dest, message):
        """
        Download a remote file to a local temp file and raise an error
        if that failed.
        Note(vdloo): This method is tested as part of
        the verify_download_magento_{1,2}_file tests
        :param str source_file: The remote file to download
        :param str temp_dest: The name of the temp file to
        download it to
        :param str message: Message of the error to raise if
        the sync fails
        :return None:
        """
        remote_verification_file = join(
            self.file_transfer_connection.importer.remote_path,
            source_file
        )
        local_verification_file = join('/tmp', str(uuid4()) + '-' + temp_dest)
        try:
            self.download(remote_verification_file, local_verification_file)
            if not isfile(local_verification_file):
                raise RemoteRsyncVerificationFailed(message)
        finally:
            with suppress(FileNotFoundError):
                remove(local_verification_file)

    def verify_download_magento_1_file(self):
        """
        Download a file from the magento 1 installation to verify
        we are actually capable of downloading files from the remote
        server.
        :return None:
        """
        self.verify_download_file(
            'app/Mage.php', 'Mage.php',
            message="Failed to sync Magento 1 test file from remote host"
        )

    def verify_download_magento_2_file(self):
        """
        Download a file from the magento 2 installation to verify
        we are actually capable of downloading files from the remote
        server.
        :return None:
        """
        self.verify_download_file(
            'pub/static.php', 'static.php',
            message="Failed to sync Magento 2 test file from remote host"
        )

    def verify_download_shopware_5_file(self):
        """
        Download a file from the shoppware 5 installation to verify
        we are actually capable of downloading files from the remote
        server.
        :return None:
        """
        self.verify_download_file(
            'shopware.php', 'shopware.php',
            message="Failed to sync Shopware 5 test file from remote host"
        )

    def verify_download_shopware_6_file(self):
        """
        Download a file from the shopware 6 installation to verify
        we are actually capable of downloading files from the remote
        server.
        :return None:
        """
        self.verify_download_file(
            'public/index.php', 'index.php',
            message="Failed to sync Shopware 6 test file from remote host"
        )

    def verify_can_download(self):
        """
        Download a file to verify that we can actually download using
        this driver.
        :return None:
        """
        log.debug("Verifying file transfer is possible by syncing test file")
        if self.file_transfer_connection.importer.site_type == 'magento1':
            self.verify_download_magento_1_file()
        elif self.file_transfer_connection.importer.site_type == 'magento2':
            self.verify_download_magento_2_file()
        elif self.file_transfer_connection.importer.site_type == 'shopware5':
            self.verify_download_shopware_5_file()
        elif self.file_transfer_connection.importer.site_type == 'shopware6':
            self.verify_download_shopware_6_file()
        else:
            raise NotImplementedError(
                "Downloading a test file for {} has not been implemented "
                "yet".format(self.file_transfer_connection.importer.site_type)
            )


class AbstractRsyncDriver(FileTransferDriver, ABC):
    """
    Abstract class for rsync based file transfer drivers
    """
    @property
    def rsync_io_error_return_code(self):
        return 23

    @property
    def rsync_file_vanished_return_code(self):
        return 24

    def compose_rsync_command(self, source, destination, rsync_options=tuple(),
                              ssh_options=tuple(), protected_files=tuple()):
        """
        Create the rsync command for the specified source, destination and options
        :param str source: The path on the remote host to download
        :param str destination: The path on the local host to sync it to
        :param iter rsync_options: Things like extra --excludes in an iterable
        :param iter ssh_options: Options for the inner '-e ssh' command.
        For example string -oProxyCommand=none in an interable.

        WARNING: The SSH OPTIONS and RSYNC OPTIONS arl NOT SANITIZED. Do not put
        user input here when running in an environment where the user is not allowed
        to do local code execution. You might be able to break out of the shell if you do.
        For example, don't let the service panel or hyperformance web interface
        fill in this importer options flag in the argparser.

        :param iter protected_files: iterable of protected files. These files are
        excluded and filtered from the sync.
        :return tuple rsync_command: A tuple containing all command sections.
        """
        host = quote(self.file_transfer_connection.importer.host)
        rsync_source_segment = "{}@{}:{}".format(
            quote(self.file_transfer_connection.importer.user),
            host,
            quote(source),
        )

        protected_filters = ['--filter={}'.format('P {}'.format(f)) for f in protected_files]
        excludes = ['--exclude={}'.format(f) for f in protected_files]

        ssh_command = 'ssh {} -o StrictHostKeyChecking=no ' \
                      '-o UserKnownHostsFile=/dev/null ' \
                      '-o VerifyHostKeyDNS=no ' \
                      '-o Loglevel=error -p {}'.format(' '.join(ssh_options),
                                                       self.file_transfer_connection.importer.port)

        if self.file_transfer_connection.importer.ssh_password:
            ssh_pass_command = ('sshpass', '-e', '--')
        else:
            ssh_pass_command = tuple()

        if self.file_transfer_connection.importer.force:
            update_option = tuple()
        else:
            update_option = ('--update',)

        if self.file_transfer_connection.importer.quiet:
            progress_option = tuple()
        else:
            progress_option = ('--info=progress2',)

        return (
            *ssh_pass_command, 'rsync', *progress_option,
            '--archive', '--copy-unsafe-links',
            '--delete', *update_option, rsync_source_segment, destination,
            *excludes, *protected_filters, *rsync_options,
            '-e', ssh_command,
        )

    def rsync_download(self, source, destination, rsync_options=tuple(),
                       ssh_options=tuple(), protected_files=tuple()):
        """
        Use the local shell rsync to download a file.
        :param str source: The path on the remote host to download
        :param str destination: The path on the local host to sync it to
        :param iter rsync_options: Things like extra --excludes in an iterable
        :param iter ssh_options: Options for the inner '-e ssh' command.
        For example string -oProxyCommand=none in an interable.
        :param iter protected_files: iterable of protected files. These files are
        excluded and filtered from the sync.
        :return NoneType None: Raises error on failure
        """

        rsync_command = self.compose_rsync_command(
            source, destination, rsync_options=rsync_options,
            ssh_options=ssh_options, protected_files=protected_files
        )

        rsync_environment = compose_sshpass_env(
            password=self.file_transfer_connection.importer.ssh_password
        )

        return_code = run_local_command(
            rsync_command,
            # This is important. If we give the command a full shell people
            # might be able to escape the command and perform undesired commands
            # based on user input.
            shell=False,
            environment=rsync_environment,
            timeout=self.file_transfer_connection.importer.file_sync_timeout
        )

        if return_code != 0:
            if return_code in (self.rsync_io_error_return_code, self.rsync_file_vanished_return_code):
                log.warning("File transfer: some files could not be transfered.")
            else:
                raise RsyncFailed(
                    "Rsync failed to transfer the files"
                )

    def try_rsync_download(self, source, destination, rsync_options=tuple(), protected_files=tuple()):
        """
        Attempt to download files using rsync. If the proxy fails (if there is one),
        it will retry without the proxy. If it still fails it will raise.
        :param str source: The path on the remote host to download
        :param str destination: The path on the local host to sync it to
        :param iter rsync_options: Things like extra --excludes in an iterable
        :param iter protected_files: iterable of protected files. These files are
        excluded and filtered from the sync.
        :return NoneType None: Raises error on failure
        """
        try:
            self.rsync_download(
                source, destination, rsync_options=rsync_options,
                protected_files=protected_files
            )
        except RsyncFailed:
            log.debug(
                "Failed syncing the files. Trying again with different "
                "SSH config options (forcing ProxyCommand off)"
            )
            self.rsync_download(
                source, destination, rsync_options=rsync_options,
                ssh_options=('-oProxyCommand=none',), protected_files=protected_files
            )

    @abstractmethod
    def download(self, source, destination):
        """
        All AbstractRsyncDrivers must implement download which takes
        a source and destination path.
        :param str source: The source path on the remote host
        :param str destination: The destination path on the local host
        :return None:
        """
        pass


class FileTransferRsyncDriver(AbstractRsyncDriver):
    """
    Sync files from the remote host using the remote available rsync
    """
    def __init__(self, file_transfer_connection):
        """
        Initialize the FileTransferRsyncDriver. Will throw an error if
        rsync was not found on the remote host or syncing a test file fails.
        :param obj file_transfer_connection: The File Transfer connection
        object with the details of how to use rsync for this host with.
        """
        log.debug("Attempting to sync a file using rsync")
        super(FileTransferRsyncDriver, self).__init__(
            file_transfer_connection=file_transfer_connection
        )
        if not self.file_transfer_connection.importer.remote_rsync_available:
            raise RemoteRsyncNotAvailable("Rsync is not available on the remote host!")

    def download(self, source, destination, protected_files=None):
        """
        Use rsync to download files from the remote host
        """
        if not protected_files:
            protected_files = []
        protected_files.extend(self.file_transfer_connection.importer.protected_file)
        self.try_rsync_download(source, destination, protected_files=protected_files)


class FileTransferConnection(ImporterSubClass):
    driver = None

    def __init__(self, importer):
        """
        Initialize the FileTransferConnection.
        :param obj importer: The Importer object where we will put the
        keep the global state and will put the file transfer connection
        object if we can transfer files.
        :return None:
        """
        super(FileTransferConnection, self).__init__(importer=importer)

    def catch_known_file_sync_errors(self, e):
        """
        This method checks a caught FileSyncNotAvailable for known errors
        that we can translate to something more user-friendly for the user
        of the program. Sets specific exit codes.
        :param obj SSHException e: A caught FileSyncNotAvailable as e.
        :return None:
        """
        if isinstance(e, RemoteRsyncNotAvailable):
            self.importer.return_code = NO_FILE_COPYING_TOOL_RC
        if isinstance(e, RemoteRsyncVerificationFailed):
            self.importer.return_code = CAN_NOT_SYNC_FILES_RC

    def connect(self):
        """
        Connect to a remote SSH server. Returns the SSHConnection
        object and a password if one was prompted for and allowed
        the program to connect successfully to the remote server.

        Use the specified connection details to connect to the remote
        server and instantiate the first SSH driver object that works.
        We attempt to use Paramiko first but in case that does not work
        for any reason we try to fall back to connecting using the
        commandline SSH utility. If we can't connect with any of the SSH
        keys available to the program, we prompt for a password and try
        again once more.

        :return tuple (file_transfer_connection, password): The SSHConnection
        object and the password if any.
        """
        log.info("Creating a file transfer connection")
        with catch_and_raise_known_errors(
                exceptions=(FileSyncNotAvailable,),
                catch_function=self.catch_known_file_sync_errors
        ):
            self.driver = FileTransferRsyncDriver(file_transfer_connection=self)
            # TODO(vdloo): implement alternative upload
            # statically compiled rsync driver (nice to have)
            # TODO(vdloo): implement alternative upload
            # using paramiko scp driver (nice to have)
        return self
