import logging
from os import mkdir
from os.path import join
from contextlib import contextmanager
from shutil import rmtree
from subprocess import check_output

from kamikaze3.settings import HYPERNODE_NGINX_PATH, \
    HMV_STATEFILE_PATH, HYPERNODE_HMV_IMPORTED_PATH


logger = logging.getLogger(__name__)
MAGENTO_2_NGINX_FLAG = '/data/web/nginx/magento2.flag'
NGINX_DOWNLOAD_EXCLUDE = ['whitelist-development-exception.conf', 'http.magerunmaps', 'htpasswd-development',
                          'letsencrypt.flag', 'ssl/*']


def copy_app_nginx_files(importer):
    """
    Copy over the nginx configuration from an existing
    Hypernode(like) host where the hardcoded /data/web/nginx
    path exists. This strategy is used if the Hypernode that
    is being imported from does NOT use hypernode-managed-vhosts
    but instead has the legacy Nginx configuration.
    :param obj importer: The instantiated hypernode-importer object
    :return None:
    """
    logger.info("Downloading nginx directory from remote host")
    importer.file_transfer_conn.driver.download(
        importer.remote_nginx_path, HYPERNODE_NGINX_PATH,
        protected_files=NGINX_DOWNLOAD_EXCLUDE,
    )


def download_hmv_nginx_configs_for_import(importer):
    """
    Download the nginx config from the remote hmv
    enabled hypernode into the directory hmv will
    later import from.
    :param obj importer: The instantiated hypernode-importer object
    :return None:
    """
    logger.info("Downloading hmv-enabled nginx directory from remote host")
    imported_path = join(HYPERNODE_HMV_IMPORTED_PATH, 'nginx')
    importer.file_transfer_conn.driver.download(
        importer.remote_nginx_path, imported_path,
        protected_files=NGINX_DOWNLOAD_EXCLUDE,
    )


def download_hmv_statefile_for_import(importer):
    """
    Download the hmv statefile from the remote hmv
    enabled hypernode into the directory hmv will
    later import from.
    :param obj importer: The instantiated hypernode-importer object
    :return None:
    """
    logger.info("Downloading hmv statefile from remote host")
    importer.file_transfer_conn.driver.download(
        HMV_STATEFILE_PATH, HYPERNODE_HMV_IMPORTED_PATH
    )


@contextmanager
def ensure_hmv_imported_path_cleaned_up():
    """
    Make sure that the hmv imported path
    is cleaned up before we download the
    files to be imported and that that
    dir is cleaned up after we do.
    :return None:
    """
    rmtree(HYPERNODE_HMV_IMPORTED_PATH, ignore_errors=True)
    mkdir(HYPERNODE_HMV_IMPORTED_PATH)
    try:
        yield
    finally:
        rmtree(HYPERNODE_HMV_IMPORTED_PATH, ignore_errors=True)


def run_hmv_import():
    """
    Perform the hmv import. This will overwrite the old
    appname.hypernode.io to the new appname.hypernode.io
    in all references of that string in the nginx config
    content and the vhost names. A side-effect of this is
    also that the hmv statefiles and the nginx configuration
    currently active on the node that is imported on will
    be moved away to a backup directory (those configs are
    then no longer active).
    :return None:
    """
    logger.info("Importing hypernode-managed-vhosts configuration")
    hmv_import_command = [
        'hypernode-manage-vhosts',
        '--import-dir',
        join(HYPERNODE_HMV_IMPORTED_PATH, 'nginx'),
        '--protected-files',
        *NGINX_DOWNLOAD_EXCLUDE,
        '--import-statefile',
        join(HYPERNODE_HMV_IMPORTED_PATH, 'nginx_vhosts.json'),
        '--disable-https',
        '--disable-force-https',
    ]
    output = check_output(hmv_import_command)
    logger.info(output.decode('utf-8'))


def copy_app_nginx_hmv_files(importer):
    """
    Copy over the nginx configs and hmv statefile from the
    other Hypernode that has hmv enabled. These configs
    will then be imported using a hypernode-managed-vhosts
    --import command. In the configs and vhosts the old
    appname.hypernode.io will be replaced with the new
    appname.hypernode.io. The most common use-case here
    will be people importing a shop from their production
    node to a dev plan
    :param obj importer: The instantiated hypernode-importer object
    :return None:
    """
    logger.info(
        "Will now attempt to import the hypernode-managed-vhosts "
        "configuration from the remote server"
    )
    with ensure_hmv_imported_path_cleaned_up():
        download_hmv_nginx_configs_for_import(importer)
        download_hmv_statefile_for_import(importer)
        run_hmv_import()
