import smtplib
import paramiko
import logging
import json
from sys import exc_info, argv
from os import environ
from shlex import quote
from traceback import format_tb

from kamikaze3.common.env import get_app_name, host_is_a_docker
from kamikaze3.common.slack import try_send_slack_message
from kamikaze3.importer.error_codes import CATCHALL_RC, IMPORTER_ERROR_MESSAGES
from kamikaze3.importer.ssh import host_is_a_vagrant
from email.mime.text import MIMEText
from socket import gethostname

from kamikaze3.onboarding import call_event_hook

log = logging.getLogger(__name__)

NOTIFICATION_IGNORE_EXCEPTIONS = (paramiko.AuthenticationException,)

MAIL_EXCEPTION_SENDER = 'HN Importer <noreply@hypernode.com>'
MAIL_EXCEPTION_RECIPIENT = ['hypernode@byte.nl', 'onboarding@byte.nl']
IMPORTER_RESULT_CHANNELS = ('#onboarding-actie',)
MAIL_BODY = """
Hypernode importer failed. Output:

%(output)s

*** Exception:

%(exc_type)s: %(exc)s\n
%(traceback)s

*** CLI arguments:

%(cli_args)s

*** Using password: %(used_password)s

*** ENV:

%(env)s
"""


def mail_exception(output, importer):
    """
    Mail information about the failed import
    :param str output: The Exception output
    :param obj importer: The importer object with all the gathered import state
    :return None:
    """
    if not host_is_a_vagrant() and not host_is_a_docker():
        exc_type, exc, traceback = exc_info()
        if exc.__class__ in NOTIFICATION_IGNORE_EXCEPTIONS:
            return
        error = {
            'exc': exc,
            'output': output,
            'exc_type': exc_type,
            'traceback': "".join(format_tb(traceback)),
            'used_password': bool(importer.ssh_password),
            'env': json.dumps(dict(environ), indent=4, sort_keys=True),
            'cli_args': ' '.join(quote(x) for x in argv)
        }
        msg = MIMEText(MAIL_BODY % error)

        subject = "hypernode-importer crashed on {}".format(gethostname())
        if importer.return_code == CATCHALL_RC:
            subject = "Uncaught exception: {}".format(subject)

        msg['Subject'] = subject
        msg['From'] = MAIL_EXCEPTION_SENDER
        msg['To'] = ", ".join(MAIL_EXCEPTION_RECIPIENT)
        smtp = smtplib.SMTP('localhost')
        try:
            smtp.sendmail(MAIL_EXCEPTION_SENDER, MAIL_EXCEPTION_RECIPIENT, msg.as_string())
        except smtplib.SMTPException:
            log.error("Import failed and the script failed to send the exception to Byte technicians, if you think "
                      "this is something we should take a look at please copy this output and create a ticket.")


def send_importer_result_to_slack(site_type, success=False, return_code=CATCHALL_RC):
    app_name = get_app_name()
    app_identifier = 'app {}'.format(app_name) if app_name else 'an app with an unparseable hostname'

    emoticon = ":party:" if success else ":anguished:"
    success_message = "was successfully imported"
    failure_message = "failed to import because {}".format(IMPORTER_ERROR_MESSAGES[return_code])
    result = success_message if success else failure_message
    for channel in IMPORTER_RESULT_CHANNELS:
        try_send_slack_message(
            "{} A {} shop on {} {}.".format(
                emoticon,
                site_type or 'magento',
                app_identifier,
                result
            ),
            channel=channel
        )


def send_hypernode_importer_success_event():
    call_event_hook(['hypernode_importer_success'])


def send_hypernode_importer_failure_event():
    call_event_hook(['hypernode_importer_failure'])


def send_failure_notifications(output, importer):
    """
    Send information about the failed import
    :param str output: The Exception output
    :param obj importer: The importer object with all the gathered import state
    :return None:
    """
    mail_exception(output, importer)
    send_importer_result_to_slack(importer.site_type, success=False, return_code=importer.return_code)
    send_hypernode_importer_failure_event()


def send_success_notifications(importer):
    """
    Send information about the successful import
    :param obj importer: The importer object with all the gathered import state
    :return None:
    """
    send_importer_result_to_slack(importer.site_type, success=True)
    send_hypernode_importer_success_event()
