import logging
from contextlib import contextmanager

from kamikaze3.common.slack import try_send_slack_message
from kamikaze3.importer.db import LocalDBConnection, RemoteDBConnection
from kamikaze3.importer.discover import DiscoverConnection
from kamikaze3.importer.error_codes import NOERROR_RC, CATCHALL_RC, USER_ABORTED_RC
from kamikaze3.importer.file_transfer import FileTransferConnection
from kamikaze3.importer.notifications import send_failure_notifications, send_success_notifications
from kamikaze3.importer.ssh import SSHConnection
from kamikaze3.importer.strategy.determine import ImportStrategy

log = logging.getLogger(__name__)


class Importer(object):
    """
    The hypernode-importer. Imports a Magento shop from a remote server or archive.
    There is a lot of state in this applications, which is why the interface to
    this piece of functionality is an object. With a more imperative / functional
    approach the result would be that a lot of variables have to be threaded into
    lower functions. Because we often need to detect and register information about
    the environment on which the code is running (both locally and remote) it is
    convenient to have a place to put all information without interrupting the
    control flow of the program by passing data through multiple levels of function
    calls.
    """

    # These are the static constant variables that don't change. They are properties
    # because that's as good as it gets in python for making object attributes read
    # only using the standard lib.
    @property
    def local_db_cnf_file(self):
        return '/data/web/.my.cnf'

    @property
    def local_db_host(self):
        return 'mysqlmaster'

    @property
    def known_site_types(self):
        """
        NOTE(vdloo): These are the known site-types, we can expand this in
        the future. These are not the site types the importer supports, those
        are defined in the importer.discover.find_remote_magento_installation
        logic. We also identify not supported site types so we can display
        more specific errors (like "can't import, Magento 2 not supported).
        It also gives us insight in what types of sites people try to import
        and if we maybe should start writing support for those types. The
        idea is that if the importer crashes we can analyse the type of import
        performed based on the attributes set on the importer object.

        :return iter site_types: An iterable of site types
        """
        return (
            'magento1',
            'magento2',
            'shopware5',
            'shopware6'
        )

    def set_site_type(self, site_type):
        """
        Set the site type but validate if it is a know site type, if not,
        someone probably made a programming error. This setter serves as a
        sanity check. If it breaks during runtime the customer should report
        this error to use (or even better, we should find out from the trace)
        and then we should debug.
        :param str site_type: The site type that was discovered.
        :return None:
        """
        if site_type not in self.known_site_types:
            try_send_slack_message(
                ":bug: Looks like someone shipped a bug in hypernode-importer. "
                "An unsupported site-type was set. If you implement a "
                "find_remote_*_installation in importer.discover.py, you "
                "also need to add that site type to importer.known_site_types "
                "and implement a strategy for this site type in the "
                "importer.strategy.determine.ImportStrategy module."
            )
            raise RuntimeError(
                "Discovered site type {}. The importer does not know "
                "how to import these types of sites.".format(site_type)
            )
        self.site_type = site_type

    @property
    def cms_default_locations(self):
        """
        TODO(vdloo): Nice to have, we should keep track of what locations are most
        often used and then order this list based on the most common paths for faster
        cms path discovery.

        Paths where the importer will look for cms in if autodetect is enabled.
        :return iter paths: An iterable of paths
        """
        return (
            '$HOME',
            '/var/www',
            '/var/html',
            '/www',
            '/app',  # Magento 2 enterprise cloud servers
        )

    @property
    def file_sync_timeout(self):
        """
        If the file sync takes longer than this amount of seconds
        it will be killed.
        :return int file_sync_timeout: The time a sync is allowed to take
        """
        return 43200  # 12 hours

    # These configurables must be passed to the init
    # They are declared here because if we only set them
    # dynamically on self.__dict__ as we do in __init__ then
    # pycharm won't understand that these attributes are
    # available on the object and make everything orange.
    verbose = None
    quiet = None
    host = None
    port = None
    user = None
    path = None
    set_default_url = None
    skip_db_import = None
    skip_file_import = None
    skip_nginx_import = None
    dry_run = None
    skip_autodetect = None
    protected_file = None
    dev = None
    use_settings_file = None
    site_type = None
    remote_db_info = None
    force = None
    single_threaded = None

    # These are the variables that are configured during
    # runtime based on the environment on/against which the
    # importer is running.
    skip_file_sync = None
    skip_db_sync = None
    skip_remote_conn = None
    ssh_conn = None
    ssh_password = None
    local_db_conn = None
    discover_conn = None
    remote_path = None
    remote_env_path = None
    remote_rsync_available = None
    file_transfer_conn = None
    remote_db_conn = None
    local_hmv_active = None
    remote_hmv_active = None
    import_strategy = None
    import_nginx_strategy = None
    import_warnings = []

    def check_can_skip_remote_conn(self):
        """
        Check if we can skip creating a remote connection based on the
        configured importer options. If we don't need to sync the files
        or the database because we are importing from a local archive
        instead, then we don't need to log in to any remote machine over
        SSH so we can skip checking if we can.
        :return bool skip_remote_conn: True if we can skip creating a
        remote connection, False if not.
        """
        self.skip_file_sync = self.skip_file_import
        self.skip_db_sync = self.skip_db_import
        self.skip_remote_conn = self.skip_file_sync and self.skip_db_sync

    def check_can_skip_default_url(self):
        """
        Check if we can skip set_default_url based on the configured
        importer options. If we are setting maintenance mode on the source
        environment we want to migrate the shop and use the urls from the
        source environment. We dont want to set a default url then.
        :return none
        """
        self.set_default_url = self.set_default_url and not self.set_maintenance_on_source

    def __init__(self, arguments):
        """
        Constructor for the Importer object. Sets the instantiation time
        options on the object and defines a zero exit code which can be
        overwritten later by the program if any failure occurs.
        :param obj arguments: The parsed arguments which will be used
        to configure importer functionality based on user input.
        """

        # Shallow copy all argument attributes to the Importer object.
        # This sets all the flags and variables.
        self.__dict__ = arguments.__dict__.copy()
        self.check_can_skip_remote_conn()
        self.check_can_skip_default_url()
        self.return_code = NOERROR_RC

    def initiate_connections(self):
        """
        Instantiate connections to the required remote services.
        :return None:
        """
        self.local_db_conn = LocalDBConnection(importer=self).connect()
        if not self.skip_remote_conn:
            self.ssh_conn, self.ssh_password = SSHConnection(importer=self).connect()
            self.discover_conn = DiscoverConnection(importer=self).connect()
        if not self.skip_file_sync:
            self.file_transfer_conn = FileTransferConnection(importer=self).connect()
        if not self.skip_db_import:
            self.remote_db_conn = RemoteDBConnection(importer=self).connect()

    def prepare_import(self):
        """
        Prepare the actions that the importer should take to import the site
        based on the discovered information about the shop on the remote side.
        :return None:
        """
        self.import_strategy = ImportStrategy(importer=self)

    @contextmanager
    def flatten_exception(self):
        """
        NOTE(vdloo): This method is tested as part of the run method
        testcase in tests/importer/perform_import/test_importer_run.py
        Handle any exceptions and log nicer output based
        on the exception message content.
        :yield None:
        """
        try:
            yield
        except Exception as e:
            if self.return_code == NOERROR_RC:
                self.return_code = CATCHALL_RC
            # subprocess saves output in the CalledProcessError exception
            output = getattr(e, "output", "")
            log.critical('Aborted: {}'.format(e))
            if output:
                log.critical("Output:\n\n{}".format(output))
            if not self.dev:
                send_failure_notifications(output, importer=self)

    def run(self):
        """
        Perform the import
        :return int return_code: The return code of the import. Zero on
        success, nonzero on any type of failure. Check
        kamikaze3/importer/error_codes.py for a list of exit codes.

        """
        with self.flatten_exception():
            try:
                self.initiate_connections()
                self.prepare_import()
                if self.dry_run:
                    log.info("\nProfit: looks like we are ready to import!\n")
                else:
                    self.import_strategy.execute()
                    send_success_notifications(self)
            except KeyboardInterrupt:
                self.return_code = USER_ABORTED_RC
        return self.return_code
