import logging
from os import path
from pathlib import Path
from urllib.parse import urlparse

from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.shell import run_local_command, replace_in_file
from kamikaze3.importer.nginx import MAGENTO_2_NGINX_FLAG
from kamikaze3.importer.strategy.cms import ImportMagento

log = logging.getLogger(__name__)


class ImportMagento2(ImportMagento):
    local_database_config = '/data/web/magento2/app/etc/env.php'
    magerun = 'magerun2'

    @property
    def source_maintenance_path(self):
        return path.join(self.importer.remote_path, 'var/.maintenance.flag')

    @property
    def destination_maintenance_path(self):
        return '/data/web/magento2/var/.maintenance.flag'

    def fixup_local_database(self):
        replace_in_file(
            '/data/web/magento2/app/etc/config.php',
            self.importer.remote_db_info['host'].encode('utf-8'),
            b'mysqlmaster'
        )

    def import_filesystem(self):
        """
        Import the filesystem of the shop
        :return None:
        """
        log.info("Starting filesystem synchronization")
        self.importer.file_transfer_conn.driver.download(
            self.importer.remote_path, '/data/web/magento2'
        )
        super(ImportMagento2, self).import_filesystem()

    def sync_data(self, extra_mysqldump_arguments=None, cms_type=None):
        """
        Sync the shop data. The filesystem and the database in parallel.
        :param cms_type: cms type and version for skip table arguments
        :param str extra_mysqldump_arguments: A string of extra arguments.
        Is only added on the real sync, not the initial structure only sync
        :return None:
        """
        log.debug("Initiating data sync")
        super(ImportMagento2, self).sync_data(
            cms_type="magento2"
        )
        log.debug("Removing any old /data/web/public")
        run_local_command("rm -Rf /data/web/public", shell=True)
        log.debug("Symlink /data/web/magento2/pub to /data/web/public")
        run_local_command("ln -sf /data/web/magento2/pub /data/web/public", shell=True)

    @staticmethod
    def parse_url(select_query_result, force_https=False):
        """
        :param select_query_result:dict results from exec_query:
        :param list[dict, ..] select_query_result: The query result
        :param force_https:bool force the parser to use https
        :return string | None: Updated host for base_link url,
        or None in case we could not retrieve it from the select query
        """
        parsed_hypernode_name = urlparse(get_hypernode_name())
        if len(select_query_result):
            parsed_unsecure_url = urlparse(select_query_result[0]['value'])
            scheme = 'https' if force_https else str(parsed_unsecure_url.scheme)
            updated_unsecure_url = scheme + "://" + str(parsed_hypernode_name.path) + str(parsed_unsecure_url.path)
            return updated_unsecure_url

    def set_web_secure_base_url(self):
        """
        Updates the web/secure/base_url, this function will make all HTTPS links point to the new server
        """
        select_query = 'SELECT value from core_config_data where path = \"web/secure/base_url\"'
        secure_url_result = self.importer.local_db_conn.driver.exec_query(select_query)
        updated_unsecure_url = self.parse_url(secure_url_result, force_https=True)
        if updated_unsecure_url:
            update_query = "UPDATE core_config_data SET value = \"{}\" where path = \"web/secure/base_url\""
            self.importer.local_db_conn.driver.exec_query(update_query.format(updated_unsecure_url))
        else:
            self.importer.import_warnings.append(
                "Looks like there was no web/secure/base_url we could update, if you have any issues opening "
                "{} in your browser check out https://support.hypernode.com/knowledgebase/change-baseurl-magento2/"
                "".format(get_hypernode_name())
            )

    def exec_update_web_base_link_url_query(self, select_query, url_type):
        """
        find(select query), compose result for url and execute an UPDATE query.
        :param str select_query: A select query to find url value
        :param str url_type: A string with the url type in database like "web/secure/base_url".
        :return None:
        """
        query_result = self.importer.local_db_conn.driver.exec_query(select_query)
        updated_url = self.parse_url(query_result)
        if updated_url:
            url_update_query = "UPDATE core_config_data SET value = \"{}\" where path = \"{}\"".format(updated_url,
                                                                                                       url_type)
            self.importer.local_db_conn.driver.exec_query(url_update_query)
        else:
            self.importer.import_warnings.append(
                "Looks like there was no {} we could update, if you have any issues opening "
                "{} in your browser check out https://support.hypernode.com/knowledgebase/change-baseurl-magento2/"
                "".format(url_type, get_hypernode_name())
            )

    def set_web_base_link_url(self):
        """
        Updates the web/unsecure/base_link_url, this function will make all HTTPS links point to the new server
        """
        query_dict = {
            'unsecure_select_query': 'SELECT value from core_config_data where path = \"web/unsecure/base_link_url\"',
            'secure_select_query': 'SELECT value from core_config_data where path = \"web/secure/base_link_url\"'
        }
        self.exec_update_web_base_link_url_query(select_query=query_dict['unsecure_select_query'],
                                                 url_type="web/unsecure/base_link_url")
        self.exec_update_web_base_link_url_query(select_query=query_dict['secure_select_query'],
                                                 url_type="web/secure/base_link_url")

    def set_base_link_url(self):
        """
        Set the store base_link_url. Updates web/unsecure/base_link_url and web/secure/base_link_url.
        :return None:
        """
        log.info("Setting base_link_url for Magento 2")

        self.set_web_secure_base_url()
        self.set_web_base_link_url()

    def purge_redis_cache(self):
        """
        Use magerun2 to flush the magento 2 cache
        :return None:
        """
        log.info("Flushing the Magento cache")
        magento_flush_command = ['n98-magerun2', 'cache:flush', '--root-dir=/data/web/magento2']
        run_local_command(magento_flush_command)
        super(ImportMagento2, self).purge_redis_cache()

    @staticmethod
    def ensure_magento_2_nginx_flag():
        """
        Ensure that the Magento 2 nginx flag exists
        See; https://support.hypernode.com/knowledgebase/installing-magento-2-on-hypernode/#Activate_Magento_2_mode
        :return None:
        """
        Path(MAGENTO_2_NGINX_FLAG).touch(exist_ok=True)

    def update_settings(self):
        self.fix_up_local_database_config()
        self.save_last_known_base_url()
        if self.importer.set_default_url:
            self.set_base_link_url()
            self.set_default_url()
        self.purge_redis_cache()

    def execute(self):
        """
        Import a Magento 2 shop using the shop information
        stored on the passed importer object
        :return: None
        """
        log.info("Executing Magento 2 import strategy")
        super(ImportMagento2, self).execute()
        self.update_settings()
        self.ensure_magento_2_nginx_flag()
        self.print_success()
