import logging
from kamikaze3.importer.strategy.cms import ImportShopware
from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.shell import run_local_command

log = logging.getLogger(__name__)


class ImportShopware5(ImportShopware):
    local_database_config = '/data/web/public/config.php'

    @property
    def download_path(self):
        return '/data/web/public/'

    def set_base_url(self, base_url):
        """
        Set the store base url
        :param str base_url: The base url to set
        :return None:
        """
        log.debug("Setting the base url for shopware 5")
        core_config_data_table = self.importer.remote_db_info['table_prefix'] + 'core_shops'
        update_base_url_query = "UPDATE {} SET base_url = \"{}\"" \
                                "".format(core_config_data_table, base_url)
        self.importer.local_db_conn.driver.exec_query(update_base_url_query)

    def sync_data(self, extra_mysqldump_arguments=None, cms_type=None):
        """
        Sync the shop data. The filesystem and the database in parallel.
        :param cms_type: cms type and version for skip table arguments
        :param str extra_mysqldump_arguments: A string of extra arguments.
        Is only added on the real sync, not the initial structure only sync
        :return None:
        """
        super(ImportShopware5, self).sync_data(
            cms_type="shopware5"
        )

    def set_default_url(self):
        """
        Set the store default url. base_url
        :return None:
        """
        self.base_url = "http://{}/".format(get_hypernode_name())
        self.set_base_url(self.base_url)

    @staticmethod
    def remove_caches():
        """
        Removes Shopware caches in /var/cache and /web/cache
        :return: None:
        """
        run_local_command(["php", "/data/web/public/bin/console", "sw:cache:clear"])

    def set_host(self):
        """
        Set the store host in the s_core_shops table in shopware 5
        :return: None:
        """
        log.info("Updating host in database")
        core_config_data_table = self.importer.remote_db_info['table_prefix'] + 'core_shops'
        update_host_query = "UPDATE {} SET host = \"{}\"".format(core_config_data_table, get_hypernode_name())
        update_hosts_query = "UPDATE {} SET hosts = \"{}\"".format(core_config_data_table, get_hypernode_name())

        self.importer.local_db_conn.driver.exec_query(update_host_query)
        self.importer.local_db_conn.driver.exec_query(update_hosts_query)

    def save_last_known_base_url(self):
        """
        Store the last known base url from the remote shop on disk
        so we can later use this to find the shop to benchmark against
        in hypernode-control even if the customer supplied an SSH host
        that is not also the domain of the shop that is being imported.
        :return None:
        """
        log.info("Attempting to save the last known base url in /tmp/last_known_base_url")

        core_config_data_table = self.importer.remote_db_info['table_prefix'] + 'core_shops'
        last_known_query = "SELECT host FROM {}".format(core_config_data_table)
        host_query_tuple = self.importer.local_db_conn.driver.exec_query(last_known_query)
        last_known_host = host_query_tuple[0]['host']

        tmp_dir_command = "cd /tmp | touch last_known_base_url"
        try:
            if last_known_host:
                run_local_command(tmp_dir_command)
                with open('/tmp/last_known_base_url', 'w') as f:
                    f.write(last_known_host)
        except FileNotFoundError:
            log.info("Was not able to save last known base url, Thats ok.")

    def verify_set_maintenance(self):
        raise NotImplementedError("Managing maintenance is not yet implemented for Shopware 5, you can manually"
                                  " configure maintenance mode by going to your Shopware 5 backend, selecting"
                                  "Configuration --> Basic settings --> System --> Maintenance. After enabling "
                                  "maintenance mode clear your template caches for the changes to take effect. You can "
                                  "do this by going to Configuration --> Cache/Performance --> Cache")

    def clear_caches_on_source(self):
        pass

    def compose_query_set_maintenance(self):
        pass

    def compose_query_unset_maintenance(self):
        pass

    def update_configuration(self):
        """
        Update the imported shop settings to fit the local environment
        :return None:
        """
        self.save_last_known_base_url()
        self.fix_up_local_database_config()
        self.purge_redis_cache()
        if self.importer.set_default_url:
            self.set_host()
        self.remove_caches()

    def execute(self):
        """
        Import a Shopware 5 shop using the shop information
        stored on the passed importer object
        :return: None
        """
        log.info("Executing Shopware 5 import strategy")
        super(ImportShopware5, self).has_mysql_57()
        super(ImportShopware5, self).execute()
        super(ImportShopware5, self).update_settings()
