import socket
from contextlib import contextmanager
from copy import deepcopy
from os import environ
from re import match

SAFE_CHARACTER_REGEX = r'^[.a-zA-Z0-9\-_ =^/@%]+$'


@contextmanager
def catch_and_raise_known_errors(exceptions, catch_function=lambda e: None):
    """
    Try the body of the context and catch any specified errors and
    translate them to something more humanly readable before re-raising
    any not caught errors in the catch function.
    :yield None:
    """
    try:
        yield
    except exceptions as e:
        catch_function(e)
        raise


def compose_sshpass_env(password=None):
    """
    Compose the SSHPASS environment
    :param str password: The password to put in the environment dict if any
    :return dict env: The environment with SSHPASS added
    """
    environment = dict(deepcopy(environ))
    if password:
        environment["SSHPASS"] = password
    return environment


def get_free_port():
    """
    Find a random free port on the system and return it
    Tries to open a port, Linux will allocate a new free
    one and then close it immediately. The allocated port
    will be returned so it can be used later by the program
    with a high likeliness of it being free.
    :return int port: A random open port
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(('127.0.0.1', 0))
    _, random_port = sock.getsockname()
    sock.close()
    return random_port


def has_only_safe_chars(text):
    """
    Check if the provided string has only safe characters. We can't escape
    the table and db names for example, so we need to make sure nothing goofy
    is going on there with the user input.
    :param str text: The text to check for special chars
    :return bool safe: True if safe, False if not
    """
    return not text or bool(match(SAFE_CHARACTER_REGEX, text))
