"""
kamikaze3.importer.validators
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Validation functions for importer operations
"""
from re import search
from subprocess import check_output

MYSQL_VERSION_PATTERN = r'version\s*([\d.]+)'
PHP_VERSION_PATTERN = r'PHP\s*([\d.]+)'


def get_mysql_major_version():
    """
    Return the MySQL major version
    :return str mysql_version: Like '5.7' or '5.6'
    """
    mysql_version_command = ['mysql', '-e', 'SHOW VARIABLES WHERE Variable_name = "version"', '--batch']
    raw_mysql_version_output = check_output(
        mysql_version_command
    )
    mysql_version_output = raw_mysql_version_output.decode('utf-8')
    mysql_version_groups = search(MYSQL_VERSION_PATTERN, mysql_version_output)
    mysql_version = mysql_version_groups.group(1)
    mysql_major_version = '.'.join(mysql_version.split('.')[:2])
    return mysql_major_version


def check_mysql_major():
    if float(get_mysql_major_version()) < 5.7:
        return "Database was successfully imported, however... your MySQL installation is still at version {}\n" \
               "Shopware will likely not work with 5.6 or lower. You can upgrade yourself by\n" \
               "checking out https://support.hypernode.com/changelog/release-6583-upgrading-to-mysql-5-7/\n" \
               "and running \"hypernode-systemctl settings mysql_version --value 5.7\" to upgrade" \
               "".format(get_mysql_major_version())


def db_username_fits_in_hypernode(username):
    # MySQL 5.7 and MariaDB max username limit is 32 unicode characters
    # (even multibyte chars) for MySQL 5.6 this is 16
    max_chars = 16 if get_mysql_major_version() == "5.6" else 32
    return len(username) <= max_chars


def db_name_fits_in_hypernode(db_name):
    return len(db_name) <= 64


def get_php_major_version():
    """
    Return the PHP major version
    :return str php_major_version: Like '7.1' or '7.2'
    """
    php_version_command = ['php', '--version']
    raw_php_version_output = check_output(
        php_version_command
    )
    php_version_output = raw_php_version_output.decode('utf-8')
    php_version_groups = search(PHP_VERSION_PATTERN, php_version_output)
    php_version = php_version_groups.group(1)
    php_major_version = '.'.join(php_version.split('.')[:2])
    return php_major_version


def check_php_major():
    php_major_version = get_php_major_version()
    if float(php_major_version) < 7.2:
        return "Your installation was successfully imported, however... your PHP installation is still at" \
               " version {}\nShopware will likely not work with 7.1 or lower. You can upgrade yourself by\n" \
               "checking out https://support.hypernode.com/knowledgebase/hypernode-systemctl-cli-tool/\n" \
               "and running \"hypernode-systemctl settings php_version --value 7.2\" to upgrade" \
               "".format(php_major_version)
