import logging
import socket
from argparse import ArgumentParser

import requests

from kamikaze3.common.api import get_hypernode_api_token, HYPERNODE_API_URL
from kamikaze3.common.env import get_app_name
from kamikaze3.log.log import configure_logging

log = logging.getLogger(__name__)


HYPERNODE_API_FLOW_LIST_URL = HYPERNODE_API_URL + '/logbook/v1/logbooks/{}/flows/'


def parse_arguments():
    """
    Parse the hypernode-log arguments
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser(
        description="List pending and previous system "
                    "actions for this Hypernode"
    )
    parser.add_argument(
        '--verbose', '-v',
        help='Display debug output',
        action='store_true'
    )
    parser.add_argument('--lines', default=50, required=False, type=int,
                        help='Amount of tasks to display. Uses 50 lines by default')
    parsed_args = parser.parse_args()
    return parsed_args


def _try_get(url):
    response = requests.get(url, headers={
        'Authorization': 'Token {}'.format(get_hypernode_api_token()),
        'Content-Type': 'application/json',
        'Accept': 'application/json',
        'User-Agent': '{}:{}'.format(
            socket.getfqdn(), 'hypernode-kamikaze3'
        )
    })
    try:
        response.raise_for_status()
        return response.json()
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 404:
            print("No recent activity")
            exit(0)
        else:
            raise


def get_latest_flows(amount=50):
    flows = list()
    resp_handle = {'next': HYPERNODE_API_FLOW_LIST_URL.format(get_app_name())}
    while len(flows) < amount and resp_handle.get('next'):
        next_url = resp_handle['next']
        resp_handle = _try_get(next_url)
        flows.extend(resp_handle['results'])
    return flows[:amount]


def print_list_line(flow_name='ACTION', state='STATE', start='START', end='END',
                    progress='TASKS', running="RUNNING", finished=False):

    print("{:<30}\t{:<20}\t{:<20}\t{:<8}\t{:<4}\t{}".format(
        flow_name[:30],
        start or '',
        end or '',
        state or 'WAITING',
        progress or '',
        running or ("finished" if finished else "waiting")
    ))


def last_jobs(lines):
    """
    List the last flows run for this app. This is the same as ./bin/last_jobs in
    hn-control but then only for this specific app.
    :param int lines: The amount of flows to display
    :return None:
    """
    latest_flows = get_latest_flows(amount=lines)

    print_list_line()
    for f in latest_flows:
        failed = f['state'] in ('reverted',)
        all_tasks_complete = f['progress']['completed'] == f['progress']['total']
        finished = f['progress']['total'] > 0 and all_tasks_complete
        flow_name = f['name']
        state = f['state']
        if flow_name == 'create_backup' and state == 'reverted':
            state = 'background'
        print_list_line(
            flow_name=flow_name,
            state=state,
            start=f['created_at'],
            end=f['updated_at'],
            progress="{}/{}".format(
                f['progress']['completed'],
                f['progress']['total']
            ),
            running=' '.join(
                f['progress']['running']
            ),
            finished=finished or failed
        )


def main():
    args = parse_arguments()
    configure_logging('kamikaze3.log', args.verbose)
    last_jobs(lines=args.lines)
