import subprocess
import logging
from argparse import ArgumentParser
from contextlib import suppress


from kamikaze3.common.env import get_app_name, get_public_ip
from kamikaze3.common.slack import try_send_slack_message
from kamikaze3.log.log import configure_logging
from kamikaze3.manage_floating_ip.manage_ip_configurations import manage_route_to_digitalocean_api, \
    manage_floating_ip_loopback, ensure_idempotency, manage_mail_route_table, manage_mail_route_via_instance_gateway, \
    manage_ip_rule_for_mail_route, manage_mail_mangle_output_mark, manage_mail_snat_to_instance_ip, manage_arp_entry
from kamikaze3.manage_floating_ip.metadata import retrieve_instance_metadata, get_current_default_gateway, \
    floating_ip_active
from kamikaze3.utils import wait_for

logger = logging.getLogger(__name__)


def set_mail_configuration(enable):
    """
    Calls all relevant functions for setting outbound traffic over the given ports via the instance IP
    :param bool enable: Whether to enable or disable the outbound mail traffic via the instance IP
    :rtype: None
    """
    # DigitalOcean blocks outgoing traffic on port 25 and 587 (smtp ports). They only block
    # it on floating IPs, but not on instance IPs. This might cause people to experience mail
    # problems when xgrading from a non-floating IP Hypernode to a floating IP Hypernode. A
    # mail configuration that worked before the migration might not anymore after, and the
    # effects might not be immediately clear. It can take a while before people realize no
    # mail has been sent for a while. So to get these ports to work as before we make sure
    # that outbound traffic over these ports is sent through the instance IP instead of the
    # floating IP. To do so we create a new routing table, rules to send packets marked with
    # a certain value to the new routing table and use iptables to mark packets that are
    # outgoing on these ports with the special value. For more information on how this works see
    # https://unix.stackexchange.com/questions/21093/output-traffic-on-different-interfaces-based-on-destination-port/21118#21118
    mail_configuration_funcs = (
        manage_mail_route_table, manage_mail_route_via_instance_gateway,
        manage_ip_rule_for_mail_route, manage_mail_mangle_output_mark,
        manage_mail_snat_to_instance_ip
    )
    for func in mail_configuration_funcs if enable else reversed(mail_configuration_funcs):
        ensure_idempotency(func, enable=enable)


def change_gateway(new_gateway):
    """
    Replace the old gateway with the new gateway.
    This is done in a single command, because if separate
    these into two commands, the node might become unreachable
    after the first command.
    :param str new_gateway: The gateway to change it to
    :return: None
    """
    current_gateway = get_current_default_gateway()

    logger.info("Changing gateway to {}".format(new_gateway))
    command = ''
    add_gateway = 'ip route add default via {}'.format(new_gateway)
    if current_gateway:
        del_gateway = 'ip route del default via {}'.format(current_gateway)
        command += del_gateway + '; '
    command += add_gateway

    # Will throw an exception when the second command (ip route add) returns nonzero.
    # So if the del fails, that is silently ignored. The state can then be that at
    # the moment of the add the gateway either exists or not, if it does exist because
    # the del failed then it will fail adding the new one (because one already exists),
    # and if it does not exist yet and the add fails it will except because of that.
    subprocess.check_call(command, shell=True)


def verify_outbound_ip_is_expected(expected_ip):
    """
    Check if the public IP is the expected IP. Checks a couple of
    times over max 5 seconds because the route change might not
    take effect immediately. Don't want to give up too soon.
    :param str expected_ip: The expected public IP
    :return str actual_ip: The expected IP if it matches.
    """
    logger.info("Verifying if networking is working as intended")
    return wait_for(
        get_public_ip,
        value=expected_ip,
        check_interval=1,
        timeout=10
    )


def manage_floating_ip(enable, expected_ip=None, send_slack_message=True):
    """
    Attempt to retrieve the instance metadata required to either enable or disable a floating IP (by changing
    the default gateway) and if successful perform the change. If not all metadata could be retrieved nothing
    will be done. This makes it so that only at DigitalOcean (where solely required) some action is performed.
    :param bool enable: Enable or disable the floating IP
    :param str expected_ip: the expected IP if any. we use this to verify the outbound IP against. If it is
    not specified we only check if we can do outbound requests but don't assert if the IP is something specific.
    :param bool send_slack_message: Send a slack message on failure or not
    :return None:
    """
    public_ip = None

    # Cache on enable, use cache on disable. If because of some inconsistent state
    # the route to the instance metadata api becomes unavailable when we want to
    # detach a floating IP and route the traffic back through the instance IP
    # we should not risk getting hung up on the metadata API connection if we
    # have the cached data on disk anyway.
    instance_metadata = retrieve_instance_metadata(use_cache=not enable)

    if all(instance_metadata.values()):
        logger.info("Instance requires manual floating IP management. Changing default gateway now..")
        old_gateway = instance_metadata['old_gateway']
        new_gateway = instance_metadata['anchor_gateway'] if enable else instance_metadata['instance_gateway']
        try:
            ensure_idempotency(manage_floating_ip_loopback, enable=enable,
                               ip_address=expected_ip or instance_metadata['ip_address'])
            ensure_idempotency(manage_arp_entry, enable=enable)
            ensure_idempotency(manage_route_to_digitalocean_api, enable=enable,
                               instance_gateway=instance_metadata['instance_gateway'])
            change_gateway(new_gateway=new_gateway)
            set_mail_configuration(enable=enable)
            verify_outbound_ip_is_expected(expected_ip)
        except Exception as e:
            message = "Networking validation checks failed after changing default " \
                      "gateway to {}, setting default gateway back to {}" \
                      "".format(new_gateway, old_gateway)

            with suppress(Exception):
                # Failing to get the public IP should not crash the revert under any
                # circumstances.
                public_ip = get_public_ip()

            if expected_ip:
                message += '. Expected public IP {}, but it was {}'.format(expected_ip, public_ip or 'unknown')
            logger.info(message)
            logger.exception(e)
            ensure_idempotency(manage_arp_entry, enable=not enable)
            change_gateway(new_gateway=old_gateway)
            if send_slack_message:
                try_send_slack_message(':bug: {} on {}'.format(message, get_app_name()))
            ensure_idempotency(manage_route_to_digitalocean_api, enable=not enable,
                               instance_gateway=instance_metadata['instance_gateway'])
            ensure_idempotency(manage_floating_ip_loopback, enable=not enable,
                               ip_address=instance_metadata['ip_address'])
            set_mail_configuration(enable=not enable)
            logger.info("Raising original exception now")
            raise e
    else:
        logger.info("Instance does not require manual floating IP management, skipping..")


def parse_arguments(args=None):
    """
    Parse the hypernode-manage-floating-ip arguments
    :param list args: Optional, defaults to sys.argv
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--enable',
        action='store_true',
        help='Enable the use of floating IPs for Digital Ocean'
    )
    group.add_argument(
        '--disable',
        action='store_true',
        help='Disable the use of floating IPs for Digital Ocean'
    )
    parser.add_argument(
        '--verbose', '-v',
        help='Display debug output',
        action='store_true'
    )
    parser.add_argument(
        '--expected-ip',
        help='The expected IP (optional). Will revert if the public IP '
             'is not this IP after performing the networking changes.'
    )
    parser.add_argument(
        '--dev',
        help='Use this flag when testing. Disables slack alerts.',
        action='store_true'
    )
    parsed_args = parser.parse_args(args)
    return parsed_args


def main():
    """
    The entrypoint for this command
    :return None:
    """
    args = parse_arguments()
    configure_logging('kamikaze3.manage_floating_ip', args.verbose)
    if args.enable and not floating_ip_active():
        logger.info(
            "There is no floating IP attached to this instance. "
            "You can not run --enable until one is attached."
        )
        exit(1)
    # If args.enable is not True, then --disable is set
    manage_floating_ip(
        enable=args.enable,
        expected_ip=args.expected_ip,
        send_slack_message=not args.dev
    )


if __name__ == '__main__':
    main()
