import json
import logging
import re
from contextlib import suppress
from http import HTTPStatus

import requests
from os.path import isfile

import subprocess

CLOUD_PLATFORM_API = 'http://169.254.169.254/metadata/v1/'
INSTANCE_GATEWAY = 'interfaces/public/0/ipv4/gateway'
INSTANCE_ADDRESS = 'interfaces/public/0/ipv4/address'
ANCHOR_GATEWAY = 'interfaces/public/0/anchor_ipv4/gateway'
FLOATING_IP_ACTIVE = 'floating_ip/ipv4/active'
FLOATING_IP_ADDRESS = 'floating_ip/ipv4/ip_address'
INSTANCE_METADATA_CACHE = '/etc/hypernode/instance_metadata'
GATEWAY_L3_MAC = "fe:00:00:00:01:01"  # Static MAC defined by DO.

logger = logging.getLogger(__name__)


def attempt_get_metadata(url):
    """
    Attempt getting the specified metadata from the cloud provider platform API.
    Returns None in case of error.
    :param str url: The metadata location to retrieve
    :return str metadata_value || nonetype None: The metadata on success, or None
    """
    logger.info("Attempting to get instance metadata from the API")
    with suppress(Exception):
        response = requests.get(CLOUD_PLATFORM_API + url)
        if response.status_code == HTTPStatus.OK and response.text.strip():
            return response.text.strip()


def retrieve_instance_metadata(use_cache=True):
    """
    Get all required instance metadata from the system and the cloud platform metadata API
    :param bool use_cache: Use cached instance metadata if available. Can be handy if
    for some reason the metadata API is not available in the current state of the system
    anymore.
    :return dict metadata: All relevant instance metadata
    """
    logger.info("Retrieving instance metadata from the cloud platform metadata API")

    if use_cache and isfile(INSTANCE_METADATA_CACHE):
        # Broad exception clause because if anything goes wrong we need to attempt to fall back on the actual API.
        # Only if the routes are misconfigured that API would be unavailable. If both the cache is for some reason
        # not available and the API can't be reached as well (because of an incorrect routing state) this
        # function will hang and block until a socket timeout occurs.
        with suppress(Exception):
            with open(INSTANCE_METADATA_CACHE, 'r') as f:
                instance_metadata = json.load(f)
                logger.info("Using instance metadata cached on the filesystem")
                return instance_metadata

    instance_metadata = {
        'old_gateway': get_current_default_gateway(),
        'anchor_gateway': attempt_get_metadata(ANCHOR_GATEWAY),
        'instance_gateway': attempt_get_metadata(INSTANCE_GATEWAY),
        'ip_address': attempt_get_metadata(FLOATING_IP_ADDRESS)
    }
    logger.info("Caching instance metadata from the API on the filesystem")
    with open(INSTANCE_METADATA_CACHE, 'w') as f:
        json.dump(instance_metadata, f)

    logger.info("Using instance metadata retrieved from the metadata API")
    return instance_metadata


def floating_ip_active():
    """
    Checks the DigitalOcean API whether the node has an active floating IP.
    :rtype: bool
    """
    logger.info('Checking if instance metadata API reports an attached floating IP')
    response = requests.get(CLOUD_PLATFORM_API + FLOATING_IP_ACTIVE)
    if response.status_code == HTTPStatus.OK and response.text.strip() == 'true':
        return True
    return False


def get_current_default_gateway():
    """
    Get the default gateway from 'ip route'
    :return str default_gateway || nonetype None: The default gateway if any
    """
    logger.info("Getting current gateway")
    command = ['ip', 'route']
    iproute_proc = subprocess.check_output(command)
    match = re.search(r'default via (\d+\.\d+\.\d+\.\d+)', iproute_proc.decode())
    if match:
        return match.group(1)


