"""
kamikaze3.metrics.imageoptimizer-cron
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Detect whether or not a cron job for the hypernode-image-optimizer is installed
"""

import sys
import subprocess
from os import path, EX_TEMPFAIL, EX_OK, EX_SOFTWARE
from argparse import ArgumentParser
from logging import getLogger, DEBUG, INFO
from kamikaze3.common.log import create_syslog_handler
from kamikaze3.common.csv import print_dicts_as_csv

scriptname = "hypernode-image-optimizer"
crontab_command = "/usr/bin/crontab"
logger = getLogger(__name__)


def parse_args(args):
    parser = ArgumentParser(prog=scriptname,
                            description='Check if image optimizer cron is running')
    parser.add_argument('--verbose', action='store_true', help='print verbose messages to stderr')
    return parser.parse_args(args)


def get_crontab_command():
    if path.exists(crontab_command):
        return crontab_command
    raise RuntimeError("Crontab failed, could not find binary")


def read_crontab():
    """
    Fetch the contents of the cron file for the 'app' user and return as string
    """
    proc = subprocess.Popen(
        ['sudo', '-u', 'app', get_crontab_command(), '-l'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    stdout, stderr = proc.communicate()
    if proc.returncode != 0 and not stderr.decode('utf-8').startswith('no crontab for '):
        raise RuntimeError("Crontab failed ({}): {}".format(proc.returncode, stderr))

    return stdout.decode('utf-8').split('\n')


def cronjob_exists():
    """
    Detect if a cronjob for scriptname exists and is not commented out
    :return: 1 if the cronjob exists, 0 otherwise (telegraf uses these as bools)
    :rtype: int
    """
    for line in read_crontab():
        if scriptname in line:
            if not line.strip().startswith('#'):
                return 1
    return 0


def gather_imageoptimizer_metrics():
    return {scriptname: cronjob_exists()}


def main(args=None):
    """Program entry point, check if a cron is installed for the hypernode-image-optimizer script

    :return: the exit code
    :rtype: int
    """

    opts = parse_args(args)

    logger.setLevel(DEBUG if opts.verbose else INFO)
    logger.addHandler(
        create_syslog_handler(
            name_override=scriptname if __name__ == '__main__' else None
        )
    )

    try:
        print_dicts_as_csv([gather_imageoptimizer_metrics()])
        return EX_OK
    except KeyboardInterrupt:
        logger.warning('process interrupted by user')
        return EX_TEMPFAIL
    except Exception as err:
        logger.exception(err)
        return EX_SOFTWARE


if __name__ == '__main__':
    sys.exit(main())
