"""
kamikaze3.metrics.long_processes
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Provide metrics about long-running processes
"""

import sys
from os import EX_TEMPFAIL, EX_OK, EX_SOFTWARE
from logging import getLogger, DEBUG, INFO
from argparse import ArgumentParser
from kamikaze3.common.log import create_syslog_handler
from kamikaze3.common.csv import print_dicts_as_csv
from kamikaze3.longest_process_time.main import get_longest_process_time
import MySQLdb
from MySQLdb.cursors import DictCursor
import configparser
import psutil
from time import time


program_name = 'hypernode-long-running-procs'
logger = getLogger(__name__)


def parse_args(args):
    parser = ArgumentParser(prog=program_name,
                            description='Provide metrics about long-running processes')
    parser.add_argument('--verbose', action='store_true', help='print verbose messages to stderr')
    return parser.parse_args(args)



def sql():

    try:
        config = configparser.ConfigParser()
        config.read('/etc/telegraf/telegraf.d/mysqlcred.cfg')
        dbcred = (
            config['Main']['mysql_server'],
            config['Main']['mysql_user'],
            config['Main']['mysql_pass']
        )

        cur = MySQLdb.connect(*dbcred).cursor(DictCursor)
        cur.execute('show processlist')
        times = [x['Time'] for x in cur if x['Command'] == 'Query']
        return int(max(times)) if times else 0
    except Exception as e:
        logger.exception(e)

    return None


def filter_procs(proc):

    USER = 'app'
    REQUIRED_PROC_NAMES = ('php',)

    if proc.username() != USER:
        return False
    if proc.name() not in REQUIRED_PROC_NAMES:
        return False

    return True


def shell():

    try:
        durations = [
            time() - p.create_time()
            for p in psutil.process_iter() if filter_procs(p)
        ]
        return int(max(durations)) if durations else 0
    except Exception as e:
        logger.exception(e)

    return None


def fpm():
    try:
        longest = get_longest_process_time()
        return int(longest)
    except Exception as e:
        logger.exception(e)

    return None


def gather_long_processes_metrics():
    return dict(
        sql=sql(),
        shell=shell(),
        fpm=fpm(),
    )


def main(args=None):
    """Program entry point, create a report of longest running processes:
    prints cvs of the duration of the longest-running MySQL query, shell PHP process, and fpm process, in seconds.

    :return: the exit code
    :rtype: int
    """

    opts = parse_args(args)

    logger.setLevel(DEBUG if opts.verbose else INFO)
    logger.addHandler(
        create_syslog_handler(
            name_override=program_name if __name__ == '__main__' else None
        )
    )

    try:
        result = gather_long_processes_metrics()
        print_dicts_as_csv([result])
        return EX_OK
    except KeyboardInterrupt:
        logger.warning('process interrupted by user')
        return EX_TEMPFAIL
    except Exception as err:
        logger.exception(err)
        return EX_SOFTWARE


if __name__ == '__main__':
    sys.exit(main())
