"""
kamikaze3.monitoring.check_db
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
monitor DB consistency and notify client
"""
import sys
import os
import re
import traceback
from subprocess import run, PIPE, SubprocessError
from argparse import ArgumentParser
from logging import getLogger, DEBUG, INFO
from kamikaze3.common.log import create_syslog_handler, create_stream_handler
from kamikaze3.monitoring import NAGIOS_PLUGINS_PATH, PLUGIN_CODE_OK
from kamikaze3.onboarding import try_call_event_hook


NAGIOS_DB_PLUGIN = os.path.join(NAGIOS_PLUGINS_PATH, 'check_db')
logger = getLogger(__name__)


IGNORED_DB_ISSUES_PATTERNS = (
    r"CRITICAL.+\(2003, .+Can't connect to MySQL server on.+\(111\)",
)


def parse_args(args):
    parser = ArgumentParser(description="monitor DB consistency and notify client")
    mut_exclusive_opts = parser.add_mutually_exclusive_group()
    mut_exclusive_opts.add_argument('-v', '--verbose', action='store_true', help='run in verbose mode')
    mut_exclusive_opts.add_argument('-q', '--quiet', action='store_true', help='run in quiet mode')
    parser.add_argument('--no-notify', action='store_true', help='do not notify client')
    return parser.parse_args(args)


def get_db_consistency_issue():
    """Returns a message regarding DB consistency issues, or None
    if DB is OK
    :return: str|None
    """
    check_db_cmd = [NAGIOS_DB_PLUGIN]
    logger.debug('running DB consistency check: {}'.format(' '.join(check_db_cmd)))
    check_proc = run(check_db_cmd, stdout=PIPE, stderr=PIPE)
    logger.debug('DB consistency check code: {}'.format(check_proc.returncode))
    if check_proc.returncode != PLUGIN_CODE_OK:
        db_issue = check_proc.stderr.decode('utf-8', errors='replace').strip()
        for pattern in IGNORED_DB_ISSUES_PATTERNS:
            if re.match(pattern, db_issue):
                return None
        return db_issue


def notify_client_of_db_issue(issue):
    """Send an event to event API to notify client of the issue

    :param str issue: the DB issue description
    :return: None
    """
    event = {'monitoring-db-consistency': issue}
    logger.debug('sending DB consistency issue event to control ...')
    try_call_event_hook(event)
    logger.info('DB consistency issue event sent!')


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    opts = parse_args(args)
    logger.setLevel(DEBUG)
    if not opts.quiet:
        log_level = DEBUG if opts.verbose else INFO
        logger.addHandler(create_stream_handler(level=log_level))
    logger.addHandler(
        create_syslog_handler(level=DEBUG,
                              name_override='kamikaze3.monitoring.check_db' if __name__ == '__main__' else None))
    try:
        db_issue = get_db_consistency_issue()
        if db_issue:
            logger.warning('DB has consistency issue: "{}"'.format(db_issue))
            if not opts.no_notify:
                notify_client_of_db_issue(db_issue)
        else:
            logger.info('DB consistency is OK')
    except KeyboardInterrupt:
        logger.warning('process interrupted!')
        return os.EX_TEMPFAIL
    except (RuntimeError, OSError, SubprocessError) as err:
        logger.error('failed to check DB: {}'.format(err))
        logger.debug(traceback.format_exc())
        return os.EX_SOFTWARE
    return os.EX_OK


if __name__ == '__main__':
    sys.exit(main())
