import json
from multiprocessing import Lock
from os import fsync, rename
from kamikaze3.mysqlbackup.logger import get_logger
from kamikaze3.mysqlbackup.mysql import fetch_list_of_databases
from kamikaze3.mysqlbackup.settings import BACKUP_MONITOR_FILE
from datetime import datetime
from random import randint

log = get_logger(__name__)


class BackupState(object):
    def __init__(self):
        self._read_state_from_file()
        self.lock = Lock()

    def backup_succeeded(self, database):
        """
        Saves current timestamp for successful backup
        """
        self.lock.acquire()
        self._read_state_from_file()
        self.state[database] = {'last_backup': datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%f')}
        self._write_state_to_file()
        self.lock.release()

    def remove_obsolete_entries(self):
        try:
            self._read_state_from_file()
            self._remove_obsolete_entries()
        except Exception as e:
            log.warning("Removing obsolete entries from BackupState failed: '{}'".format(e))

        self._write_state_to_file()

    def get_state(self):
        return self.state

    def _remove_obsolete_entries(self):
        """
        Cleans up the backup state, removing data about databases that do not exist anymore.
        """
        databases_on_server = set(fetch_list_of_databases())  # Does an actual call to mysql!
        databases_in_backup_state = set(self.state.keys())
        non_existing_dbs = databases_in_backup_state - databases_on_server

        for db in non_existing_dbs:
            del(self.state[db])

    def _read_state_from_file(self):
        """
        Reads backup state from file and stores in self.state
        """
        try:
            with open(BACKUP_MONITOR_FILE, 'r') as f:
                data = f.read()
                self.state = json.loads(data)
        except IOError:
            self.state = {}  # No state file (yet)

    def _write_state_to_file(self):
        """
        Writes backup_state dict atomically to state file
        """
        data = json.dumps(self.state)

        # Atomic write.
        # Write data to temp file and move it to real target file
        tmp_file = BACKUP_MONITOR_FILE + "-new" + str(randint(1000, 9999))

        fh = open(tmp_file, 'w')
        fh.write(data)

        fh.flush()             # make sure that all data is on disk
        fsync(fh.fileno())     # see http://stackoverflow.com/questions/7433057/is-rename-without-fsync-safe
        fh.close()

        rename(tmp_file, BACKUP_MONITOR_FILE)
