import json
from pwd import getpwnam

from kamikaze3.mysqlbackuprestore.logger import get_logger
from kamikaze3.mysqlbackuprestore.printer import slack_restore_db_datetimes
from kamikaze3.mysqlbackuprestore.restore_helpers import start_decompressor, start_mysql_importer, \
    stream_backup_from_s3, assert_database_does_not_exist, ensure_database, filter_databases, find_closest_backups
from kamikaze3.mysqlbackuprestore.settings import RESTORE_PATH, FULL_RESTORE_STATE_FILE
from kamikaze3.mysqlbackuprotator.s3_commands import list_database_directories, get_bucket_and_config
import os
from kamikaze3.mysqlbackuprotator.settings import S3_MYSQL_BACKUP_FOLDER

log = get_logger("hypernode.mysql.backuprestore")


def download_or_import_databases(arguments):
    """
    This function either downloads or imports a single database.
    Alternatively it can also import all databases from backup into the local mysql

    :param arguments: the arguments dict from argparse
    """
    bucket, config = get_bucket_and_config()
    databases = list_database_directories(bucket, **config)

    if arguments.latest:
        backups_to_be_restored = find_closest_backups(bucket, databases, config['appname'],
                                                      max_delta=arguments.max_backup_delta)
    else:  # select backups nearest to specific date/time
        backups_to_be_restored = find_closest_backups(bucket, databases, config['appname'], target_date=arguments.date,
                                                      max_delta=arguments.max_backup_delta)

    if arguments.database is not None:  # download/import a specific database
        backups_to_be_restored = filter_databases([arguments.database], backups_to_be_restored)

        if arguments.download:
            check_enough_diskspace_available_to_download_backups(backups_to_be_restored, bucket, config)
        for backup in backups_to_be_restored:
            if arguments.download:
                download_database(bucket, backup, **config)
            else:
                import_backup(bucket, backup, overwrite=arguments.force, destination_database=arguments.import_as,
                              **config)

    if arguments.exclude_database:  # remove specific databases which have been excluded
        backups_to_be_restored = filter_databases(arguments.exclude_database, backups_to_be_restored, exclude=True)

    if arguments.restore_all_databases:
        restore_all_databases(backups_to_be_restored, bucket, config, arguments.force)

    if arguments.download_all_databases:
        check_enough_diskspace_available_to_download_backups(backups_to_be_restored, bucket, config)
        for backup in backups_to_be_restored:
            download_database(bucket, backup, **config)


def import_database(bucket, backup, destination_database=None, **kwargs):
    """
    Import a database straight to mysql. The target database is :destination_database:

    :param bucket: Boto bucket object
    :param MysqlBackup backup: the backup which will be imported
    :param str destination_database: the database name which we're restoring the backup towards
    :param dict kwargs: arguments which get passed on to sub functions. i.e. the output from parse_json_file
    """
    destination_database = destination_database or backup.database

    log.info("importing database backup {} into database {}".format(backup.filename, destination_database))
    decompressor_process = start_decompressor()
    mysqlimport_process = start_mysql_importer(decompressor_process.stdout, destination_database)
    stream_backup_from_s3(decompressor_process.stdin, bucket, backup, appname=kwargs['appname'])

    decompressor_process.stdin.close()
    decompressor_exit_code = decompressor_process.wait()

    if decompressor_exit_code != 0:
        raise RuntimeError("Decompressor exited with non-zero exit code!")
    mysqlimport_stderr = mysqlimport_process.communicate()[1]
    mysqlimport_exit_code = mysqlimport_process.poll()
    if mysqlimport_exit_code != 0:
        raise RuntimeError("Importing database failed! Mysql exited with non-zero exit code! {}".format(mysqlimport_stderr))

    decompressor_process.stdout.close()
    mysqlimport_process.stdout.close()


def download_database(bucket, backup, **kwargs):
    """
    Download a mysql backup straight to disk

    :param bucket: Boto bucket object
    :param str database: database name from which we're fetching a backup
    :param MysqlBackup backup: file name of the backup to be fetched
    :param dict kwargs: arguments which get passed on to sub functions. i.e. the output from parse_json_file
    """
    log.info("Downloading database backup {} to {}".format(backup.filename, RESTORE_PATH))
    target_file = os.path.join(RESTORE_PATH, backup.filename)
    with open(target_file, "w+b") as file:
        stream_backup_from_s3(file, bucket, backup, appname=kwargs['appname'])
    app_id = getpwnam("app")
    os.chown(target_file, app_id.pw_uid, app_id.pw_gid)


def import_backup(bucket, backup, overwrite=False, destination_database=None, **kwargs):
    """
    Fetch a backup and directly import it to mysql as :destination_database:

    :param bucket: Boto bucket object
    :param MysqlBackup backup: file name of the backup to be fetched
    :param bool overwrite: whether or not over write existing database
    :param str destination_database: the database name which we're restoring the backup towards
    :param dict kwargs: arguments which get passed on to sub functions. i.e. the output from parse_json_file
    """
    if destination_database is None:
        destination_database = backup.database

    if not overwrite:
        assert_database_does_not_exist(destination_database)

    ensure_database(destination_database)
    import_database(bucket, backup, destination_database=destination_database, **kwargs)


def list_databases_in_latest_backup():
    """
    :return: list(str) containing all database names existing in the backups
    """
    bucket, config = get_bucket_and_config()
    databases = list_database_directories(bucket, **config)
    return databases


def restore_all_databases(backups_to_be_restored, bucket, config, force):
    """
    Import all database backups directly to MySQL

    :param list(MysqlBackup) backups_to_be_restored:
                a list of specific backups which will be imported
    :param bucket: Boto bucket object
    :param dict config: the output from parse_json_file
    :param bool force: force overwriting of existing backups
    :return:
    """

    # We are restoring all databases. The "mysql" database must be restored first
    # since other databases may depend on a "definer" user
    for backup in backups_to_be_restored:
        if backup.database == "mysql":
            import_backup(bucket, backup, overwrite=force, **config)

    for backup in backups_to_be_restored:
        if backup.database == "mysql":
            continue
        import_backup(bucket, backup, overwrite=force, **config)

    slack_restore_db_datetimes(backups_to_be_restored)
    store_restore_db_datetimes_on_disk(backups_to_be_restored)


def store_restore_db_datetimes_on_disk(restored_backups):
    """
    Store restore databases + datetimes on disk

    :param list(MysqlBackup) restored_backups:
                a list of backups which were restored
    """

    data = {}
    for b in restored_backups:
        data[b.database] = b.backup_date.isoformat(' ')  # UTC

    with open(FULL_RESTORE_STATE_FILE, "w") as f:
        f.write(json.dumps(data))


def compute_total_size_of_backup_archives(backups, bucket, config):
    """
    Calculate total size of all .sql.gz archives inside :backups:

    :param list(MysqlBackup) backups:
                a list of specific backups for which we compute the total size
    :param bucket: Boto bucket object
    :param dict config: the output from parse_json_file
    :return int: required total size for all backups
    """
    result = 0  # bytes

    for backup in backups:
        key = bucket.get_key("{}{}/{}".format(S3_MYSQL_BACKUP_FOLDER.format(appname=config["appname"]),
                                              backup.database, backup.filename))
        result += key.size

    return result


def check_enough_diskspace_available_to_download_backups(backups, bucket, config):
    """
    Check if enough disk space is avaiable. Raises RuntimeError if not.

    :param list(MysqlBackup) backups: a list of the selected backups which will be downloaded
    :param bucket: Boto bucket object
    :param dict config: the output from parse_json_file
    :raises RuntimeError: if not enough diskspace is available
    """

    size_needed = compute_total_size_of_backup_archives(backups, bucket, config)  # bytes
    st = os.statvfs(RESTORE_PATH)
    free_size = st.f_bavail * st.f_frsize  # result is in bytes

    if size_needed > free_size:
        # convert size needed to MB
        size_needed = int(size_needed / 1024 / 1024)
        free_size = int(free_size / 1024 / 1024)
        raise RuntimeError("Not enough disk space available. Needed: {} Available: {}".format(size_needed, free_size))
