from copy import copy
from datetime import timedelta, datetime
from kamikaze3.mysqlbackuprotator.settings import BACKUP_RETENTION, MAX_MATCH_DIFFERENCE

NOW = datetime.utcnow()


def calculate_hourly(retention=BACKUP_RETENTION, from_datetime=NOW):
    """
    Given a :retention: object, a list of datetime objects is calculated and returned for
    the hourly backups. The term 'hourly' may be misleading, because they can be n-hourly backups
    (e.g. every three hours for SLA customers)
    """
    amount = retention.hourly_amount
    backup_interval = retention.hourly_interval

    return [from_datetime - timedelta(hours=h * backup_interval) for h in range(0, amount)]


def calculate_daily(retention=BACKUP_RETENTION, from_datetime=NOW):
    """
    Given a :retention: object, a list of datetime objects is calculated and returned for
    the daily backups.
    """
    offset_hours = retention.get_daily_offset()
    amount = retention.daily_amount
    hour = retention.DAILY_PREFERED_HOUR
    minute = retention.DAILY_PREFERED_MINS

    start_datetime = from_datetime - timedelta(hours=offset_hours)
    start_datetime = start_datetime.replace(hour=hour, minute=minute, second=0, microsecond=0)

    return [start_datetime - timedelta(days=d) for d in range(0, amount)]


def calculate_weekly(retention=BACKUP_RETENTION, from_datetime=NOW):
    """
    Given a :retention: object, a list of datetime objects is calculated and returned for
    the weekly backups.
    """
    offset_hours = retention.get_weekly_offset()
    amount = retention.weekly_amount
    hour = retention.DAILY_PREFERED_HOUR
    minute = retention.DAILY_PREFERED_MINS
    weekday = retention.WEEKLY_BACKUP_PREFERRED_WEEKDAY

    start_datetime = from_datetime - timedelta(hours=offset_hours)
    start_datetime = start_datetime.replace(hour=hour, minute=minute, second=0, microsecond=0)

    # if calculated day is not the expected weekday. pick the previous date which matches the weekday
    if start_datetime.weekday() != weekday:
        start_weekday = start_datetime.weekday()
        difference = start_weekday - weekday
        start_datetime = start_datetime - timedelta(days=difference)

    return [start_datetime - timedelta(weeks=w) for w in range(0, amount)]


def generate_keep_dates(retention=BACKUP_RETENTION, from_datetime=NOW):
    """
    Generates the keep dates for a given :Retention: objects

    Calls the previously defined calculator functions and returns
    concatenation of all output. It also shifts the datetimes 24
    hours into the future, which ensures that the newest backup is always kept!

    returns:
        a generator of datetime objects
    """
    keepdates = calculate_hourly(from_datetime=from_datetime, retention=retention) \
        + calculate_daily(from_datetime=from_datetime, retention=retention)\
        + calculate_weekly(from_datetime=from_datetime, retention=retention)

    for date_time in keepdates:  # shift the dates into the future, so we never delete the latest backup!
        yield date_time + timedelta(days=1)


def find_closest(target_date, available_backups, max_delta=None):
    """
    Finds the closest matching backup

    :param datetime target_date: target datetime. non timezone aware object in UTC
    :param list(MysqlBackup) available_backups: a list of available backups
    :param int max_delta: maximum difference in days between :target_date: and the closest available backup

    returns:
        closest backup MysqlBackup match
        or None if no match is found
    :rtype MysqlBackup|None:
    """
    difference = None
    result = None
    max_delta = max_delta if max_delta else MAX_MATCH_DIFFERENCE

    for b in available_backups:
        if result is None:
            if abs(b.backup_date - target_date) < timedelta(days=max_delta):
                result = b
                difference = abs(b.backup_date - target_date)
        elif abs(b.backup_date - target_date) < difference and \
                abs(b.backup_date - target_date) < timedelta(days=max_delta):
            result = b
            difference = abs(b.backup_date - target_date)

    return result


def find_matches(keep_dates, backup_list):
    """
    Matches backup files to dates. Uses find_closest() to find matches

    :param list(datetime) keep_dates: list of datetimes to keep backups for
    :param list(MysqlBackup) backup_list: a list of available backups

    returns:
        list of tuples (datetime, filename)
    """
    backup_list = copy(backup_list)  # To prevent removing items in outer scope

    matches = []
    for k in keep_dates:
        closest = find_closest(k, backup_list)
        if closest is not None:
            matches.append(closest)
            backup_list.remove(closest)  # pop it
            if not backup_list:
                # no more left
                break

    return matches


def backups_to_remove(keep_dates, backup_list):
    """
    Returns a list of backups that are superfluous after matching and can thus be deleted.

    :param list(datetime) keep_dates: list of datetimes to keep backups for
    :param list(MysqlBackup) backup_list: a list of available backups
    """
    matches = find_matches(keep_dates, backup_list)
    return [b for b in backup_list if b not in matches]


def backups_to_keep(keep_dates, backup_list):
    """
    Returns a list of backups that are should be kept.

    :param list(datetime) keep_dates: list of datetimes to keep backups for
    :param list(MysqlBackup) backup_list: a list of available backups
    """
    return find_matches(keep_dates, backup_list)
