from kamikaze3.mysqlbackuprotator.helpers import parse_json_file
from kamikaze3.mysqlbackuprotator.logger import get_logger
from kamikaze3.mysqlbackuprotator.settings import DB_NAME_REGEX, S3_MYSQL_BACKUP_FOLDER, DB_BACKUP_REGEX, VARS_FILE
from re import match
from boto.s3.connection import S3Connection

log = get_logger(__name__)


def list_database_directories(bucket, **kwargs):
    """
    Returns a list of all databases/directories which are currently on s3
    Uses the :DB_NAME_REGEX: to skip any database that does not match expected format

    :param bucket: a boto s3 bucket object

    :rtype [string]:
    :return a list of strings containing database names
    """
    get_db_regex = ".*/({})/".format(DB_NAME_REGEX)
    bucket_result = bucket.list(S3_MYSQL_BACKUP_FOLDER.format(appname=kwargs["appname"]), "/")
    databases = [match(get_db_regex, res.name).group(1) for res in bucket_result]

    result = []
    for db in databases:
        if match(DB_NAME_REGEX, db):
            result.append(db)
        else:
            log.warning("Invalid database directory encountered: {}. Skipping!".format(db))
    return result


def list_files(bucket, database, **kwargs):
    """
    Returns a list of all files (no directories) for :database:

    :param database: A subdirectory inside the mysqlbackups also representing the database name
    """
    get_filename_regex = ".*/(.*)$"
    bucket_result = bucket.list("{}{}/".format(S3_MYSQL_BACKUP_FOLDER.format(appname=kwargs["appname"]), database), "/")
    files = [match(get_filename_regex, res.name).group(1) for res in bucket_result]

    result = []
    for f in files:
        if match(DB_BACKUP_REGEX, f):
            result.append(f)
        else:
            log.warning("Invalid database backup file encountered: {}. Skipping!".format(f))

    return result


def delete_file(bucket, backup, **kwargs):
    """
    Deletes a given file from the backups

    :param bucket: boto bucket object on which we operate
    :param MysqlBackup backup:  the database the file belongs to
    :param kwargs:  general kwargs. appname key is required for this function
    :return: None
    """
    bucket.delete_key("{}{}/{}".format(
        S3_MYSQL_BACKUP_FOLDER.format(appname=kwargs["appname"]), backup.database, backup.filename))


def create_s3_connection(s3_user, s3_pass, bucketname):
    """
    Creates an s3 boto connection and returns the bucket object

    :param s3_user: the AWS id
    :param s3_pass: the AWS pass
    :param bucketname: the bucket to connect to
    """

    conn = S3Connection(s3_user, s3_pass)
    log.debug("made connection to bucket {}".format(bucketname))
    return conn.get_bucket(bucketname)


def get_bucket_and_config():
    """
    Parse secrets and create a boto s3 bucket object using those secrets.

    :return: a tuple containting of:
        Boto bucket object
        kwargs dict
    """
    global __bucket_and_config
    if get_global_bucket_and_config() is None:
        config = parse_json_file(VARS_FILE)
        bucket = create_s3_connection(config['s3_user'], config['s3_pass'], config['backup_bucket_name'])
        set_global_bucket_and_config((bucket, config))
    return get_global_bucket_and_config()


def get_global_bucket_and_config():
    global __bucket_and_config
    return __bucket_and_config


def set_global_bucket_and_config(bucket_and_config):
    global __bucket_and_config
    __bucket_and_config = bucket_and_config


__bucket_and_config = None
