from argparse import ArgumentParser
import configparser
from MySQLdb import connect
from MySQLdb import OperationalError

from kamikaze3.importer.validators import get_mysql_major_version
from kamikaze3.mysqlfixer.settings import ACCESS_USERS, SET_CORRECT_SUPERUSER_PERMISSIONS, GET_ACTUAL_USER_PERMISSIONS, \
    EXPECTED_SUPERUSER_PERMISSIONS
from kamikaze3.mysqlfixer.logger import get_logger

log = get_logger()


"""
What do I do?

1. Ensure that the root, app, debian accounts have proper privileges
2. Ensure that the configured passwords on disk are correct

TODO:

- [ ] Restore root account if invalid (mysql restart with --skip-grant-tables)
- [ ] Recreate app/debian accounts if they got deleted
"""


def main():
    process_args()
    fix_mysql()


def process_args():
    parser = ArgumentParser(
        description='Fix the Mysql user permissions and ensure that '
                    'the configured passwords on disk are correct.'
    )
    parser.parse_args()


def fix_mysql():
    cursor = get_root_cursor()

    for user, cnf in ACCESS_USERS:

        if not are_user_permissions_correct(user=user, cursor=cursor):
            fix_mysql_user_permissions(user=user, cursor=cursor)

        if not user_can_login(user, cnf):
            fix_mysql_user_password(user, cnf, cursor)

    log.info("Fixing is finished!")


def change_mysql_password(cursor, user: str, password: str):
    """
    Change mysql password. Behaviour changes based on mysql version
    """
    mysql_version = get_mysql_major_version()
    log.warn("Setting password {} for {} on MySQL {}".format(password, user, mysql_version))
    if mysql_version == '5.6':
        rows_updated = cursor.execute(
            "UPDATE `mysql`.`user` SET `password` = PASSWORD(%s) WHERE `user` = %s", (password, user)
        )
    else:
        rows_updated = cursor.execute(
            "ALTER USER '{user}' IDENTIFIED BY '{passw}'".format(user=user, passw=password)
        )

    cursor.execute("FLUSH PRIVILEGES")
    return rows_updated


def fix_mysql_user_password(user, cnf, cursor):

    assert user != 'root', "safeguard: am not changing password for root"

    config = configparser.SafeConfigParser()
    config.read(cnf)
    password = config.get('client', 'password')

    assert len(password) >= 8, "password too short"
    rows_updated = change_mysql_password(cursor=cursor, user=user, password=password)

    return rows_updated


def user_can_login(user, cnf):
    try:
        get_mysql_conn(cnf)
    except OperationalError as e:
        # (1045, "Access denied for user 'debian-sys-maint'@'localhost' (using password: YES)")
        code, msg = e.args
        if code == 1045:
            log.warn("User {} cannot login".format(user))
            return False
        else:
            raise

    return True


def get_root_cursor():
    """
    Tries the various root accounts and returns a root cursor if found
    :return: MySQLdb.cursor with root privileges
    """
    log.info("Testing for root account")
    for user, cnf in ACCESS_USERS:
        try:
            cursor = get_mysql_conn(cnf)
            cursor.execute("SELECT * FROM user")  # simple test to check if we have superuser privileges
            log.info("Found root account: {}".format(user))
            return cursor
        except OperationalError:
            pass
    raise RuntimeError("No working root mysql credentials were found!")


def get_mysql_conn(my_cnf_location=None):
    """
    Login to mysql and return a cursor with credentials from :my_cnf_location:

    :param  str my_cnf_location: the path to the my.cnf file with login credentials
    :return: MySQLdb.cursor
    """
    db = connect(db="mysql", host="localhost", read_default_file=my_cnf_location)
    return db.cursor()


def fix_mysql_user_permissions(user, cursor):
    """
    Set's correct superuser equivalent privileges for :user:
    :param str user: the mysql username
    :param cursor: a MysqlDB cursor with root permissions
    """

    log.info("Fixing mysql permissions for user {}".format(user))

    if user == "app":
        cursor.execute("DELETE FROM user WHERE user = '{}' AND host != '%'".format(user))
        if cursor.rowcount != 0:
            log.info("Deleted %d excess entries for user {}".format(cursor.rowcount, user))
    cursor.execute(SET_CORRECT_SUPERUSER_PERMISSIONS, (user, ))
    cursor.execute("flush privileges")


def are_user_permissions_correct(user, cursor):
    """
    Checks if user's permissions are correctly configured

    :returns bool: True if permissions are correct, false if permissions are false
    """
    log.info("checking MySQL permissions of user: {}".format(user))

    cursor.execute(GET_ACTUAL_USER_PERMISSIONS, (user, ))
    result = cursor.fetchall()

    for permissions_tuple in result:
        if permissions_tuple != EXPECTED_SUPERUSER_PERMISSIONS:
            log.warning("MySQL permissions of user {} not what we expected!".format(user))
            return False

    log.info("MySQL permissions for user {} look good!".format(user))
    return True


if __name__ == "__main__":
    # execute only if run as a script
    main()
