import logging
import os

from logging import getLogger
from contextlib import suppress
from datetime import datetime
from dateutil.tz import tzutc
from subprocess import CalledProcessError, check_output

from kamikaze3.settings import USR_CNF
from kamikaze3.common.env import get_app_name
from kamikaze3.common.settings import ONBOARDING_STATE_PATH
from .checks import service_checks
from kamikaze3.privileged.servicectl import _try_post

logger = getLogger(__name__)
FLAG = '-n'
EVENTS_URL = None  # module level var, cache URL avoid multiple calls to USR_CNF


class ImproperlyConfigured(RuntimeError):
    pass


def _get_events_url():
    global EVENTS_URL
    if not EVENTS_URL:
        try:
            format_url = check_output((USR_CNF, FLAG)).decode('utf-8').strip()
            EVENTS_URL = format_url.format(get_app_name())
        except (FileNotFoundError, CalledProcessError):
            raise ImproperlyConfigured()
    return EVENTS_URL


def call_event_hook(events):
    """
    Call an event hook.
    :param dict|list events:
    """
    if isinstance(events, list):
        now = datetime.utcnow().replace(tzinfo=tzutc()).isoformat()
        events = {event: now for event in events}

    if not isinstance(events, dict):
        raise TypeError('events must be a list or dict, %s given' % events)

    try:
        url = _get_events_url()
    except ImproperlyConfigured:
        return

    logger.debug("Calling event hook with events: %s" % events)
    data = {
        'app': get_app_name(),
        'events': events
    }
    _try_post(url, data=data)


def try_call_event_hook(events):
    """
    Ignore errors when calling the event hook.
    :param dict|list events:
    """
    with suppress(Exception):
        call_event_hook(events)


def state_path_parent_exists():
    return os.path.isdir(os.path.dirname(ONBOARDING_STATE_PATH))


def ensure_working_directory_exists():
    if not os.path.isdir(ONBOARDING_STATE_PATH):
        os.mkdir(ONBOARDING_STATE_PATH)


def main():
    logging.basicConfig(level=logging.ERROR)
    logger.debug("Starting Onboarding script")

    if state_path_parent_exists():
        ensure_working_directory_exists()
        successful_checks = []
        for check_cls in service_checks:
            check = check_cls()
            success, success_time = check.execute()
            if success:
                successful_checks.append((check, success_time))

        if successful_checks:
            call_event_hook({c.name: t.isoformat() for c, t in successful_checks})
            for check, _ in successful_checks:
                check.mark_check_done()
    else:
        logger.debug(
            "The parent of the state path does not exist. "
            "Perhaps this is a Vagrant that has been messed "
            "with or a not fully provisioned Hypernode? Skipping"
        )
