from contextlib import suppress
from ctypes import cdll

from os.path import isfile
from psutil import NoSuchProcess, TimeoutExpired
from threading import Thread, Lock
import argparse
import gc
import logging
import os
import psutil
import sys
from time import time

from kamikaze3.onboarding import try_call_event_hook

logger = logging.getLogger(__name__)

# todo: re-enable this until we find out what's causing the socket binding problem on respawn
MASTER_PIDFILES = [
    '/var/run/nginx.pid',
]
MYSQL_PIDFILE = '/var/run/mysqld/mysqld.pid'
OOMKILLED_MARKER = '/data/web/.last_oomkilled'
APP_CGROUP = 'limited.slice'
MEM_PERCENT_THRESHOLD = 70
OOM_SCORE_ADJ_MIN = -1000

# If the permissive_memory_management file exists on disk (placed as a flag by hn-control
# if enabled for the app in the eav settings), then we will not preventively kill app user
# processes as quickly. For some shops this gives better results (if they have heavy one-time
# processes like CRONs that make the system go out of memory once in a while but not structurally).
# By default we have 'strict' memory management where we are quick to kill user level processes as
# soon as one out of memory event is detected to prevent an avalanche of memory hungry processes.
# Both settings can make sense depending on the shop, so YMMV. For people who experience trouble
# with one or the other configuration, they can try to switch and see if that makes a difference.
# If not, then the easiest solution is to just upgrade to a larger plan or find out what processes
# are using so much memory.
CONFIGURED_MAX_KILLS_PER_SECOND = 5 if isfile('/etc/hypernode/permissive_memory_management') else 1


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--verbose', action='store_true')
    args = parser.parse_args(args)

    logging.basicConfig(
        format='%(asctime)s %(levelname)s %(name)s: %(message)s',
        level=logging.DEBUG if args.verbose else logging.INFO,
    )
    logger.info("Starting OOMMonitor")
    sys.exit(start_app_oom_monitor())


def start_app_oom_monitor():
    OOMMonitor(cgroup=APP_CGROUP).run()


# We need eventfd (http://man7.org/linux/man-pages/man2/eventfd.2.html) to use
# the cgroups notification API, we could either 1. create a c extension (or use
# one like 'butter') or 2. use the one in the glibc shared library which is
# available on all our systems anyway. For more info see:
# http://sgros-students.blogspot.nl/2013/05/calling-eventfd-from-python.html
def eventfd(init_val, flags):
    libc = cdll.LoadLibrary("libc.so.6")
    return libc.eventfd(init_val, flags)


class AffectedProcesses(object):
    """Data structure to represent affected processes
    attributes:
      * master: list of master processes affected (killed, etc.)
      * app: list of app user processes affected (killed, etc.)
      * mysql: MySQL process if affected, or None
    """

    def __init__(self, master=None, app=None, mysql=None):
        """
        :param list master:
        :param list app:
        :param psutils.Process mysql:
        """
        self.master = master or []
        self.app = app or []
        self.mysql = mysql


class OOMMonitor(object):
    def __init__(self,
                 cgroup='',
                 kill_all_app_processes_delay=1,
                 max_kills_per_second=CONFIGURED_MAX_KILLS_PER_SECOND):
        self.event_control_path = '/sys/fs/cgroup/memory/{}/cgroup.event_control'.format(
            cgroup)
        self.oom_control_path = '/sys/fs/cgroup/memory/{}/memory.oom_control'.format(
            cgroup)
        self.kill_all_app_processes_delay = kill_all_app_processes_delay
        self.max_kills_per_second = max_kills_per_second
        self.kill_count = 0
        self.kill_stamp = 0
        self.kill_restart_stamp = 0
        # keep track of the time events are sent, to throttle sent events
        self.last_events_timestamp = {}
        self.send_events_cooldown_secs = 30
        self.cache = None
        self._thread_kill_mysql = None  # thread used to stop MySQL service
        self._lock_kill_mysql = Lock()

    def attempt_populating_emergency_memory_cache(self):
        if not self.cache:
            with suppress(MemoryError):
                self.cache = bytearray(2 * 1024 * 1024)  # 2M cache

    def dump_emergency_memory_cache(self):
        del self.cache
        # force the python gc to return the deallocated memory to the os
        # the interpreter does this on its own after 700 object deallocations
        # but we don't want to wait for that
        gc.collect()

        self.cache = None

    @staticmethod
    def try_kill(process):
        with suppress(NoSuchProcess):
            logger.info("killing process {}: {}".format(
                process.pid, process.name()))
            process.kill()

    @staticmethod
    def try_terminate(process):
        with suppress(NoSuchProcess):
            logger.info("terminating process {}: {}".format(
                process.pid, process.name()))
            process.terminate()

    @staticmethod
    def try_soft_kill(process, timeout=300):
        """Try to terminate the process, if after timeout process is not terminated
        then kill the process.
        """
        with suppress(NoSuchProcess):
            logger.info(
                "soft killing process {} ({}), with terminating".format(
                    process.name(), process.pid))
            process.terminate()
            try:
                process.wait(timeout)
            except TimeoutExpired:
                logger.info(
                    "Process {} ({}) was not terminated after {} seconds. killing the process!".
                    format(process.pid, process.name(), timeout))
                process.kill()

    @staticmethod
    def read_file_contents(filename):
        try:
            with open(filename) as f:
                return int(f.read())
        except (OSError, IOError):
            logger.error(
                "Can't read file, there is probably too little memory. Skipping"
            )

    @staticmethod
    def get_oom_score_adj(pid):
        """
        Get the oom_score_adj from the process. psutil does not provide an interface for this.
        The oom_score_adj influences how likely the process is to be oomkilled by the kernel
        in the out of memory killer. If the oom score is -1000 the kernel will never oomkill
        this process. In that case we should also not kill the process preventively here as
        it is 'protected'.
        For implementation details in the kernel
        see https://github.com/torvalds/linux/search?q=OOM_SCORE_ADJ_MIN&unscoped_q=OOM_SCORE_ADJ_MIN
        :param int pid: The pid to get the oom_score_adj for.
        :return int oom_score_adj: The oom_score_adj of the process.
        """
        try:
            return int(open('/proc/{}/oom_score_adj'.format(pid)).readline())
        # Broad exception clause because we never want this method to raise
        except Exception:
            # If we can not read the oom_score_adj for whatever reason assume the process is protected
            # When the system is super out of memory even reading the oom_score_adj here might not work
            # in that case a later iteration might get it if needed.
            return OOM_SCORE_ADJ_MIN

    def kill_processes(self):
        """Try to terminate/kill processes to free some memory.
        packs affected processes in an AffectedProcesses instance.

        :rtype: AffectedProcesses
        """
        self.dump_emergency_memory_cache()

        all_processes = list(psutil.process_iter())
        master_pids = [
            self.read_file_contents(pidfile) for pidfile in MASTER_PIDFILES
        ]

        mysql_pid = self.read_file_contents(MYSQL_PIDFILE)
        mysql_process = None
        affected_procs = AffectedProcesses()

        # kill master processes
        for process in all_processes:
            with suppress(NoSuchProcess):
                if process.pid == mysql_pid:
                    mysql_process = process
                    continue
                if process.pid not in master_pids:
                    continue
            self.try_terminate(process)
            affected_procs.master.append(process)

        # kill processes that belong to the "app" user
        for process in all_processes:
            with suppress(NoSuchProcess):
                if process.username() != 'app':
                    continue
                if self.get_oom_score_adj(process.pid) == OOM_SCORE_ADJ_MIN:
                    continue
            self.try_kill(process)
            affected_procs.app.append(process)

        if mysql_process and self.stop_mysql_if_necessary(mysql_process):
            affected_procs.mysql = mysql_process
        return affected_procs

    def stop_mysql_if_necessary(self, process):
        """Stop MySQL once per second
        We had situations where MySQL hogged most system resources for long time,
        and since OOM kill is disabled on MysQL, OOM kill was not helping at all.
        So if necessary, on OOM kill per second tries
        helping the system by ensuring it's not MySQL that's hogging system resources.

        Later systemd will notice MySQL is terminated and starts it again.

        :param psutil.Process process: the MySQL process
        :return: True if initiated to stop MySQL or False otherwise
        """
        # let first OOM kill try reclaiming resources. but if still low memory,
        # maybe it's MySQL hugging system resources.
        if self.kill_count != 2:
            return False

        if process.memory_percent() < MEM_PERCENT_THRESHOLD:
            return False

        if not self._lock_kill_mysql.acquire(False):
            logger.info(
                "skipped stopping MySQL, a previous attempt to stop MySQL is still ongoing"
            )
            return False

        self._thread_kill_mysql = Thread(
            target=self._soft_kill_mysql_and_unlock,
            args=(process, ),
            name='oom_monitor_stop_mysql')
        self._thread_kill_mysql.start()
        return True

    def _soft_kill_mysql_and_unlock(self, process):
        """Try soft killing MySQL and ensure the related lock is released.
        Note: Only call if the MySQL kill lock is acquired
        """
        try:
            self.try_soft_kill(process)
        finally:
            self._lock_kill_mysql.release()

    # oomkills in the last second
    def update_kill_count(self):
        cur_stamp = int(time())
        self.kill_count = (
            self.kill_count + 1) if self.kill_stamp == cur_stamp else 1
        return cur_stamp

    def kill_processes_if_needed(self):
        """Tries to kill master/app processes if necessary.
        Packs the result of the action in an AffectedProcesses instance.

        :rtype: AffectedProcesses
        """
        affected_procs = AffectedProcesses()
        if self.kill_count >= self.max_kills_per_second:
            logger.warning("Kill count of {} exceeded max kills per second {}"
                           "".format(self.kill_count,
                                     self.max_kills_per_second))
            if (self.kill_stamp -
                    self.kill_restart_stamp) > self.kill_all_app_processes_delay:
                self.kill_restart_stamp = self.kill_stamp
                affected_procs = self.kill_processes()
        return affected_procs

    @staticmethod
    def touch_oomkilled_flag():
        # If anything goes wrong writing the marker, ignore it.
        # This happens in low memory conditions so it's best effort.
        with suppress(Exception):
            with open(OOMKILLED_MARKER, 'w+'):
                os.utime(OOMKILLED_MARKER)

    def start_oom_event_listener(self):
        # We will use the cgroup notification API. 1. start eventfd subsystem. 2. pass 2 control file descriptors by
        # writing them as integers to the special event_fd socket. See:
        # https://access.redhat.com/documentation/en-US/Red_Hat_Enterprise_Linux/6/html/Resource_Management_Guide/sec-Using_the_Notification_API.html
        event_fd = eventfd(0, 0)
        cfd = os.open(self.event_control_path, os.O_RDWR)
        ofd = os.open(self.oom_control_path, os.O_RDONLY)
        # Note that there's no .format for bytestrings, there was a PEP for that but
        # that was rejected: https://www.python.org/dev/peps/pep-0461/ printf style
        # formatting it is then.
        os.write(cfd, b'%d %d' % (event_fd, ofd))
        return event_fd

    def should_send_handle_oomkill_event(self, events):
        """Return True if we should send handle OOMKill event or not.
        We throttle sending events to prevent creating many requests
        when the node is being OOMKilled too fast, consuming resources
        from the node itself and hammering our APIs

        :return:  bool
        """
        if not self.last_events_timestamp:
            return True

        for event in events:
            if event not in self.last_events_timestamp:
                return True
            last_event_timestamp = self.last_events_timestamp[event]
            return (time() - last_event_timestamp) > self.send_events_cooldown_secs

    def handle_oomkill(self):
        logger.info("Caught OOM event")
        self.kill_stamp = self.update_kill_count()
        affected_procs = self.kill_processes_if_needed()
        self.touch_oomkilled_flag()
        events = ['oom-killer-invoked']
        if affected_procs.mysql:
            events.append('oom-killer-slayed-mysql')
        if self.should_send_handle_oomkill_event(events):
            logger.info("Sending OOM handled event")
            try_call_event_hook(events)
            # record timestamp of sent events for future reference
            now = time()
            for event in events:
                self.last_events_timestamp[event] = now
        else:
            logger.info("Skipped sending OOM handled event")

    def run(self):
        event_socket = self.start_oom_event_listener()

        # loop over event listener by reading from socket
        while True:
            self.attempt_populating_emergency_memory_cache()
            os.read(event_socket, 8)  # events are 8 bytes
            self.handle_oomkill()
