import json
from ipaddress import IPv4Network
from kamikaze3.common.filesystem import read_file
from threading import RLock

INVENTORY_FILE = '/data/openvpn/openvpn_client_configs/inventory.json'
DEFAULT_CONTENT = {}
DEFAULT_CLIENT_CONFIG_NAME = 'openvpnclient'
OPENVPN_KEYS_PATH = '/data/openvpn/keys'
INV_LOCK = RLock()


def get_inventory(filepath=INVENTORY_FILE):
    try:
        with open(filepath, 'r') as inv:
            content = json.load(inv)
    except FileNotFoundError:
        # Assuming empty
        content = DEFAULT_CONTENT
    return content


def write_inventory(content, filepath=INVENTORY_FILE):
    """
    Write a dict of settings to the inventory file
    :param content: dict: the settings to write
    :param str filepath: The path to write to
    """
    with INV_LOCK:
        with open(filepath, 'w') as inv:
            json.dump(content, inv)


def add_to_inventory(name, value, filepath=INVENTORY_FILE):
    with INV_LOCK:
        content = get_inventory(filepath)
        content[name] = value
        write_inventory(content, filepath)


def delete_from_inventory(name, filepath=INVENTORY_FILE):
    """
    Delete a element from the inventory
    :param name: The name (key) of the element to delete
    :param str filepath: The path to delete from
    :return: True is successful, False if failure (lookup error)
    """
    with INV_LOCK:
        content = get_inventory(filepath)
        try:
            del content[name]
        except KeyError:
            return False
        write_inventory(content, filepath)
        return True


def get_inventory_ips():
    """
    Get the inventory and RESERVED_IPS and return that combined list
    :return: ips: list: the combined list with IP's
    """
    inventory = get_inventory()
    return calculate_reserved_ips() + list(inventory.values())


def get_next_available_ip():
    network = IPv4Network(get_client_subnet())
    ips = get_inventory_ips()
    hosts_iterator = (host for host in network.hosts() if str(host) not in ips)
    return str(next(hosts_iterator))


def get_client_certificate(name, path=OPENVPN_KEYS_PATH):
    """
    Gets TLS auth key, ca cert, client cert, client key from disk
    :param name: str: the name of the certificate (for exmaple customclient2.erikhyperdev.hypernode.io)
    :param path: str: the directory to get the keys from
    :return: str: ta key, ca cert, client cert, client key
    """
    ta = read_file('{}/ta.key'.format(path))
    ca = read_file('{}/ca.crt'.format(path))
    cert = read_file('{}/{}.crt'.format(path, name))
    key = read_file('{}/{}.key'.format(path, name))
    return ta, ca, cert, key


def calculate_reserved_ips():
    """
    We don't want people to use the first few IP's in the subnet, as that might collide with other connected users.
    """
    network = str(get_client_subnet().split('/')[0])
    first_octects = network.split('.')[0:3]
    reserved = [network]
    for item in range(1, 8):
        reserved.append('.'.join(first_octects) + '.{}'.format(str(item)))
    return reserved


def get_client_subnet():
    return '{}/24'.format(read_setting_from_file('/etc/openvpn/server.conf', 'server'))


def read_setting_from_file(filepath, setting_name, split_char=' '):
    """
    Read a setting from a file on disk, returns the value after the delimiter.
    :param filepath: str: The filepath of the settings file
    :param setting_name: str: The name of settings to read (key)
    :param split_char: str: The value to split on (default space)
    :return: str: The value of the setting
    """
    with open(filepath, 'rt') as fh:
        for line in fh:
            line_content = line.strip()
            if line_content.startswith(setting_name):
                check_list = line_content.split(split_char)
                if setting_name in check_list:
                    return line_content.split(split_char, 2)[1]
        raise KeyError("'{}' not found in file {}".format(setting_name, filepath))
