import logging
from kamikaze3.utils import request_confirmation
from kamikaze3.common.env import host_is_a_docker, host_is_a_vagrant, get_hypernode_name
from kamikaze3.openvpn_config_generator.client import generate_client_config, \
    generate_openvpn_client_cert, DEFAULT_PATH
from kamikaze3.openvpn_config_generator.inventory import get_next_available_ip, get_inventory_ips, get_client_subnet, \
    DEFAULT_CLIENT_CONFIG_NAME, OPENVPN_KEYS_PATH
from kamikaze3.openvpn_config_generator.client import remove_client_config, \
    remove_openvpn_client_config, place_openvpn_client_config
from kamikaze3.openvpn_config_generator.inventory import add_to_inventory, delete_from_inventory, \
    get_inventory
import kamikaze3.common.log
from argparse import ArgumentParser
from os.path import isfile
import os

logger = logging.getLogger(__name__)


def purge_openvpn_client_config(name=None):
    """
    Delete a openvpn client config and remove from inventory.
    :param name: str: The name of the custom client config to remove
    :return: None
    """
    name = name or DEFAULT_CLIENT_CONFIG_NAME
    hn_name = get_hypernode_name()
    # A config name is present, so we are removing a custom config
    logger.info('Removing {} openvpn client config'.format(name))
    if not remove_client_config(name, hn_name):
        logger.warning('FAIL: could not remove {}!'.format(name))
        exit(os.EX_IOERR)
    elif name != DEFAULT_CLIENT_CONFIG_NAME:
        if not remove_openvpn_client_config(name, hn_name):
            logger.critical('!!! Removed user client config, but unable to remove OpenVPN client config. '
                            'This should not happen. Operator please investigate !!!')
            exit(os.EX_IOERR)
        logger.info('Deleting {} from inventory'.format(name))
        if not delete_from_inventory(name):
            logger.warning('Could not delete {} from inventory!'.format(name))


def create_openvpn_client_config(name=None, ip=None):
    """
    Create a openpvn client config based on the given parameters and add to inventory
    :param name: str: The name of the openvpn client config, when None the default config will be generated
    :param ip: str: The IP that should be given to the custom config
    :return: None
    """
    hn_name = get_hypernode_name()
    if name:
        # A config name is present, so we are making a custom config
        # IP related logic
        if ip:
            # Check that the IP is in the the right subnet
            if get_client_subnet().split('.')[:-1] != ip.split('.')[:-1]:
                logger.critical('Given IP not in the {} subnet!'.format(get_client_subnet()))
                exit(os.EX_DATAERR)
            # Check if IP is not RESERVED or already used
            elif ip in get_inventory_ips():
                logger.critical('Given IP is already used or reserved, exiting...')
                exit(os.EX_DATAERR)
        else:
            ip = get_next_available_ip()
            logger.info('IP no specified, using next available which is {}'.format(ip))

        # Certificate generation
        # Check if certificate is already generated, if not generate it
        if not isfile('{}/{}.{}.crt'.format(OPENVPN_KEYS_PATH, name, hn_name)):
            logger.info('Calling pkitool to generate client certificate for {}'.format(name))
            if generate_openvpn_client_cert(name, hn_name) == 0:
                logger.info('Certificate generation for {} successful'.format(name))
            else:
                logger.warning('Could not generate certificate for {}, please investigate'.format(name))
                exit(os.EX_NOINPUT)
        else:
            logger.info('Client cert with name {} already exists ... skipping generation'.format(name))

        # Client config generation
        logger.info('Generating client config for {} with IP {}'.format(name, ip))
        if generate_client_config(name, hn_name):
            logger.info('Client config generation successful, it can be found in {}'.format(DEFAULT_PATH))
            # Finally place the Openvpn client config to assign the IP to the client on connection
            logger.info('Placing OpenVPN client config')
            if not place_openvpn_client_config(name, hn_name, ip):
                logger.critical('!!! Generated client certificate, but unable to place OpenVPN client config '
                                'client will not get the requested static IP !!!')
                exit(os.EX_SOFTWARE)
        else:
            logger.warning('Client config generation failed, please investigate')
            exit(os.EX_SOFTWARE)

        logger.info('Adding {} with IP {} to inventory'.format(name, ip))
        add_to_inventory(name, ip)
    else:
        # No config name is present, so we are making the default config
        logger.info('Generating default openvpn client config')
        generate_client_config(DEFAULT_CLIENT_CONFIG_NAME, hn_name)


def parse_args():
    parser = ArgumentParser(description='Hypernode openvpn client config management. '
                                        'Generate or delete openvpn client configs')
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-n', '--name', help='The Common Name (CN) of the client configuration. When no name is given '
                                            'the default config will generated/deleted.')
    group.add_argument('-a', '--purge-all', action='store_true', help='Purge all client configs, '
                                                                      'except the default')

    parser.add_argument('-p', '--purge', action='store_true', help='Delete the given client config')
    parser.add_argument('--ip', help='The IP to assign to a certain client config')
    parser.add_argument('-y', '--yes', action='store_true', help='Answer yes to all questions')
    return parser.parse_args()


def main():
    """
    Command entrypoint
    Calls purge_openvpn_client_config when --purge goven. Otherwise call create_openvpn_client_config.
    """
    logger.setLevel(logging.DEBUG)
    logger.addHandler(kamikaze3.common.log.create_stream_handler())
    logger.addHandler(kamikaze3.common.log.create_syslog_handler(
        name_override='OpenVPN_client_config_generator' if __name__ == '__main__' else None,
        level=logging.WARNING
    ))
    if host_is_a_docker() or host_is_a_vagrant():
        logger.critical('This command can only be executed on non virtual environments')
        exit(os.EX_USAGE)

    args = parse_args()
    if args.ip and not args.name:
        logger.warning('The default client config can not be assigned a IP. Continuing like argument IP was not given')

    if args.purge_all:
        if not args.yes:
            request_confirmation('You are about to delete all custom client configs, are you sure?')
        inventory = get_inventory()
        logger.info('Deleting {} custom client configs'.format(len(inventory)))
        for name, _ in inventory.items():
            purge_openvpn_client_config(name=name)
    elif args.purge:
        purge_openvpn_client_config(
            name=args.name,
        )
    else:
        create_openvpn_client_config(
            name=args.name,
            ip=args.ip,
        )
    logger.info('All done ... Have a secure day!')
    return os.EX_OK


if __name__ == '__main__':
    exit(main())
