from argparse import ArgumentParser
from subprocess import check_output, check_call


def get_php_versions():
    """
    Get a list of all the installed php versions on the Hypernode.
    :return list: list of the php versions e.g. ['7.3', '7.1']
    """
    return check_output(
        "find /lib/systemd/system/php*-fpm.service | grep -o '[0-9][\\.][0-9]\\+'",
        shell=True
    ).decode().strip().split('\n')


def process_args():
    """
    Process the arguments given to the php switch command
    :return ArgumentParser: parsed arguments for the switch php command
    """
    parser = ArgumentParser(
        description='Set the PHP version'
    )
    parser.add_argument(
        'version', help='The PHP version',
        choices=get_php_versions()
    )
    return parser.parse_args()


def disable_fpm_service(old_version):
    """
    Disable php fpm for a php version
    :param str old_version: the php version to disable fpm for
    """
    check_call(['systemctl', 'disable', 'php{}-fpm'.format(old_version)])


def stop_fpm_service(old_version):
    """
    Stop php fpm for a php version
    :param str old_version: the php version to stop fpm for
    """
    check_call(['systemctl', 'stop', 'php{}-fpm'.format(old_version)])


def update_alternatives(new_version):
    """
    Set the node to a new php version
    :param str new_version: the new php version to set for the node
    """
    new_php_location = check_output(['which', 'php' + new_version]).strip()

    check_call(['update-alternatives',
                '--set',
                'php',
                new_php_location,
                ])


def enable_fpm_service(new_version):
    """
    Enable php fpm for a php version
    :param str new_version: the php version to enable fpm for
    """
    check_call(['systemctl', 'enable', 'php{}-fpm'.format(new_version)])


def start_fpm_service(new_version):
    """
    Start php fpm for a php version
    :param str new_version: the php version to start fpm for
    """
    check_call(['systemctl', 'start', 'php{}-fpm'.format(new_version)])


def switch_php(new_version):
    """
    Switch the php version for this node
    :param str new_version: the php version to switch this node to
    """
    php_versions_to_disable = get_php_versions()
    php_versions_to_disable.remove(new_version)

    for old_version in php_versions_to_disable:
        disable_fpm_service(old_version)
        stop_fpm_service(old_version)

    update_alternatives(new_version)
    enable_fpm_service(new_version)
    start_fpm_service(new_version)


def main():
    """
    Main function for the php switch command
    """
    args = process_args()
    switch_php(new_version=args.version)


if __name__ == '__main__':
    main()
