#!/usr/bin/env python3
"""
hypernode.privileged.fail2ban_purge
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Tool to purge fail2ban state on Hypernodes
"""
import sys
from os import EX_OK, EX_SOFTWARE, EX_TEMPFAIL, EX_NOINPUT
import os.path
from argparse import ArgumentParser
from logging import getLogger, DEBUG
from kamikaze3.common.log import create_syslog_handler, create_stream_handler
from kamikaze3.common.sqlite import can_vacuum_database
from kamikaze3.common.filesystem import capture_file_size_change
from kamikaze3.system_services.fail2ban import purge_fail2ban_database


program_name = 'hypernode-fail2ban-purge'
__version__ = '0.1.0'
logger = getLogger(__name__)


def parse_args(args):
    parser = ArgumentParser(prog=program_name,
                            description='Purge Fail2ban state')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='print verbose output')
    parser.add_argument('--db', default='/var/lib/fail2ban/fail2ban.sqlite3',
                        help='fail2ban SQLite database path')
    return parser.parse_args(args)


def main(args=None):
    """Run the program accepting arguments or use sys.argv by default.

    :param list args: list of CLI args (default is sys.argv[1:])
    :returns: the exit code.
    :rtype: int
    """
    args = parse_args(args)

    logger.setLevel(DEBUG)
    logger.addHandler(
        create_syslog_handler(
            name_override=program_name if __name__ == '__main__' else None
        )
    )
    if args.verbose:
        logger.addHandler(create_stream_handler())

    return_code = EX_OK
    try:
        # we purge the database directly since current version of fail2ban itself
        # does not support purging yet: https://github.com/fail2ban/fail2ban/issues/1267
        if os.path.exists(args.db):
            can_vacuum = can_vacuum_database(args.db)
            with capture_file_size_change(args.db) as file_sizes:
                purge_fail2ban_database(args.db, vacuum=can_vacuum)
            size_before, size_after = file_sizes[0], file_sizes.pop()
            logger.info(
                'Purged fail2ban database. Freed {:.2f} MB disk, file size changed from {:.2f} to {:.2f} MB'.format(
                    (size_before - size_after), size_before, size_after))
            if not can_vacuum:
                logger.warning('There was not enough resources available to vacuum fail2ban database!')
                return_code = EX_TEMPFAIL
        else:
            logger.warning('Skipped purging fail2ban database. File "{}" does not exist'.format(args.db))
            return_code = EX_NOINPUT
    except (OSError, RuntimeError) as err:
        logger.exception('Failed to purge fail2ban database: {}'.format(err))
        return EX_SOFTWARE

    return return_code


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        sys.exit(EX_TEMPFAIL)
