import json
import re
from argparse import Namespace
from glob import glob
from logging import getLogger
from typing import Optional
from shutil import move, copy, copytree

import os

from kamikaze3.privileged.manage_vhosts import program_name, settings
from kamikaze3.privileged.manage_vhosts.manage_configs import delete_vhost
from kamikaze3.privileged.manage_vhosts.state import safe_write_root_file
from kamikaze3.utils import get_app_name

logger = getLogger(program_name)


def find_appname_in_dir(import_dir: str) -> Optional[str]:
    """
    Attempt to find the original appname in the imported nginx dir
    :param str import_dir: Imported app nginx directory
    :rtype: str|None
    """
    for file in glob(os.path.join(import_dir, '*'), recursive=True):
        if os.path.isfile(file):
            appname = find_appname_in_file(file)
            if appname:
                return appname
        else:
            match = settings.APPNAME_PATTERN.search(file)
            if match:
                return match.group(1)


def find_appname_in_file(file: str) -> Optional[str]:
    """
    Attempt to find the imported appname in the given file
    :param str file: Imported HMV statefile
    :rtype: str|None
    """
    with open(file, 'r') as fp:
        match = settings.APPNAME_PATTERN.search(fp.read())
        if match:
            return match.group(1)


def find_imported_appname(import_dir: str, import_statefile: str) -> Optional[str]:
    """
    Attempt to find the imported appname in the statefile. If not found,
    try to find it in the imported directory.
    :param str import_dir: Imported app nginx directory
    :param str import_statefile: Imported HMV statefile
    :rtype: str|None
    """
    original_appname = find_appname_in_file(import_statefile)
    if original_appname:
        return original_appname
    else:
        return find_appname_in_dir(import_dir)


def move_nginx_to_backup(timestamp):
    """
    Move current nginx dir to backup dir
    :param str timestamp: Vhost timestamp
    :rtype: None
    """
    if os.path.exists(settings.APP_NGINX_DIR):
        backup_nginx_dir = settings.APP_IMPORT_BACKUP_DIR.format(timestamp=timestamp)
        logger.info("Backing up {} to {}".format(settings.APP_NGINX_DIR, backup_nginx_dir))
        os.makedirs(os.path.dirname(backup_nginx_dir), exist_ok=True)
        move(settings.APP_NGINX_DIR, backup_nginx_dir)


def copy_statefile_to_backup(timestamp) -> Optional[str]:
    """
    Create a backup of the current statefile
    :param str timestamp: Vhost timestamp
    :return str|None: Path to backup statefile if there is a current statefile
    """
    if os.path.exists(settings.STATE_FILE_PATH):
        backup_statefile = settings.STATEFILE_IMPORT_BACKUP.format(timestamp=timestamp)
        logger.info("Backing up {} to {}".format(settings.STATE_FILE_PATH, backup_statefile))
        os.makedirs(os.path.dirname(backup_statefile), exist_ok=True)
        copy(settings.STATE_FILE_PATH, backup_statefile)
        return backup_statefile


def remove_all_vhosts(statefile_path: str):
    """
    Remove all vhosts defined in the given statefile
    :param str statefile_path: Path to statefile
    :rtype: None
    """
    if os.path.exists(statefile_path):
        with open(statefile_path, 'r') as fp:
            for sn in json.load(fp):
                delete_vhost(sn)


def replace_inline_str(filepath: str, old_str: str, new_str: str):
    """
    Find and replace all occurences of `old_str` in `filepath` and replace it with `new_str`
    :param str filepath: File to replace lines in
    :param str old_str: Pattern to replace
    :param str new_str: Substitute for the replaced pattern
    :rtype: None
    """
    with open(filepath, 'r') as fp:
        contents = fp.read()

    substituted_contents = re.sub(old_str, new_str, contents)
    with open(filepath, 'w') as fp:
        fp.write(substituted_contents)


def substitute_in_filecontents(old_str, new_str):
    """
    Replace substr in filecontents in nginx dir
    :param str old_str: Substr to replace
    :param str new_str: Str to place instead of old_str
    :rtype: None
    """
    for file in glob(os.path.join(settings.APP_NGINX_DIR, '*'), recursive=True):
        if os.path.isfile(file):
            replace_inline_str(file, old_str, new_str)


def substitute_in_dirnames(old_str, new_str):
    """
    Replace substr in dirnames of nginx dir
    :param str old_str: Substr to replace
    :param str new_str: Str to place instead of old_str
    :rtype: None
    """
    for file in glob(os.path.join(settings.APP_NGINX_DIR, '*'), recursive=True):
        if os.path.isdir(file):
            dst_file = re.sub(old_str, new_str, file)
            logger.debug("Moving {} to {}".format(file, dst_file))
            move(file, dst_file)


def substitute_in_filenames(old_str, new_str):
    """
    Replace substr in filenames of nginx dir
    :param str old_str: Substr to replace
    :param str new_str: Str to place instead of old_str
    :rtype: None
    """
    for file in glob(os.path.join(settings.APP_NGINX_DIR, '*'), recursive=True):
        if os.path.isfile(file):
            dst_file = re.sub(old_str, new_str, file)
            logger.debug("Moving {} to {}".format(file, dst_file))
            move(file, dst_file)


def replace_appname(import_appname: str):
    """
    Replace occurences of the given appname with the existing appname.
    This looks at the appname in config files, dir names and file names
    :param str import_appname: Imported appname
    :rtype: None
    """
    import_fqdn = '{}.hypernode.io'.format(import_appname)
    new_fqdn = '{}.hypernode.io'.format(get_app_name())
    logger.info("Replacing {} with {}".format(import_fqdn, new_fqdn))
    # Replace appname in /data/web/nginx file contents
    substitute_in_filecontents(import_fqdn, new_fqdn)

    # First rename dirs
    substitute_in_dirnames(import_fqdn, new_fqdn)

    # Rename files afterwards because the dir they're in can contain the import_fqdn
    substitute_in_filenames(import_fqdn, new_fqdn)

    # Replace appname in statefile
    replace_inline_str(settings.STATE_FILE_PATH, import_fqdn, new_fqdn)


def restore_file(filename, timestamp, nginx_directory):
    backup_nginx_dir = settings.APP_IMPORT_BACKUP_DIR.format(timestamp=timestamp)
    path = os.path.join(backup_nginx_dir, filename)
    try:
        copy(path, nginx_directory)
    except FileNotFoundError:
        logger.warning('File {} not found, skipping restore'.format(path))


def import_hmv(import_dir: str, import_statefile: str, args: Namespace, protected_files: list = None):
    # Find original appname
    import_appname = find_imported_appname(import_dir=import_dir, import_statefile=import_statefile)
    # Move current nginx dir to backup
    move_nginx_to_backup(args.updated_at)

    # Move current statefile to backup
    backup_statefile = copy_statefile_to_backup(args.updated_at)
    if backup_statefile:
        remove_all_vhosts(backup_statefile)

    # Copy import_directory to /data/web/nginx
    logger.info("Copying {} to {}".format(import_dir, settings.APP_NGINX_DIR))
    copytree(import_dir, settings.APP_NGINX_DIR)

    # Copy statefile to old statefile location
    logger.info("Copying {} to {}".format(import_statefile, settings.STATE_FILE_PATH))
    with open(import_statefile, 'r') as fp:
        safe_write_root_file(settings.STATE_FILE_PATH, fp.read())

    if import_appname:
        replace_appname(import_appname)

    if protected_files:
        logger.info("Copying protected files back from backup to {}...".format(settings.APP_NGINX_DIR))
        for f in protected_files:
            logger.info("Copying {} to {}".format(f, settings.APP_NGINX_DIR))
            restore_file(f, args.updated_at, settings.APP_NGINX_DIR)
