import re

from argparse import Action, ArgumentParser
from datetime import datetime
from logging import getLogger
from os import listdir
from os.path import isdir
from pytz import utc

from kamikaze3.privileged.manage_vhosts import program_name, settings
from kamikaze3.systemctl.list_backups import ensure_datetime_microseconds_stripped

logger = getLogger(program_name)


def check_conditions_before_starting():
    """
    Check required conditions before allowing this command to be run.
    If conditions are not met, return False. Else, return True.
    :rtype: bool
    """
    if not isdir(settings.ROOT_SITES_DIR):
        print("{} is not created. Check if the managed_vhosts_enabled setting is set in the app configuration.".
              format(settings.ROOT_SITES_DIR))
        return False
    return True


def parse_args(args):
    """
    Parse input from the CLI
    :param list args: Input arguments for testing
    :rtype: obj
    """
    class NegateAction(Action):
        """
        Helper class to deal with negating arguments.
        If set as action in an `add_argument` call, this will
        set False if the argument is prefixed with 'disable'.
        It will return True in all other cases.
        """
        def __call__(self, parser, ns, values, option=None):
            setattr(ns, self.dest, not option.startswith('--disable-'))

    parser = ArgumentParser(description='Manage the nginx config for one or more given servernames')

    servername_group = parser.add_mutually_exclusive_group(required=False)
    servername_group.add_argument('servernames', nargs="*", help='Servernames to manage the root nginx config for',
                                  default=[])
    servername_group.add_argument("--all", action='store_true',
                                  help="Apply the configuration for all vhosts defined in the statefile.")

    parser.add_argument("-v", "--verbose", help="Be more verbose", action='store_true')
    parser.add_argument("--varnish", "--disable-varnish", help="Manage varnish for the servername",
                        dest='varnish', action=NegateAction, nargs=0)
    parser.add_argument("--default-server", help="Use this servername as default server", action='store_true')
    parser.add_argument("--type", choices=[f for f in listdir(settings.SERVERNAME_CONFIG_TEMPLATE_DIR)
                                           if isdir(settings.SERVERNAME_CONFIG_TEMPLATE_DIR + f)],
                        help="Create configuration based on the given type")
    parser.add_argument("--delete", action='store_true', help="Delete the servername from the root nginx config")
    confirmation_group = parser.add_mutually_exclusive_group(required=False)
    confirmation_group.add_argument("-y", "--yes", action='store_true', help="Answer yes on any confirmation requests")
    confirmation_group.add_argument("-n", "--no", action='store_true', help="Answer no on any confirmation requests")
    ssl_group = parser.add_argument_group("SSL related configs")
    ssl_group.add_argument("--https", "--disable-https", dest='https', action=NegateAction, nargs=0,
                           help="Sets the use of https. Will create and use a letsencrypt certificate if there is "
                                "no other SSL certificate present")
    ssl_group.add_argument("--force-https", "--disable-force-https", dest='force_https', action=NegateAction, nargs=0,
                           help="Force http to be redirected to https")
    ssl_group.add_argument("--ssl-config", choices=['modern', 'intermediate'],
                           help="Use the given Mozilla SSL standard config")
    ssl_group.add_argument("--ssl-noclobber", action='store_true', help="Do not overwrite existing SSL certificate")
    listen_group = parser.add_argument_group("Listen ports")
    listen_group.add_argument("--port-http", type=int, help="Port that listens to http traffic")
    listen_group.add_argument("--port-https", type=int, help="Port that listens to https traffic")
    listen_group.add_argument("--port-http-staging", type=int,
                              help="Port that listens to http staging traffic")
    listen_group.add_argument("--port-https-staging", type=int,
                              help="Port that listens to https staging traffic")
    list_group = parser.add_argument_group("Information about active VHosts")
    list_group.add_argument("--list", metavar="servername", nargs="*",
                            help='Servernames to list. Will show all managed VHosts if none given')
    list_group.add_argument("--format", dest='list_format', choices=['table', 'json', 'raw'], default='table',
                            help='What way should the output be listed')
    import_group = parser.add_argument_group("Importer")
    import_group.add_argument("--import-statefile", help="Location of the imported statefile")
    import_group.add_argument("--protected-files", nargs='+', help="Files to protect: these will not be overwritten")
    import_group.add_argument("--import-dir", help="Location of the imported nginx directory")
    parsed_args = parser.parse_args(args)
    parsed_args.updated_at = ensure_datetime_microseconds_stripped(datetime.now(tz=utc).isoformat())

    if parsed_args.yes or parsed_args.no:
        parsed_args.automatic_confirmation = "yes" if parsed_args.yes else "no"
    else:
        parsed_args.automatic_confirmation = None
    return parsed_args


def strip_protocol_servername(servername):
    stripped_servername = re.sub(r'^https?://', '', servername)
    if stripped_servername != servername:
        logger.warning(
            "Found HTTP(S) protocol in servername. Stripping {} to {}".format(servername, stripped_servername)
        )
    return stripped_servername


def validate_servername(servername):
    return settings.SERVERNAME_PATTERN.match(servername)


def validate_input(args):
    """
    Validate input before placing configs
    :param obj args: Arguments from the argparser in the parse_args function.
    :rtype: bool
    """
    if args.list is not None:
        return True

    if not all([args.import_statefile, args.import_dir]) and any([args.import_statefile, args.import_dir]):
        logger.warning("When specifying either import_statefile or import_dir it is required to specify the other "
                       "component too")
        return False

    if all([args.import_statefile, args.import_dir]):
        logger.info("Importing files, assuming args 'all' and 'no'")
        args.all = True
        args.no = True
        args.automatic_confirmation = 'no'

    if not (args.servernames or args.all):
        logger.warning("Could not find an action to perform. Give a servername or specify --all to manage VHosts or "
                       "specify --list to get more information about existing VHosts")
        return False

    # Check if the given servername is a valid servername
    if args.servernames:
        for i, sn in enumerate(args.servernames):
            # Make sure that the args are mutated when stripping so that it's passed to the main function
            stripped_sn = strip_protocol_servername(sn)
            args.servernames[i] = stripped_sn

            if not validate_servername(stripped_sn):
                logger.warning("{} is not a valid servername".format(stripped_sn))
                return False

    if args.force_https and not args.https:
        logger.warning("Cannot force HTTPS if HTTPS is not enabled")
        return False

    if args.default_server and (args.all or len(args.servernames) > 1):
        logger.warning("Cannot set default_server for more than one VHost. Please specify one specific servername "
                       "when using the '--default-server' option.")
        return False

    if args.default_server and args.delete:
        logger.warning("You're not allowed to delete the default_server.")
        return False

    return True
