#!/usr/bin/env python3
"""
hypernode.privileged.manage_vhosts
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Tool to manage nginx vhosts on the host by placing nginx config files in an
idempotent manner. This will place both root nginx config files and default
app nginx config files if able.
"""
import subprocess
import sys
from contextlib import suppress
from os import EX_OK, EX_USAGE, EX_TEMPFAIL, remove
from logging import getLogger, INFO, DEBUG
from pathlib import Path

from kamikaze3.common.env import host_is_a_docker
from kamikaze3.common.log import create_syslog_handler, create_stream_handler
from kamikaze3.privileged.manage_vhosts import program_name, settings
from kamikaze3.privileged.manage_vhosts.default_server import manage_default_server
from kamikaze3.privileged.manage_vhosts.importer import import_hmv
from kamikaze3.privileged.manage_vhosts.input import check_conditions_before_starting, parse_args, validate_input
from kamikaze3.privileged.manage_vhosts.list import list_vhosts
from kamikaze3.privileged.manage_vhosts.manage_configs import delete_vhost, manage_http_config, manage_https_config, \
    manage_varnish_config
from kamikaze3.privileged.manage_vhosts.state import update_config, update_statefile, get_all_configs

logger = getLogger(program_name)


def manage_vhost(new_config, **kwargs):
    """
    Manage the entire VHost config on the node based on the given setup.
    :param dict new_config: VHost parameters
    :param dict kwargs: Any additional kwargs
    :return dict: Updated config
    """
    new_config = manage_http_config(new_config, **kwargs)
    new_config = manage_https_config(new_config)
    manage_varnish_config(new_config)

    # TODO: Manage ports
    return new_config


def trigger_nginx_config_reloader():
    """
    The nginx config reloader (https://github.com/ByteInternet/nginx_config_reloader)
    uses inotify to trigger a reload. We touch and delete a file to make sure the
    nginx config reloader is triggered.
    :return:
    """
    if host_is_a_docker():
        subprocess.check_call(['sudo', '/usr/bin/nginx_config_reloader'])
    else:
        Path(settings.NGINX_CONFIG_RELOAD_FLAG).touch()
        with suppress(FileNotFoundError):
            remove(settings.NGINX_CONFIG_RELOAD_FLAG)

    with suppress(FileNotFoundError):
        with open(settings.NGINX_CONFIG_RELOADER_ERROR_FILE) as f:
            error_content = f.read()
            logger.warning("The current NGINX config is not valid and will not be loaded:")
            logger.warning(error_content)


def main(args=None):
    if not check_conditions_before_starting():
        return EX_TEMPFAIL
    args = parse_args(sys.argv[1:] if args is None else args)

    logger.setLevel(DEBUG)
    logger.addHandler(create_syslog_handler())
    logger.addHandler(create_stream_handler(level=DEBUG if args.verbose else INFO))

    if not validate_input(args):
        return EX_USAGE

    if args.list is not None:
        return list_vhosts(args.list, args.list_format)

    if args.import_dir:
        import_hmv(
            import_dir=args.import_dir,
            import_statefile=args.import_statefile,
            args=args,
            protected_files=args.protected_files,
        )

    if args.all:
        all_configs = get_all_configs()
        args.servernames = all_configs.keys()

    if args.default_server:
        # We know that there is only one given servername thanks to validate_input
        args.servernames = manage_default_server(args.servernames[0])

    for servername in args.servernames:
        logger.info("Managing configs for {}".format(servername))
        new_config = update_config(servername, vars(args))
        if args.delete:
            delete_vhost(servername)
        else:
            new_config = manage_vhost(new_config, automatic_confirmation=args.automatic_confirmation)
        update_statefile(new_config)
    trigger_nginx_config_reloader()

    return EX_OK


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        sys.exit(EX_TEMPFAIL)
