import json
import subprocess
from copy import deepcopy
from logging import getLogger
from os import chmod

from kamikaze3.privileged.manage_vhosts import program_name, settings

# Statefile options
from kamikaze3.settings import HYPERNODE_PRIV_GROUP

logger = getLogger(program_name)


def get_all_configs():
    """
    Retrieve all current configs
    :rtype: dict[dict, ..]
    """
    try:
        with open(settings.STATE_FILE_PATH, 'r') as fp:
            return json.load(fp)
    except FileNotFoundError:
        return {}


def get_current_config(servername):
    """
    Get the active configuration for the given servername
    :param str servername: Servername that we configure a VHost for.
    :rtype: dict
    """
    try:
        with open(settings.STATE_FILE_PATH, 'r') as fp:
            config = json.load(fp)
            if servername in config:
                logger.debug("Found existing config for {}".format(servername))
                return config[servername]
            else:
                logger.info("No existing config for {}, starting with default options".format(servername))
                return settings.DEFAULT_VHOST_OPTIONS
    except FileNotFoundError:
        return deepcopy(settings.DEFAULT_VHOST_OPTIONS)


def _filter_config(config):
    """
    Remove any variables from the config that don't need to be saved
    in the statefile.
    :param dict config: Config that is to be saved in the statefile.
    :rtype: dict
    """
    # The default_server option is handled in manage_default_server
    filter_keys = ['servernames', 'default_server', 'all', 'list', 'list_format', 'yes', 'no', 'automatic_confirmation',
                   'verbose', 'import_dir', 'import_statefile']
    filter_values = [None]

    filtered_config = {}
    for k, v in config.items():
        if k not in filter_keys and v not in filter_values:
            filtered_config[k] = v
    return filtered_config


def _update_wildcard_options(config):
    """
    Adjust config if the servername has a wildcard.
    :param dict config: Config of the servername
    :rtype: dict
    """
    servername = config['servername']
    config['is_wildcard'] = servername.startswith('*')
    config['config_name'] = servername.replace('*', 'wildcard') if config['is_wildcard'] else config['servername']
    return config


def _is_legit_to_quit(servername):
    """
    Check if the servername can be safely deleted.
    :param str servername: Servername that is about to be deleted
    :rtype: bool
    """
    config = get_current_config(servername)
    return not config.get('default_server')


def update_config(servername, new_config):
    """
    Find differences between the current config and the config that is
    defined in the given args.
    Return the keys that changed with their new values defined in the args.
    :param str servername: Name that should match the config
    :param dict new_config: Config with new values
    :rtype: dict
    """
    current_config = get_current_config(servername)
    logger.debug("Current config: {}".format(current_config))
    # New config has the leading values
    current_config.update(_filter_config(new_config))

    # Add servername to config to make it easier for templating
    current_config['servername'] = servername
    current_config = _update_wildcard_options(current_config)
    return current_config


def update_statefile(new_servername_config):
    """
    Write the new config to the statefile. Create a statefile if it didn't exist yet.
    :param dict new_servername_config: Updated config for the servername
    :rtype: None
    """
    servername = new_servername_config['servername']
    current_config = get_all_configs()

    if new_servername_config.get('delete'):
        if not _is_legit_to_quit(servername):
            logger.info("Cannot delete default_server {}. Please first change the default_server to another VHost".
                        format(servername))
            return
        if servername in current_config:
            del current_config[servername]
        else:
            logger.info("Attempted to delete {} from the vhosts statefile while it wasn't defined.".
                        format(servername))
    else:
        current_config[servername] = new_servername_config

    safe_write_root_file(settings.STATE_FILE_PATH, json.dumps(current_config))


def safe_write_root_file(path, contents):
    """
    Write the given contents to a directory that is safe for the app user,
    and give it the right permissions for root before moving it to the correct
    location owned by root.
    :param str path: Path of the file
    :param str contents: Contents of the file
    :rtype: None
    """
    with open(settings.TMP_ROOT_FILE, 'w') as fp:
        fp.write(contents)
    chmod(settings.TMP_ROOT_FILE, 0o664)
    try:
        subprocess.check_call(["sudo", "/bin/chown", "root:{}".format(HYPERNODE_PRIV_GROUP), settings.TMP_ROOT_FILE])
        subprocess.check_call(["sudo", "/bin/mv", settings.TMP_ROOT_FILE, path])
    except subprocess.CalledProcessError as e:
        logger.exception("Could not properly write file to {}".format(path))
        raise e
