#!/usr/bin/env python3
"""
kamikaze3.privileged.postsuper
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This wrapper to postsuper command to manage postfix queues, accessible by app user
"""
import sys
import traceback
import re
from os import EX_OK, EX_TEMPFAIL, EX_DATAERR, EX_SOFTWARE, EX_UNAVAILABLE
from logging import getLogger, DEBUG, INFO, WARNING
from argparse import ArgumentParser
from subprocess import call, run
from kamikaze3.common.log import create_stream_handler, create_syslog_handler


PROGRAM_NAME = 'hypernode-postsuper'
SUDO = '/usr/bin/sudo'
POSTSUPER = '/usr/sbin/postsuper'
logger = getLogger(__name__)


def parse_args(args=None):
    parser = ArgumentParser(prog=PROGRAM_NAME,
                            description='Hypernode wrapper for postfix postsuper command')
    parser.add_argument('-v', '--verbose', action='store_true', help='Enable verbose logging')
    parser.add_argument('-d', '--delete', required=True,  # currently we only support deleting mails
                        help='Delete a message with the named queue ID, or ALL to clear queues'
                        ', set - to specify queue id from stdin')
    return parser.parse_args(args)


def validate_message_queue_id(msg_queue_id):
    """Validates a postfix message queue ID, raise if invalid

    :param str: msg_queue_id
    :raises: ValueError
    """
    if re.search(r'\W', msg_queue_id):
        raise ValueError('Invalid postfix message queue ID: {}'.format(msg_queue_id))


def clear_mailq(verbose=False):
    """Clear the mail queues,
    Returns True if postsuper command exited succesfully or False
    otherwise.

    :param bool verbose: If should enable verbose output from postsuper
    """
    call_proc_args = [SUDO, POSTSUPER]
    if verbose:
        call_proc_args.append('-v')
    call_proc_args.extend(['-d', 'ALL'])

    logger.warning('Clearing out postfix mail queues')
    logger.debug('Calling postsuper as: {}'.format(' '.join(call_proc_args)))

    ret_code = call(call_proc_args)
    return ret_code == 0


def delete_from_mailq(msg_queue_id, verbose=False):
    """Delete a single message from mail queues,
    Returns True if postsuper command exited succesfully or False
    otherwise.

    :param str msg_queue_id: queue ID of a message
    :param bool verbose: If should enable verbose output from postsuper
    :raises: ValueError if msg_queue_id is invalid
    """
    validate_message_queue_id(msg_queue_id)
    # postsuper is a setuid root command, we're going to run it with sudo
    # as hn-priv group. Since the queue id is a variable value, we can't
    # safe guard a fix value in sudoers file config file, so would have to use
    # wildcards in sudoers config file. To avoid those wildcards
    # we're going to send the queue_id to postsuper from stdin, so the actual
    # executed command by sudo stays the same, and we avoid the wildcards.
    call_proc_args = [SUDO, POSTSUPER]
    if verbose:
        call_proc_args.append('-v')
    call_proc_args.extend(['-d', '-'])  # postsuper would read queue_id from stdin

    logger.info('Removing postfix message queue ID "{}"'.format(msg_queue_id))
    logger.debug('Calling postsuper as: {} (stdin="{}")'.format(' '.join(call_proc_args), msg_queue_id))

    proc = run(call_proc_args, input=msg_queue_id.encode('utf-8'))  # input should be bytes
    return proc.returncode == 0


def main(args=None):
    """Run program with specified arguments. Returns a proper code as exit code.

    :param Iterable args: lis of CLI args
    :return int: exit code
    """
    opts = parse_args(args)
    verbose = opts.verbose

    logger.setLevel(DEBUG if verbose else INFO)
    logger.addHandler(create_stream_handler(level=INFO if verbose else WARNING))
    logger.addHandler(
        create_syslog_handler(
            name_override=(PROGRAM_NAME if __name__ == '__main__' else None)
        )
    )

    exit_code = EX_OK
    try:
        if opts.delete == 'ALL':
            if not clear_mailq(verbose):
                exit_code = EX_UNAVAILABLE
        else:
            queue_id = sys.stdin.readline().strip() if opts.delete == '-' else opts.delete
            if not delete_from_mailq(queue_id, verbose):
                exit_code = EX_UNAVAILABLE
    except ValueError as error:
        logger.error('{}'.format(error))
        logger.debug(traceback.format_exc())
        exit_code = EX_DATAERR
    except KeyboardInterrupt:
        logger.info('processs interrupted')
        exit_code = EX_TEMPFAIL
    except Exception as error:
        logger.error('Uncaught error: {}'.format(error))
        logger.debug(traceback.format_exc())
        exit_code = EX_SOFTWARE

    return exit_code


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
