#!/usr/bin/env python3
"""
hypernode.privileged.servicectl
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Tool to control services running on the host, while limiting the options
and making sure the operation is logged.

Users can use this tool to do allowed action (restart/reload) on a limited
set of services.
"""
import socket
import sys
from os import EX_OK, EX_USAGE, EX_SOFTWARE, EX_TEMPFAIL, EX_UNAVAILABLE
from datetime import datetime
from argparse import ArgumentParser
from logging import getLogger, DEBUG
from subprocess import check_output, CalledProcessError, PIPE, STDOUT
import requests
from os.path import isfile

from kamikaze3.common.api import get_hypernode_api_token
from kamikaze3.settings import USR_CNF
from kamikaze3.common.log import create_syslog_handler
from kamikaze3.common.env import (get_app_name, host_is_a_vagrant, is_varnish_enabled,
                                  is_persistent_redis_enabled, get_active_php_version_info,
                                  host_is_a_docker, is_elasticsearch_enabled, is_supervisor_enabled,
                                  is_rabbitmq_server_enabled)


__version__ = '0.1.0'

logger = getLogger(__name__)

program_name = 'hypernode-servicectl'
all_actions = ('restart', 'reload')


def get_active_php_version():
    """
    Return the active PHP version on the system
    :return str php_version: The active php version.
    For example 7.1
    """
    try:
        php_version_info = get_active_php_version_info()
        return '.'.join(php_version_info[0:2])
    except RuntimeError:
        return '7.1'


available_services = {
    'mysql': ('restart',),
    'redis-server': ('restart',),
    'supervisor': ('restart', 'reload'),
    'nginx': ('restart', 'reload'),
    'php{}-fpm'.format('.'.join(get_active_php_version().split('.')[0:2])): ('restart', 'reload')
}

if is_varnish_enabled():
    available_services['varnish'] = ('restart', 'reload')

if is_persistent_redis_enabled():
    available_services['redis-server-persistent'] = ('restart',)

if is_supervisor_enabled():
    available_services['supervisor'] = ('restart', 'reload')

if is_elasticsearch_enabled():
    available_services['elasticsearch'] = ('restart',)

if is_rabbitmq_server_enabled():
    available_services['rabbitmq-server'] = ('restart',)


FLAG = '-c'


def _try_post(url, data=None):
    response = requests.post(url, headers={
        'Authorization': 'Token {}'.format(get_hypernode_api_token()),
        'Content-Type': 'application/json',
        'Accept': 'application/json',
        'User-Agent': '{}:{}'.format(
            socket.getfqdn(), 'hypernode-kamikaze3'
        )
    }, json=data)
    response.raise_for_status()
    return response.json()


def send_event(event):
    event['time_utc'] = int(datetime.utcnow().timestamp())
    format_url = check_output([USR_CNF, FLAG], stderr=PIPE).decode('utf-8').strip()
    url = format_url.format(get_app_name())
    return _try_post(url, data=event)


def do_action_on_service(action, service):
    logger.info('running action "{}" on service "{}" ...'.format(action, service))

    action_performed = False
    event = {'message': '', 'success': False, 'action': action, 'service': service}

    try:
        systemctl_output = check_output(['sudo', 'systemctl', action, service], stderr=STDOUT)
        action_performed = True
        message = 'performed action "{}" on service "{}"'.format(action, service)
        print(systemctl_output.decode('utf-8'), end='')
    except CalledProcessError as exp:
        systemctl_output = exp.output.decode('utf-8').strip()
        message = 'failed to perform action "{}" on service "{}". command "{}" exited with "{}": "{}"'.format(
            action, service, ' '.join(exp.cmd), exp.returncode, systemctl_output)
        logger.exception(message)
        print(systemctl_output)
    except Exception as exp:
        message = 'error while running action "{}" on service "{}": {}'.format(action, service, exp)
        logger.exception(message)
    else:
        logger.info(message)

    event['success'] = action_performed
    event['message'] = message

    if host_is_a_vagrant() or host_is_a_docker():
        logger.info("event: {}".format(event))
    else:
        logger.info("sending event: {}".format(event))
        send_event(event)
        logger.info("event sent successfully: {}".format(event))

    return action_performed


def validate_action(action):
    if action == '':
        raise ValueError('no action specified')
    if action not in all_actions:
        raise ValueError('invalid action')


def validate_services(services):
    if len(services) < 1:
        raise ValueError('no services specified')
    for service in services:
        if service not in available_services:
            raise ValueError('service "{}" is not available'.format(service))


def action_allowed_on_service(action, service):
    return action in available_services.get(service, ())


def parse_args(args):
    parser = ArgumentParser(prog=program_name,
                            description='Control Hypernode services')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))
    parser.add_argument('action', nargs='?', default='',
                        help='desired action on services ({})'.format(
                            '|'.join(all_actions)))
    parser.add_argument('service', nargs='*', default=[],
                        help='target service, ({})'.format(
                            '|'.join(sorted(available_services.keys()))))
    return parser.parse_args(args)


def main(args=None):
    args = parse_args(sys.argv[1:] if args is None else args)
    action = args.action.lower().strip()
    try:
        validate_action(action)
    except ValueError as exp:
        print('{}. use -h or --help for instructions'.format(exp), file=sys.stderr)
        return EX_USAGE

    services = [srv.strip() for srv in args.service]
    try:
        validate_services(services)
    except ValueError as exp:
        print('{}.'.format(exp), file=sys.stderr)
        return EX_USAGE

    for service in services:
        if not action_allowed_on_service(action, service):
            print(
                'service "{}" does not support action "{}". choices: {}'.format(
                    service, action, ', '.join(available_services.get(service, []))),
                file=sys.stderr)
            return EX_USAGE

    logger.setLevel(DEBUG)
    logger.addHandler(
        create_syslog_handler(
            name_override=program_name if __name__ == '__main__' else None
        )
    )

    # During xgrades (and other important operations) we set
    # the /etc/nologin file to prevent anyone except for root
    # from logging in. When this file is set we should not allow
    # people to restart the services because that might mess with
    # whatever operation is in progress. For example, after stopping
    # the services during an xgrade we perform a final filesync. If
    # at that moment any service is (re)started orders might be
    # processed on the source node that would then not be copied
    # anymore to the destination node.
    if isfile('/etc/nologin'):
        print("Not allowed to restart or reload services at this time")
        return EX_UNAVAILABLE

    for service in services:
        try:
            do_action_on_service(action, service)
        except Exception:
            message = 'unhandled error while running action ' \
                      '"{}" on service "{}"'.format(action, service)
            print(message, file=sys.stderr)
            logger.exception(message)
            return EX_SOFTWARE

    return EX_OK


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        sys.exit(EX_TEMPFAIL)
