import logging
import json
from argparse import ArgumentParser
from os import EX_DATAERR

from kamikaze3.common.filesystem import get_file_hashes
from kamikaze3.common.system import user_is_root
from kamikaze3.security.slack_security import try_send_slack_security_message
from kamikaze3.common.env import get_hypernode_name
from kamikaze3.common.log import create_stream_handler

logger = logging.getLogger(__name__)


def parse_arguments():
    parser = ArgumentParser(description="Check sensitive files")
    mutexgroup = parser.add_mutually_exclusive_group(required=True)
    mutexgroup.add_argument(
        "--json-string",
        help="json of the sensitive file names and their hashes",
        type=str
    )
    mutexgroup.add_argument(
        "--json-file",
        help="The sensitive file list can also be passed as a path to a json file with this argument",
        type=str
    )
    parser.add_argument(
        "--slack",
        help="Post the results in a defined slack channel (only works as root)",
        type=str
    )
    return parser.parse_args()


def main():
    """
    Check the integrity of sensitive files and alert if it changed
    :return:
    """
    args = parse_arguments()
    logger.setLevel(logging.INFO)
    logger.addHandler(create_stream_handler())

    if args.json_string:
        try:
            sensitive_file_list = json.loads(args.json_string)
        except json.decoder.JSONDecodeError as e:
            logger.exception(e)
            return EX_DATAERR
    else:
        with open(args.json_file, "rt") as f:
            file_contents = f.read()
        sensitive_file_list = json.loads(file_contents)

    dirty_files = check_sensitive_files(sensitive_file_list)

    if len(dirty_files) > 0:
        if args.slack and user_is_root():
            try_send_slack_security_message(
                _craft_slack_message_sensitive_files_warning(), channel=args.slack)
        print_message = _craft_print_message_sensitive_files_warning(dirty_files)
    else:
        print_message = "Scan successful, all matches up :)"
    logger.info(print_message)


def check_sensitive_files(sensitive_file_list):
    dirty_files = []
    sensitive_file_names = sensitive_file_list.keys()
    file_hashes = get_file_hashes(sensitive_file_names)
    for file_path in sensitive_file_names:
        if sensitive_file_list[file_path] != file_hashes[file_path]:
            dirty_files.append(file_path)
    return dirty_files


def _craft_slack_message_sensitive_files_warning():
    return "There are sensitive files that do not match the given hash on Hypernode '" \
           + get_hypernode_name() + "'."


def _craft_print_message_sensitive_files_warning(dirty_files):
    message = "There are sensitive files that do not match on this hypernode."
    message += "\nThe following files have hashes that do not match:\n"
    for dirty_file in dirty_files:
        message += dirty_file + "\n"
    return message
