import os
import json
import logging
from argparse import ArgumentParser
from os import EX_OK, EX_DATAERR

from kamikaze3.common.filesystem import is_risky_sxid_file, get_file_hashes, \
    compare_file_hashes_to_earlier_defined_list
from kamikaze3.common.env import get_hypernode_name
from kamikaze3.security.slack_security import try_send_slack_security_message
from kamikaze3.common.log import create_stream_handler
from kamikaze3.common.system import user_is_root

logger = logging.getLogger(__name__)


def parse_arguments():
    parser = ArgumentParser(description="Run SUID and SGID audit on the hypernode")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--json-string",
        help="json of the legal file names and their hashes",
        type=str
    )
    group.add_argument(
        "--json-file",
        help="The sxid list can also be passed as a path to a json file with this argument",
        type=str
    )
    group.add_argument(
        "--generate",
        help="Generate a list of legal setuid and setgid files for hypernode control",
        action="store_true"
    )
    parser.add_argument(
        "--startpath",
        help="The path for the scanner to scan the node or for the generate to generate a list for",
        default="/",
        type=str
    )
    parser.add_argument(
        "--storepath",
        help="Where to store the generated setuid and setgid file. Default: $(pwd)/file_hashes.json",
        default=os.getcwd() + "/file_hashes.json",
        type=str
    )
    parser.add_argument(
        "--slack",
        help="Post the results in the slack channel (only works as root)",
        type=str
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    logger.setLevel(logging.INFO)
    logger.addHandler(create_stream_handler())

    # if the user wants to generate a list
    if args.generate:
        generate_sxid_list(startpath=args.startpath, storepath=args.storepath)
        return EX_OK

    # code will only reach here if the user wants to scan something
    if args.json_string:
        try:
            legal_sxid_list = json.loads(args.json_string)
        except json.decoder.JSONDecodeError as e:
            logger.exception(e)
            return EX_DATAERR
    else:
        with open(args.json_file, "rt") as f:
            file_contents = f.read()
        legal_sxid_list = json.loads(file_contents)

    illegal_files, dirty_files = scan_system_for_sxid(legal_sxid_list, args.startpath)

    if len(illegal_files) > 0 or len(dirty_files) > 0:
        if args.slack and user_is_root():
            try_send_slack_security_message(_craft_slack_message_sxid_warning(), channel=args.slack)
        print_message = _craft_print_message_sxid_warning(illegal_files, dirty_files)
    else:
        print_message = "Scan successful, all matches up :)"
    logger.info(print_message)
    return EX_OK


def generate_sxid_list(startpath="/", storepath=os.getcwd() + "/file_hashes.json"):
    """
    Generate a dict of the legal setuid/setgid files and store this list in
     a json file

    :param startpath: path to scan, by default the whole system
    :param storepath: where to store the final sxid list
    """
    sxid_files = find_sxid_files(startpath)
    sxid_file_hashes = get_file_hashes(sxid_files)
    sxid_file_hashes_json = json.dumps(sxid_file_hashes)
    with open(storepath, 'wt') as f:
        f.write(sxid_file_hashes_json)


def scan_system_for_sxid(legal_sxid_list, startpath="/"):
    """
    Scan the system for illegal setuid and setgid files
    :param legal_sxid_list:
    :param startpath:
    :return:
    """
    sxid_files = find_sxid_files(startpath)
    illegal_files, dirty_files = compare_file_hashes_to_earlier_defined_list(legal_sxid_list, sxid_files)
    return illegal_files, dirty_files


def find_sxid_files(startpath='/'):
    """
    Retrieve a list with the SXID files on the system
    :param startpath: Path to start scanning for sxid files
    :return: array of the SXID files on the system
    """
    sxid_files = []
    walker = os.walk(startpath)
    for root, dirs, files in walker:
        names = dirs + files
        for name in names:
            path = os.path.join(root, name)
            if os.path.isfile(path) and not os.path.islink(path):
                if is_risky_sxid_file(path):
                    sxid_files.append(path)
    return sxid_files


def _craft_slack_message_sxid_warning():
    """
    This function formats the message for slack
    :return: the formatted message
    """
    message = ":whip: Illegal or dirty setuid or setgid files have been found on hypernode '" \
              + get_hypernode_name() + "'. Check up required."
    return message


def _craft_print_message_sxid_warning(illegal_files, dirty_files):
    """
    This function formats the message for the print function
    :return: the formatted message
    """
    message = "On hypernode '" + get_hypernode_name() + "' the following has been found:\n\n"
    if len(illegal_files) > 0:
        message += "The following files seem to be illegal backdoors:\n"
        for file_path in illegal_files:
            message += file_path + "\n"
    if len(dirty_files) > 0:
        message += "The following files are supposed to have a SXID set but the hashes don't match:\n"
        for file_path in dirty_files:
            message += file_path + "\n"
    return message
