from argparse import ArgumentParser

from kamikaze3.systemctl.block_attack.parsing import setup_block_attack_parsing
from kamikaze3.systemctl.attach_backup.parsing import setup_attach_backup_parsing
from kamikaze3.systemctl.create_backup.parsing import setup_create_backup_parsing
from kamikaze3.systemctl.list_backups.parsing import setup_list_backups_parsing
from kamikaze3.systemctl.settings.parsing import _check_valid_attribute_values, setup_settings_parsing
from kamikaze3.systemctl.show.parsing import setup_show_parsing
from kamikaze3.systemctl.whitelist.parsing import setup_whitelist_parsing
from kamikaze3.systemctl.preinstall.parsing import setup_preinstall_parsing
from kamikaze3.utils import request_confirmation

REQUIRES_VALIDATION = {
    'settings': {
        'mysql_version': {
            '8.0': "WARNING: Upgrading to MySQL 8.0 will cause irreversible changes! This might cause "
                   "incompatibilities with some applications. For example, only from Magento 2.4 and later it has "
                   "been officially compatible. Are you sure you want to upgrade to MySQL 8.0? You can't go back "
                   "without contacting support."
        }
    }
}


def requires_validation(parsed_args):
    """
    Check if the parsed arguments require validation
    :param Namespace parsed_args: Parsed arguments from argparser
    :rtype: str|None
    """
    if parsed_args.command in REQUIRES_VALIDATION:
        requires_validation_command = REQUIRES_VALIDATION[parsed_args.command]
        if parsed_args.attribute and parsed_args.attribute in requires_validation_command:
            requires_validation_attribute = requires_validation_command[parsed_args.attribute]
            new_value = parsed_args.positional_value or parsed_args.deprecated_value
            if new_value and new_value in requires_validation_attribute:
                return requires_validation_attribute[new_value]


def parse_arguments(args=None):
    """
    Parse the hypernode-configure arguments
    :param list args: Optional, defaults to sys.argv
    :return obj args: The parsed arguments
    """
    parser = ArgumentParser()
    parser.add_argument(
        '--verbose', '-v',
        help='Display debug output',
        action='store_true'
    )

    sub_parsers = parser.add_subparsers(dest='command')
    setup_settings_parsing(sub_parsers)
    setup_show_parsing(sub_parsers)
    setup_whitelist_parsing(sub_parsers)
    setup_block_attack_parsing(sub_parsers)
    setup_attach_backup_parsing(sub_parsers)
    setup_list_backups_parsing(sub_parsers)
    setup_create_backup_parsing(sub_parsers)
    setup_preinstall_parsing(sub_parsers)

    parsed_args = parser.parse_args(args)

    if parsed_args.command == 'settings' \
            and (('positional_value' in parsed_args and parsed_args.positional_value)
                 or ('deprecated_value' in parsed_args and parsed_args.deprecated_value)):
        _check_valid_attribute_values(parsed_args)
        parsed_args.value = parsed_args.positional_value or parsed_args.deprecated_value

    if not parsed_args.command:
        parser.print_help()
        parser.exit()

    validation_string = requires_validation(parsed_args)
    if validation_string:
        request_confirmation(validation_string)

    return parsed_args
