import re
import socket
import json
import sys
import logging
from contextlib import contextmanager
from multiprocessing.pool import Pool
from time import time, sleep
from fcntl import lockf, LOCK_EX, LOCK_NB, LOCK_UN
from kamikaze3.settings import COMMON_INFO_PATH


log = logging.getLogger(__name__)

WAIT_FOR_LOCK_TIMEOUT = 10


def request_confirmation(message=None, prompt="Continue? (yes/no) ", func=sys.exit, args=None, kwargs=None):
    """
    Prompt user for confirmation before continuing program execution.

    :param message:
        An optional message to print before showing the prompt.
    :param prompt:
        The prompt to display.
    :param func:
        The function that is called when the user answers NO
    :param args:
        Positional arguments passed to the above function.
    :param kwargs:
        Keyword arguments passed to the above function.
    """
    if args is None and func == sys.exit:
        args = [1]
    elif args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    if message is not None:
        print(message)

    response = None
    while True:
        if response in ('y', 'yes'):
            return True
        elif response in ('n', 'no'):
            func(*args, **kwargs)
            return False
        elif response is not None:
            print("Please answer yes or no.")
        response = input(prompt).lower()


lock_socket = dict()


def get_lock(name=None):
    # Get lock using Linux sockets, slightly modified from:
    # http://stackoverflow.com/a/7758075/604515
    global lock_socket

    name = name or sys.argv[0]

    if name in lock_socket:
        log.warning("This process already issued a lock with name '{}', program error!".format(name))
        return False

    lock_socket[name] = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        lock_socket[name].bind('\0' + name)
        return True
    except OSError:
        return False


@contextmanager
def async_pool(max_procs=3):
    log.debug("Starting thread pool for parallel execution")
    pool = Pool(processes=max_procs)
    try:
        yield pool
    except KeyboardInterrupt:
        pool.terminate()
    finally:
        pool.close()
        pool.join()


def wait_for(resource_getter, attribute_getter=None, value=None, error_value=None, timeout=60, check_interval=1,
             expect_excs=None, stop_excs=None):
    """
    Wait until a resource becomes available.

    `resource_getter` is repeatedly called to retrieve an updated version of the
    resource. The function will return if it is successfully retrieved.

    To only return if the resource has reached a specific state, pass a `value` to
    specify what the state should be. If the value should be checked against a property
    of the resource, pass an `attribute_getter` that takes the resource and returns
    the state to check.

    If the state becomes `error_value`, the wait loop will exit with a PollingError.

    The wait loop will also stop if the resource_getter raises an exception in
    `stop_excs`. Exceptions from `expect_excs` will be ignored.

    :raises TimeoutException: if timeout is exceeded
    :raises Error: if error value is returned

    :param resource_getter: Function to retrieve resource
    :param attribute_getter: Function to retrieve attribute from resource
    :param value: Value of attribute to wait for
    :param error_value: Value of attribute that signals error
    :param int timeout: Wait timeout in seconds
    :param int check_interval: Interval between polling checks
    :param list[Exception] expect_excs: List of exceptions to ignore
    :param list[Exception] stop_excs: List of exceptions that signal stop of polling
    :return: The updated resource
    """
    if attribute_getter is None:
        def attribute_getter(x):
            return x

    # NB: catching multiple exceptions at once must be done with a tuple. Since PyCharm can't do
    # type hinting for a tuple, we request a list to be passed in, and convert it to a tuple.
    # See http://www.itmaybeahack.com/book/python-2.6/html/p02/p02c07_exceptions.html#basic-exception-handling
    stop_excs = tuple(stop_excs) if stop_excs else tuple()
    expect_excs = tuple(expect_excs) if expect_excs else tuple()

    start = time()
    while True:
        log.debug("Trying to get resource")
        try:
            upd_resource = resource_getter()
        except stop_excs:
            log.debug("Could not fetch updated resource", exc_info=True)
            break
        except expect_excs as exc:
            log.debug("Got expected exception: %s", str(exc))
        except StopIteration:
            log.debug('No more items in iterator')
        else:
            log.debug("Got resource: %s" % (upd_resource,))
            result = attribute_getter(upd_resource)
            if value is None or result == value:
                return upd_resource
            if result == error_value:
                raise PollingError("Resource fell into error state")
        sleep(check_interval)
        if time() - start > timeout:
            raise TimeoutException()


class TimeoutException(RuntimeError):
    pass


class PollingError(RuntimeError):
    pass


def camel_to_snake(string):
    """
    Converts camel-case string to snake-case string.

    Example: aCamelCaseString -> a_camel_case_string

    Source: https://gist.github.com/jaytaylor/3660565

    :param string: string in camel-case
    :return: snake-case string
    """
    underscorer1 = re.compile(r'(.)([A-Z][a-z]+)')
    underscorer2 = re.compile('([a-z0-9])([A-Z])')
    subbed = underscorer1.sub(r'\1_\2', string.replace(" ", ""))

    return underscorer2.sub(r'\1_\2', subbed).lower()


def try_get_lock(lock_filename):
    """
    Try to get an exclusive file lock
    :param str lock_filename: Name of the file to get the lock on
    :return obj fd: The lock file object
    """
    try:
        fd = open(lock_filename, 'w')
        # Locking with NB (no blocking) so then we can log that we're still
        # waiting for the lock in the wait_for in exclusive_lock, I think
        # this will make it more clear what is going on. For example, when
        # this is used in hypernode-oom-protect and it would block with a
        # blocking lock without any logging, then the user might assume
        # that the command that they are trying to run is stuck instead
        # of that we're waiting for the exclusive lock.
        lockf(fd, LOCK_EX | LOCK_NB)
    except IOError:
        log.debug('Failed to get file lock on {}'.format(lock_filename))
        raise
    return fd


def unlock(fd):
    """
    Unlock the exclusive lock file
    :param param fd: The lock file object
    :param obj fd: The lock file object to unlock
    :return None:
    """
    lockf(fd, LOCK_UN)


@contextmanager
def exclusive_lock(lock_filename, timeout=WAIT_FOR_LOCK_TIMEOUT):
    """
    Block until the lock has been acquired
    :param str lock_filename: The file to get the lock on
    :param int timeout: How long to wait for the lock before timing out
    :return None:
    """
    fd = wait_for(
        lambda: try_get_lock(lock_filename),
        expect_excs=[IOError],
        timeout=timeout,
        check_interval=1
    )
    try:
        yield
    finally:
        unlock(fd)


def parse_app_name_from_host_name():
    """
    Try parsing the app name from the hostname
    Returns None if hostname can not be parsed
    :return str | None: app_name
    """
    hostname_components = socket.gethostname().split('-')
    if len(hostname_components) > 1:
        return hostname_components[1]


def parse_short_handle_from_host_name():
    """
    Try parsing the shorthandle from the hostname
    Returns None if hostname can not be parsed
    :return str | None: app_name
    """
    hostname_components = socket.gethostname().split('-')
    if len(hostname_components) > 1:
        return hostname_components[0]


def get_app_name():
    """
    Get the app name from the node on which this script is running
    :str return: app_name
    """
    try:
        # try to get the app name from the common.json file, this is the most accurate
        with open(COMMON_INFO_PATH, 'r') as info:
            return json.load(info)['tag']
    except (IOError, ValueError, KeyError):
        # when the common.json file is not available (app user can not access it)
        # or there is no tag in the file try parsing the app name from the hostname
        return parse_app_name_from_host_name()


def noop(*args, **kwargs):
    pass
